async function digestMessage(message) {
  const msgUint8 = new TextEncoder().encode(message); // encode as (utf-8) Uint8Array
  const hashBuffer = await crypto.subtle.digest("SHA-256", msgUint8); // hash the message
  const hashArray = Array.from(new Uint8Array(hashBuffer)); // convert buffer to byte array
  const hashHex = hashArray
    .map((b) => b.toString(16).padStart(2, "0"))
    .join(""); // convert bytes to hex string
  return hashHex;
}

async function digestMatch(digest, domain) {
  const domainDigest = await digestMessage(domain);
  return domainDigest === digest;
}

document.addEventListener("DOMContentLoaded", function () {
  // Digest modal only exists on admin:
  const digestModal = document.getElementById("digestModal");
  if (!digestModal) {
    return;
  }

  const resultFlash = document.getElementById("resultFlash");
  const digestError = document.getElementById("digestError");
  const digestErrorMessage = document.getElementById("digestErrorMessage");
  const modalTitle = digestModal.querySelector(".modal-title");
  const modalBodyDigestInput = digestModal.querySelector(
    ".modal-body input#digest",
  );
  const modalBodySuggestion = digestModal.querySelector(
    ".modal-body input#suggestion",
  );

  const modalSaveButton = digestModal.querySelector(
    ".modal-footer #digestSave",
  );

  if (
    !digestError ||
    !digestErrorMessage ||
    !modalTitle ||
    !modalBodyDigestInput ||
    !modalBodySuggestion ||
    !modalSaveButton
  ) {
    throw new Error("Missing element for the domain modal");
  }

  digestModal.addEventListener("show.bs.modal", function (event) {
    // reset error
    function resetError() {
      digestError.classList.add("d-none");
      digestErrorMessage.textContent = "";
    }

    resetError();

    // Button that triggered the modal
    const button = event.relatedTarget;
    // Extract info from data-bs-* attributes
    const digest = button.getAttribute("data-bs-digest");
    const domain = button.getAttribute("data-bs-domain");

    // If necessary, you could initiate an AJAX request here
    // and then do the updating in a callback.
    //
    // Reset the modal's content.
    modalTitle.textContent = "Unknown Domain " + domain;
    modalBodySuggestion.value = "";
    modalBodyDigestInput.value = digest;

    // prevent default behavior on enter key and reset error state on keypress
    modalBodySuggestion.addEventListener("keypress", function (event) {
      if (event.key === "Enter") {
        event.preventDefault();
      }
      resetError();
    });
  });

  modalSaveButton?.addEventListener("click", async function (event) {
    event.preventDefault();

    digestError.classList.remove("d-none");

    const digest = modalBodyDigestInput.value;
    const domain = modalBodySuggestion.value;

    digestMatch(digest, domain).then(function (match) {
      if (!match) {
        // show error
        if (digestError.classList.contains("alert-success")) {
          digestError.classList.remove("alert-success");
        }

        digestError.classList.add("alert-danger");
        digestErrorMessage.textContent = "No match!";

        return;
      }

      // show success
      if (digestError.classList.contains("alert-danger")) {
        digestError.classList.remove("alert-danger");
      }

      // getCsrfToken is a global from client.js
      // @ts-ignore
      const csrfToken = getCsrfToken();
      if (!csrfToken) {
        digestError.classList.add("alert-danger");
        digestErrorMessage.textContent =
          "Invalid CSRF token, please reload the page";

        return;
      }

      digestError.classList.add("alert-success");
      digestErrorMessage.textContent = "Your suggestion is a match! Adding...";

      fetch(`/admin/match/${digest}`, {
        method: "POST",
        headers: {
          "Content-Type": "application/x-www-form-urlencoded",
          "X-CSRF-Token": csrfToken,
        },
        body: new URLSearchParams({
          domain: domain,
        }),
      })
        .then(async (res) => [res, await res.json()])
        .then(([response, json]) => {
          if (response.status === 200) {
            // close modal
            const modal = bootstrap.Modal.getInstance(digestModal);
            modal.hide();

            resultFlash.classList.remove("d-none");
            resultFlash.textContent = `Added new domain: ${json.result.domain}`;

            document.getElementById(`domain-${digest}`)?.remove();
          } else {
            if (digestError.classList.contains("alert-success")) {
              digestError.classList.remove("alert-success");
            }

            resultFlash?.classList.add("d-none");

            digestError.classList.add("alert-danger");
            digestErrorMessage.textContent = json.error;
          }
        })
        .catch(function (error) {
          if (digestError.classList.contains("alert-success")) {
            digestError.classList.remove("alert-success");
          }

          digestError.classList.add("alert-danger");
          digestErrorMessage.textContent =
            "Error communicating with the server. Please try again later. If this problem persists, please contact the administrator";
        });
    });
  });
});
