import env from "./env.js";
import { decryptString } from "./crypter.js";
import { getServerConfig } from "./queries.js";
import {
  CLIENT_NAME,
  DOMAIN_LIMIT,
  ENABLE_PUBLIC_COMMENTS,
} from "./constants.js";
import { USER_AGENT, serverFetch } from "./fetcher.js";
import { formatPrivateComment } from "./util.js";
import { captureException } from "@sentry/node";

function getRequestHeaders(token) {
  const headers = {
    Accept: "application/json",
    "User-Agent": USER_AGENT,
  };

  if (token) {
    headers["Authorization"] = "Bearer " + token;
  }

  return headers;
}

export const OAUTH_CALLBACK_PATH = "/oauth/callback";

// URL for the OAuth Redirect Callbacks:
const OAUTH_REDIRECT_URI = new URL(OAUTH_CALLBACK_PATH, env.PUBLIC_URL).href;
// Client access token scopes:
const OAUTH_SCOPES = [
  "read",
  "read:accounts",
  "admin:write:domain_blocks",
  // in order to downgrade suspends to a silence when there are existing followers on a server, we technically need admin:read:measures permissions
  // which doesn't exist yet, so we would have to request admin:read which is....a much bigger ask. At the cost of having a server suspension defederate
  // your server without warning, we can thus request a more narrow scope of permissions
  "admin:read:domain_blocks",
];

/**
 * Custom error class for handling duplicate domain block scenarios.
 * @extends Error
 */
export class DuplicateDomainBlockError extends Error {
  /**
   * Creates a new DuplicateDomainBlockError.
   * @param {string} message - The error message.
   * @param {DomainBlock} existingBlock - The existing domain block that caused the duplication error.
   */
  constructor(message, existingBlock) {
    super(message);
    this.existingBlock = existingBlock;
  }
}

// abstracts all the talking we're gonna do to the Mastodon api
export const api = {
  /**
   * @typedef ClientCredentials
   * @type {object}
   * @property {string} client_id
   * @property {string} client_secret
   */

  /**
   * Registers a new OAuth 2.0 application with the specified Mastodon server.
   * This function also ensures the application is "immortal" by calling get_client_credentials.
   *
   * @param {string} server - The domain of the Mastodon server to register with.
   * @param {import("pino").Logger} logger - A Pino logger instance for logging the registration process.
   * @returns {Promise<ClientCredentials>} A promise that resolves to an object containing client_id and client_secret.
   * @throws {Error} If the registration fails or if the server response is missing required fields.
   */
  registerApplication: async function (server, logger) {
    logger.info({ server }, "registering new oauth application");

    // FIXME: Accept serverUrl instead of server:
    const serverUrl = "https://" + server;

    const body = new FormData();
    body.set("client_name", CLIENT_NAME);
    body.set("redirect_uris", OAUTH_REDIRECT_URI);
    body.set("scopes", OAUTH_SCOPES.join(" "));

    const registration = await fetch(new URL("/api/v1/apps", serverUrl), {
      method: "POST",
      headers: getRequestHeaders(),
      body,
    }).then((res) => {
      return res.json();
    });

    if (registration.error) {
      logger.error(
        { error: registration.error },
        "Error registering oauth application",
      );
      throw new Error(
        "Failed to create oauth application (server returned an error)",
      );
    }

    if (!registration.client_id || !registration.client_secret) {
      logger.error(
        "Error registering oauth application. response missing client_id or client_secret properties",
      );
      throw new Error(
        "Failed to create oauth application (server returned an error)",
      );
    }

    const credentials = {
      client_id: registration.client_id,
      client_secret: registration.client_secret,
    };

    // calling this endpoint essentially makes the app immortal
    // preventing mastodon's oauth application vacuuming from
    // affecting our applications:
    await api.get_client_credentials(server, credentials);

    return credentials;
  },

  /**
   * Returns the authorization url for the OAuth 2.0 redirect:
   * @param {string} serverUrl
   * @param {string} clientId
   * @param {string} stateParam - A unique state parameter for CSRF protection
   * @param {boolean} forceLogin - Whether to force a login prompt even if a session exists.
   * @returns {string} The authorization URL.
   */
  authorizeUrl: function authorizeUrl(
    serverUrl,
    clientId,
    stateParam,
    forceLogin = false,
  ) {
    const redirect_url = new URL("/oauth/authorize", serverUrl);
    const oauthRedirectParams = new URLSearchParams({
      client_id: clientId,
      response_type: "code",
      force_login: forceLogin.toString(),
      redirect_uri: OAUTH_REDIRECT_URI,
      state: stateParam,
      scope: OAUTH_SCOPES.join(" "),
    });

    redirect_url.search = `?${oauthRedirectParams.toString()}`;

    return redirect_url.href;
  },

  /**
   * Adds a new domain block to a Mastodon server.
   *
   * @async
   * @param {string} server - The domain of the Mastodon server.
   * @param {string} access_token - The OAuth access token for authentication.
   * @param {string} domain - The domain to be blocked.
   * @param {string} severity - The severity level of the block (e.g., 'silence', 'suspend').
   * @param {string} [public_comment] - An optional public comment for the block.
   * @param {Object} account - The account object, likely containing information about the user performing the action.
   * @returns {Promise<Object>} The response data from the server.
   * @throws {DuplicateDomainBlockError} If the domain is already blocked on the server.
   * @throws {Error} If there's an unspecified error from the server.
   */
  add_domain_block: async (
    server,
    access_token,
    domain,
    severity,
    public_comment,
    account,
  ) => {
    const params = {
      domain,
      severity,
      public_comment,
      private_comment: formatPrivateComment(),
    };
    if (!ENABLE_PUBLIC_COMMENTS) {
      delete params["public_comment"];
    }
    const body = JSON.stringify(params);

    console.log(`Adding block entry for ${domain} at ${server}...`);

    const location = "https://" + server + "/api/v1/admin/domain_blocks";
    const headers = getRequestHeaders(access_token);
    headers["Content-Type"] = "application/json";
    const settings = {
      method: "POST",
      body,
      headers,
    };

    const fetchResponse = await serverFetch(
      server,
      location,
      settings,
      account,
    );
    const data = await fetchResponse.json();

    if (fetchResponse.status === 422) {
      // add this existing block to the database for this server
      if (data.existing_domain_block) {
        throw new DuplicateDomainBlockError(
          `A domain block for ${domain} already exists`,
          data.existing_domain_block,
        );
      }
    } else if (data.error) {
      console.error("Unspecified error", data.error);
    }

    return data;
  },

  /**
   * Updates an existing domain block on a Mastodon server.
   */
  update_domain_block: async (
    server,
    access_token,
    id,
    severity,
    public_comment,
    account,
  ) => {
    try {
      const params = {
        severity,
        private_comment: formatPrivateComment(),
      };
      if (ENABLE_PUBLIC_COMMENTS) {
        params["public_comment"] = public_comment;
      }
      const body = JSON.stringify(params);

      const location =
        "https://" + server + "/api/v1/admin/domain_blocks/" + id;
      const headers = getRequestHeaders(access_token);
      headers["Content-Type"] = "application/json";
      const settings = {
        method: "PUT",
        body,
        headers,
      };

      const fetchResponse = await serverFetch(
        server,
        location,
        settings,
        account,
      );

      const data = await fetchResponse.json();

      if (!fetchResponse.ok || data.error) {
        console.error(
          `Failed to push domain block: ${data.error || "Unknown error"}`,
        );
        // flash(
        //   "error",
        //   `Failed to push domain block: ${data.error || "Unknown error"}`,
        // );
        return data;
      }

      // flash("success", "Domain block updated successfully");
      return data;
    } catch (error) {
      console.error("Error updating domain block:", error);
      // flash(
      //   "error",
      //   "An unexpected error occurred while pushing the domain block",
      // );
      throw error;
    }
  },

  test_domain_blocks: async (server, access_token, logger) => {
    try {
      const res = await fetch(
        new URL("/api/v1/admin/domain_blocks?limit=1", `https://${server}`),
        {
          method: "GET",
          headers: getRequestHeaders(access_token),
        },
      );

      return res.status === 200;
    } catch (err) {
      logger.error(
        { error: err.toString(), stack: err.stack, server },
        "Error testing domain blocks API",
      );

      captureException(err);

      return false;
    }
  },

  get_domain_blocks: async (
    server,
    access_token,
    isAdmin,
    url,
    limit,
    account,
    req,
  ) => {
    const api_target = isAdmin ? "admin" : "instance";
    const regExp = /\<([^)]+)\>/;
    const thisLimit = limit || DOMAIN_LIMIT;

    const location = url
      ? url
      : "https://" +
        server +
        `/api/v1/${api_target}/domain_blocks?limit=` +
        thisLimit;
    // query how many domains we're blocking
    const settings = {
      method: "GET",
      headers: getRequestHeaders(isAdmin ? access_token : undefined),
    };

    let blocks = [];
    try {
      const fetchResponse = await serverFetch(
        server,
        location,
        settings,
        account,
      );
      const data = await fetchResponse.json();
      const linkHeader = fetchResponse.headers.get("link");

      if (data.error) {
        console.error("Error retrieving blocks", data.error);
        // Notify the user about the error
        // flash("error", "we cant get the block list from your s.");
        return data;
      }

      blocks = data;

      // if we're at a *user* limit requested, then stop
      if (limit && data.length >= limit) {
        // do nothing, in other words, and return what we have
        //  console.debug("Returning early with user-supplied limit.");
        // if we're at the pagination limit, look for more results from pagination header
      } else if (data.length === DOMAIN_LIMIT && linkHeader) {
        // parse out the next link (prev)
        const linkArray = linkHeader.split(",");
        const next = linkArray[0];
        const matches = regExp.exec(next);

        if (matches && matches[1]) {
          const nextUrl = matches[1];
          const newdata = await api.get_domain_blocks(
            server,
            access_token,
            isAdmin,
            nextUrl,
            undefined,
            account,
            req,
          );
          if (newdata.length > 0) {
            blocks.push(...newdata);
          }
        }
      }
    } catch (error) {
      console.error("Error fetching domain blocks", error);
      
    }

    // alpha sort blocks
    return blocks.sort((a, b) => (a.domain > b.domain ? 1 : -1));
  },

  get_access_token: async (client_id, client_secret, server, code) => {
    const body = new FormData();
    body.set("grant_type", "authorization_code");
    body.set("code", code);
    body.set("client_id", client_id);
    body.set("client_secret", client_secret);
    body.set("redirect_uri", OAUTH_REDIRECT_URI);

    const response = await fetch(new URL("/oauth/token", `https://${server}`), {
      method: "POST",
      headers: getRequestHeaders(),
      body,
    });

    if (response.status !== 200) {
      throw new Error("Failed to exchange authorization_code for access token");
    }

    return await response.json();
  },

  revoke_access_token: async (
    client_id,
    client_secret,
    server,
    token,
    account,
  ) => {
    const body = new FormData();
    body.set("client_id", client_id);
    body.set("client_secret", client_secret);
    body.set("token", token);

    const location = "https://" + server + "/oauth/revoke";
    const settings = {
      method: "POST",
      body,
      headers: getRequestHeaders(),
    };

    const fetchResponse = await serverFetch(
      server,
      location,
      settings,
      account,
    );
    const data = await fetchResponse.json();

    return data;
  },

  remove_domain_block: async (id, server, access_token, account) => {
    const location = "https://" + server + "/api/v1/admin/domain_blocks/" + id;
    const headers = getRequestHeaders(access_token);
    const settings = {
      method: "DELETE",
      headers,
    };

    const response = await serverFetch(server, location, settings, account);
    const delete_data = await response.json();

    return delete_data;
  },

  verify_account_credentials: async (server, access_token, logger) => {
    const location =
      "https://" + server + "/api/v1/accounts/verify_credentials";
    const headers = getRequestHeaders(access_token);
    const settings = {
      method: "GET",
      headers,
    };

    const fetchResponse = await serverFetch(server, location, settings);

    if (!fetchResponse.ok) {
      logger.error(
        {
          status: fetchResponse.status,
          headers: fetchResponse.headers,
          server: server,
        },
        "Error verifying account credentials",
      );
      return false;
    }

    const data = await fetchResponse.json();

    return data;
  },

  verify_app_credentials: async (server, access_token, account, logger) => {
    // Requires read:accounts scope for whatever reason or else
    // returns { error: 'This action is outside the authorized scopes' }
    const location = "https://" + server + "/api/v1/apps/verify_credentials";
    const headers = getRequestHeaders(access_token);
    const settings = {
      method: "GET",
      headers,
    };

    try {
      const fetchResponse = await serverFetch(
        server,
        location,
        settings,
        account,
      );

      if (!fetchResponse.ok) {
        logger.error(
          {
            status: fetchResponse.status,
            headers: fetchResponse.headers,
            server: server,
          },
          "Error verifying app credentials",
        );
        return false;
      }

      const data = await fetchResponse.json();

      if (!data || data?.error) {
        logger.error({ error: data?.error }, "Error verifying app credentials");
        return false;
      }

      return true;
    } catch (err) {
      next(err);
      logger.error(
        { error: err.toString(), stack: err.stack },
        "Error verifying app credentials",
      );
      return false;
    }
  },

  get_client_credentials: async (server, credentials) => {
    // Requires read:accounts scope for whatever reason or else
    // returns { error: 'This action is outside the authorized scopes' }

    const body = new FormData();
    body.set("client_id", credentials.client_id);
    body.set("client_secret", credentials.client_secret);
    body.set("grant_type", "client_credentials");

    const fetchResponse = await fetch(
      new URL("/oauth/token", `https://${server}`),
      {
        method: "POST",
        headers: getRequestHeaders(),
        body,
      },
    );

    return await fetchResponse.json();
  },
};
