import "../sentry.js";

import express from "express";

import RedisStore from "connect-redis";
import session from "express-session";
import { flashMiddleware, flash } from "@prairielearn/flash";
import cookieParser from "cookie-parser";
import { doubleCsrf } from "csrf-csrf";
import bodyParser from "body-parser";

import pino from "pino-http";
import { stdSerializers } from "pino-http";

import helmet from "helmet";
import rateLimit from "express-rate-limit";
import basicAuth from "express-basic-auth";

import env from "./env.js";
import { paths } from "./paths.js";

import { ViewEngine } from "./view_engine.js";

import { generateSecretKey } from "./crypter.js";
import { redis } from "./redis.js";

import { friendlySeverity } from "./util.js";
import {
  getDomainBlocks,
  deleteServerBlock,
  getBlocklist,
  isServerBlockRedundant,
} from "./blocklist.js";

import * as queries from "./queries.js";
import * as constants from "./constants.js";

import { adminRouter } from "./routes/admin.js";
import { sessionsRouter } from "./routes/sessions.js";
import * as SettingsController from "./controllers/SettingsController.js";
import * as AdminController from "./controllers/AdminController.js";
import { flashTemplateMiddleware } from "./middleware/flashTemplateMiddleware.js";
import { errorMiddleware } from "./middleware/errorHandler.js";
import { authenticationMiddleware } from "./middleware/authenticationMiddleware.js";
import { OAUTH_CALLBACK_PATH } from "./api.js";
import { ServerRepository } from "./repositories/ServerRepository.js";
import { DomainsRepository } from "./repositories/DomainsRepository.js";
import { createApiKey } from "./queries.js";
import publicApiRouter from "./routes/dashDNIApi.js";
import * as DomainChangeController from "./controllers/DomainChangeController.js";
import requestIp from 'request-ip';

const app = express();

// Add request-ip middleware early in the stack, before other middleware
app.use(requestIp.mw({
  // Configure request-ip to trust proxy headers
  attributeName: 'clientIp', // Default is 'clientIp'
  proxy: true // Trust proxy headers
}));

// Trust proxy settings
if (env.isProduction) {
  // In production, trust all proxies since we're behind DO's proxy
  app.set('trust proxy', true);
  
  // Enable detailed proxy logging in production temporarily
  app.use((req, res, next) => {
    console.log('Production Proxy Debug:');
    console.log('Headers:', JSON.stringify(req.headers, null, 2));
    console.log('Client IP:', req.clientIp);
    console.log('Express IP:', req.ip);
    console.log('IPs:', req.ips);
    console.log('CF-Connecting-IP:', req.headers['cf-connecting-ip']);
    console.log('X-Forwarded-For:', req.headers['x-forwarded-for']);
    console.log('X-Real-IP:', req.headers['x-real-ip']);
    next();
  });
} else {
  // In development, also trust all proxies
  app.set('trust proxy', true);
}

// Disable x-powered-by header in all environments
app.set('x-powered-by', false);

app.get("/health", function (_req, res) {
  res.send("ok");
});


// Serves client resources from assets folder
app.use(
  "/static",
  express.static(paths.assets, {
    // Prevent requests to `/static` falling through to the rest of the
    // application, preventing bugs:
    fallthrough: false,
  }),
);

const views = new ViewEngine(paths.views);
views.install(app);

views.env.addGlobal("version", constants.VERSION);
views.env.addGlobal("isDevelopment", env.isDevelopment);
views.env.addGlobal("push_frequency", constants.RUNNER_SCHEDULE_IN_MINUTES);

app.use((req, res, next) => {
  res.locals.path = req.path;
  next();
});

app.use(
  pino({
    redact: {
      paths: [
        "req.query.state",
        "req.query.code",
        "req.headers.cookie",
        "req.headers.authorization",
        "res.headers['set-cookie']",
      ],
    },
    serializers: {
      /**
       * Request log sanitizer to prevent logging access tokens in URLs
       * @param {import('http').IncomingMessage} req
       */
      req: function sanitizeRequestLog(req) {
        const log = stdSerializers.req(req);

        // If we have a url on the request and it is the OAuth Callback, then we
        // replace the URL just with the OAUTH_CALLBACK_PATH to redact the
        // secret values from that URL
        if (
          typeof log.url === "string" &&
          log.url.startsWith(OAUTH_CALLBACK_PATH)
        ) {
          log.url = OAUTH_CALLBACK_PATH;
        }

        return log;
      },
    },
  }),
);

// set up content security policy
app.use(
  helmet.contentSecurityPolicy({
    directives: {
      "script-src": ["'self'"],
      // oauth redirects make this a problem I'm not solving right now.
      "form-action": null,
    },
  }),
);

// add protective auth while in development
if (env.DEV_PASSWORD !== "") {
  app.use(
    basicAuth({
      users: { admin: env.DEV_PASSWORD },
      challenge: true, // <--- needed to actually show the login dialog!
      realm: env.PUBLIC_URL,
    }),
  );
}

app.use("/terms", (_req, res) => {
  res.render("terms.nj");
});

// Set up rate limiter: maximum of 100 requests per minute
// and apply rate limiter to all requests
app.use(
  rateLimit({
    windowMs: 1 * 60 * 1000, // 1 minute
    max: 100,
  }),
);

// Add body parsing middleware early
app.use(bodyParser.urlencoded({ extended: true }));
app.use(express.json());

// Add cookie parser before session handling
// app.use(cookieParser(generateSecretKey("http_sessions", 64).toString("base64url")));

// Setup session management
const sessionStore = new RedisStore({
  client: redis,
  prefix: "session:",
});

app.use(
  cookieParser(
    // Set the secret:
    generateSecretKey("http_sessions", 64).toString("base64url"),
  ),
);

app.use(
  session({
    name: "_oss_session",
    cookie: {
      httpOnly: true,
      // This must be 'lax' otherwise the oauth redirect flow does not work:
      sameSite: "lax",
      secure: env.isProduction,
      // Session length is 30 minutes:
      maxAge: 30 * 60 * 1000,
    },
    secret: generateSecretKey("http_sessions", 64).toString("base64url"),
    store: sessionStore,
    // Our session store supports the touch() method:
    resave: true,
   // Only save a session if it has data:
    saveUninitialized: true,
    unset: "destroy",
    // Whether to trust x-forwarded-for
    proxy: env.isProduction,
  }),
);

app.use(flashMiddleware());

// Setup CSRF middleware with bypass for public API
const csrfSecret = generateSecretKey("http_csrf", 64).toString("base64url");
const csrfMiddleware = doubleCsrf({
  getSecret: () => csrfSecret,
  cookieName: "_csrf_token",// The name of the cookie to be used, recommend using Host prefix.
  cookieOptions: {
    sameSite: "strict",
    secure: env.isProduction,
  },
  // size: 64,
  // ignoredMethods: ["GET", "HEAD", "OPTIONS"],
  getTokenFromRequest: (req) => {
    // Bypass CSRF for specific endpoints
    if (req.path === '/admin/api-keys/create-test' || 
      req.path.startsWith('/api/v1/public/') ||
      req.path.match(/^\/api\/v1\/denylist\/admin\/api-keys\/\d+$/) && req.method === 'DELETE') {
    return csrfSecret;
  }

    // FIXME: Remove once all GET mutations are replaced with POSTs
    if (req.method === "GET" && req.query.csrf) {
      return req.query.csrf;
    }

     // Ignore the default ignored requests for CSRF protection:
     if (["GET", "HEAD", "OPTIONS"].includes(req.method)) {
      return;
    }

     // Handle CSRF Token from request body:
     if (req.body && req.body._csrf_token) {
      return req.body._csrf_token;
    }

    if (req.headers["x-csrf-token"]) {
      return req.headers["x-csrf-token"];
    }
    // return req.body?._csrf_token || req.query?.csrf || req.headers["x-csrf-token"];
  },
});

// // ***** PUBLIC ROUTES *****
// // Register this routes BEFORE CSRF and auth middleware
// this allows us to use another auth method for thid api if we want
app.use('/api/v1/denylist', publicApiRouter);


// Add the test API key creation endpoint
app.post('/admin/api-keys/create-test', async (req, res) => {
  try {
    const description = req.body.description || 'Test API Key';
    const apiKey = await createApiKey(description);

    res.json({
      success: true,
      message: 'API key created',
      key: apiKey.key,
      description: apiKey.description
    });
  } catch (error) {
    console.error('Error creating API key:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to create API key'
    });
  }
});

// Apply CSRF protection to all other routes
app.use(csrfMiddleware.doubleCsrfProtection);
app.use((req, res, next) => {
  res.locals.csrf_token = csrfMiddleware.generateToken(req, res);
  next();
});

// Middleware to show flash() messages in the templates:
app.use(flashTemplateMiddleware);

// ***** PUBLIC ROUTES *****
app.use(sessionsRouter);

app.use(authenticationMiddleware);

// ***** AUTHENTICATED ROUTES *****
const authenticatedRouter = express.Router();

app.use(authenticatedRouter);

// Legacy route
app.get("/dashboard", (req, res) => {
  res.redirect("/");
});

const BLOCKLISTS = [
  {
    name: `OSS ${constants.MIN_TIER_PERCENT}%`,
    value: "minimum",
    percentage: constants.MIN_TIER_PERCENT,
  },
  {
    name: `OSS ${constants.MOD_TIER_PERCENT}%`,
    value: "moderate",
    percentage: constants.MOD_TIER_PERCENT,
  },
  {
    name: `OSS ${constants.MAX_TIER_PERCENT}%`,
    value: "maximum",
    percentage: constants.MAX_TIER_PERCENT,
  },
];

const BLOCKLIST_TYPES = BLOCKLISTS.map((list) => list.value);

authenticatedRouter.use(async function (req, res, next) {
  // @ts-ignore
  const serverId = req.session.server_id;

  const server = await queries.getServerConfigById(serverId);
  const accounts = await queries.getAccounts(serverId);
  const push_account = accounts.find(
    (account) => account.account_id === server.api_account_id,
  );

  res.locals.settings = {
    accounts,
    push_account,
    denylists: BLOCKLISTS.map((list) => {
      return { ...list, selected: list.value === server.list_type };
    }),
  };

  res.locals.denylists = BLOCKLISTS;

  next();
});

// home dashboard
authenticatedRouter.get("/", async function (req, res) {
  // @ts-ignore
  const serverId = req.session.server_id;

  // FIXME: Replace with serverId:
  const domain_blocks = await getDomainBlocks(res.locals.server, true);
  const total_blocks = domain_blocks.filter((block) => !block.removed).length;

  const blocks = await Promise.all(
    domain_blocks.map(async (block) => {
      return {
        ...block,
        friendly_severity: block.removed
          ? ""
          : friendlySeverity(block.severity),
        // redundant: await isServerBlockRedundant(serverId, block.domain),
      };
    }),
  );

  // Get recent changes - now returns { changes, pagination }
  const { changes } = await queries.getRecentDomainChanges(1, 5); // page 1, limit 5

  return res.render("dashboard/index.nj", {
    total_blocks,
    unmanaged_blocks: blocks.filter((block) => !block.is_managed),
    managed_blocks: blocks.filter(
      (block) => block.is_managed && !block.removed,
    ),
    recentChanges: changes.map(change => ({
      ...change,
      datetime: change.datetime ? new Date(change.datetime).toLocaleString() : '',
      percentChange: (change.new_percent - change.previous_percent).toFixed(2),
      absoluteChange: Math.abs(change.new_percent - change.previous_percent).toFixed(2)
    }))
  });
});

authenticatedRouter.get("/lists", async (req, res) => {
  res.render("lists/index.nj", {
    lists: BLOCKLISTS,
  });
});

authenticatedRouter.get("/lists/:list", async (req, res) => {
  const listName = req.params.list;

  if (!listName || !BLOCKLIST_TYPES.includes(listName)) {
    res.redirect("back");
    return;
  }

  const list = BLOCKLISTS.find((list) => list.value === listName);
  const blocklist = await getBlocklist(listName);

  const domain_blocks = await getDomainBlocks(res.locals.server, false);
  const blocks = new Map(domain_blocks.map((block) => [block.domain, block]));

  const managedBlocks = domain_blocks
    .filter((block) => block.is_managed)
    .map((block) => block.domain);

  const addedBlocks = blocklist
    .filter((entry) => {
      return !blocks.has(entry.domain);
    })
    .map((block) => block.domain);

  const changedBlocks = blocklist
    .filter((entry) => {
      const block = blocks.get(entry.domain);
      return block?.is_managed && block.severity !== entry.severity;
    })
    .map((block) => block.domain);

  const removedBlocks = domain_blocks.filter((block) => {
    return (
      !blocklist.find((entry) => entry.domain == block.domain) &&
      block.is_managed
    );
  });

  const removedServerBlocks = domain_blocks
    .filter((entry) => {
      const block = blocks.get(entry.domain);
      return !block?.is_managed && block?.removed;
    })
    .map((block) => block.domain);

  const removalIntel = (
    await queries.getDomainsMeta(removedBlocks.map((block) => block.domain_id))
  ).reduce((acc, block) => {
    acc[block.domain_id] = block;
    return acc;
  }, {});

  // Construct and sort an array of domain entries with enhanced details for management status and visibility
  // Domain name are sorted alhabetically
  const entries = [
    ...blocklist.map((entry) => {
      let status = "unmanaged";
      if (addedBlocks.includes(entry.domain)) {
        status = "added";
      } else if (changedBlocks.includes(entry.domain)) {
        status = "changed";
      } else if (managedBlocks.includes(entry.domain)) {
        status = "unchanged";
      } else if (removedServerBlocks.includes(entry.domain)) {
        status = "removed_server";
      }

      return {
        ...entry,
        status,
        friendlySeverity: friendlySeverity(entry.severity),
        source_comment: entry.public_comment,
      };
    }),
    ...removedBlocks.map((entry) => ({
      ...entry,
      status: "removed",
      friendlySeverity: friendlySeverity(entry.severity),
      source_comment:
        removalIntel[entry.domain_id]?.public_comment ?? "Unknown domain",
    })),
  ].sort((a, b) => {
    return a.domain.localeCompare(b.domain);
  });

  res.render("lists/show.nj", {
    list,
    list_total: blocklist.length,
    entries,
    removed_total: removedBlocks.length,
    added_total: addedBlocks.length,
  });
});

authenticatedRouter.post("/lists/:list/subscribe", async (req, res) => {
  // @ts-ignore
  const serverId = req.session.server_id;
  const listType = req.params.list;
  if (!listType || !BLOCKLIST_TYPES.includes(listType)) {
    res.redirect("back");
    return;
  }

  const denylist = BLOCKLISTS.find((list) => list.value === listType);

  if (!denylist) {
    res.redirect("back");
    return;
  }

  try {
    await ServerRepository.updateDenylist(serverId, listType);
    flash(
      "warning",
      `Subscribed to ${denylist?.name}. Please re-enable your push subscription if you are happy with your new choice`,
    );
  } catch (err) {
    res.log.error(err);
    flash("error", "Error subscribing to list");
  }

  res.redirect(`/lists/${denylist.value}`);
});

authenticatedRouter.get("/settings/reset", SettingsController.reset);
authenticatedRouter.post(
  "/settings/push_account",
  SettingsController.updatePushAccount,
);
authenticatedRouter.post("/settings/push", SettingsController.updatePushStatus);
authenticatedRouter.post("/settings/revoke", SettingsController.revoke);
authenticatedRouter.post("/settings/wipe", SettingsController.requestWipe);
authenticatedRouter.post("/admin/clear-error-logs", AdminController.clearErrorLogs);


authenticatedRouter.post("/manage_all_domains", async (req, res) => {
  let safelist = false;
  // these body params are submit button names and don't have a value
  if (
    req.body.manage === undefined &&
    req.body.lock === undefined &&
    req.body.unlock === undefined
  ) {
    req.log.error(
      "Did not specify manage or lock for manage all domains.",
      req.body,
    );
    res.redirect("/");
    return;
  } else if (req.body.lock === "") {
    safelist = true;
  }

  // enroll this block in being managed by us
  await queries.manageAllServerBlocks(res.locals.server_id, safelist);

  res.redirect("/");
  res.end();
});

const BLOCK_ACTIONS = ["manage", "unmanage"];

authenticatedRouter.post("/blocks/:block_ref", async (req, res) => {
  const block_ref = req.params.block_ref;
  const action = req.body.action;

  if (!block_ref || !action || !BLOCK_ACTIONS.includes(action)) {
    flash("error", "Error processing request");
    res.redirect("back");
    return;
  }

  if (action === "manage") {
    await queries.manageServerBlock(res.locals.server_id, block_ref, true);
    flash("success", "Block added for management");
    res.redirect("back");
    return;
  }

  if (action === "unmanage") {
    await queries.manageServerBlock(res.locals.server_id, block_ref, false);
    flash("success", "Block removed for management");
    res.redirect("back");
    return;
  }

  if (action === "remove") {
    const result = await deleteServerBlock(
      res.locals.server_id,
      block_ref,
      true,
    );

    if (result) {
      flash("success", "Removed block successfully");
    } else {
      flash("error", "Failed to remove block");
    }

    res.redirect("back");
    return;
  }
});

const DOMAIN_ACTIONS = ["lock", "unlock"];

authenticatedRouter.post("/domains/:domain_id", async (req, res) => {
  if (
    !req.params.domain_id ||
    !req.body.action ||
    !DOMAIN_ACTIONS.includes(req.body.action)
  ) {
    flash("error", "Error processing request");
    res.redirect("back");
    return;
  }

  // @ts-ignore
  const server_id = req.session.server_id;
  const action = req.body.action;
  const domain = await DomainsRepository.find(req.params.domain_id);

  if (!domain) {
    flash("error", `Unable to find domain: ${req.params.domain_id}`);
    res.redirect("back");
    return;
  }

  if (action === "lock") {
    await queries.updateServerBlockSafelist(server_id, domain.id, true);
    flash("success", `The domain ${domain.domain} has been locked`);
  } else if (action === "unlock") {
    await queries.updateServerBlockSafelist(server_id, domain.id, false);
    flash("success", `The domain ${domain.domain} has been unlocked`);
  }

  res.redirect("back");
});

authenticatedRouter.get("/domain-changes", DomainChangeController.domainChanges);


// ******* ADMIN ROUTES ******** //

authenticatedRouter.use("/admin", adminRouter);

// Error middleware must be the last in the stack, per
// https://expressjs.com/en/guide/error-handling.html
app.use(errorMiddleware);

app.use(requestIp.mw());

export default app;
