// see https://github.com/breejs/bree#cancellation-retries-stalled-jobs-and-graceful-reloading
import pino from "pino";
import { parentPort } from "worker_threads";
import { dbPool } from "./db.js";
import { redis } from "./redis.js";

export class BaseJob {
  lockName = "";
  hasLock = false;
  isRunning = false;

  constructor() {
    this.lockName = `${this.constructor.name}:lock`;
    this.logger = pino({
      name: this.constructor.name,
      redact: {
        paths: ["server.access_token"],
      },
    });

    if (parentPort) {
      parentPort.on("message", async (message) => {
        if (message === "cancel" && this.isRunning) {
          await this.cancel();
        }
      });
    }
  }

  async run() {
    this.logger.info("Starting job");
    this.isRunning = true;

    try {
      await this.handle();
      this.logger.info(`Completed job`);

      if (this.isRunning) {
        this.isRunning = false;

        if (this.hasLock) {
          await redis.del(this.lockName);
        }

        if (parentPort) {
          parentPort.postMessage("done");
        } else {
          process.exit(0);
        }
      }
    } catch (err) {
      if (this.hasLock) {
        await redis.del(this.lockName);
      }

      if (this.isRunning) {
        this.logger.error(
          { error: err.toString(), stack: err.stack },
          `Error handling job`,
        );
        process.exit(1);
      } else {
        process.exit(0);
      }
    }
  }

  async cancel() {
    try {
      this.logger.info("Received cancellation of job");
      this.isRunning = false;

      if (this.hasLock) {
        await redis.del(this.lockName);
      }

      try {
        await redis.disconnect();
      } catch (err) {
        this.logger.error(err);
      }

      try {
        dbPool.end();
      } catch (err) {
        this.logger.error(err);
      }

      if (parentPort) {
        parentPort.postMessage("cancelled");
      } else {
        process.exit(0);
      }
    } catch (err) {
      this.logger.error(
        { error: err.toString(), stack: err.stack },
        "Error cancelling job",
      );
    }
  }

  /**
   * Acquires a lock for the given timeout
   * @param {number} timeout
   */
  async acquireLock(timeout) {
    const value = Date.now() + timeout + 1;
    try {
      const result = await redis.set(this.lockName, value, "PX", timeout, "NX");

      if (result === "OK") {
        this.hasLock = true;
      } else {
        this.hasLock = false;
      }
    } catch (err) {
      this.logger.error(
        { error: err.toString(), stack: err.stack },
        "Error acquiring lock",
      );
      this.hasLock = false;
    }
  }

  async handle() {
    throw new Error("Not Implemented");
  }
}
