import { db } from "./db.js";
import { DuplicateDomainBlockError, api } from "./api.js";
import { blocks, domain_audits, servers } from "./db/schema.js";
import { and, eq, lt, sql, inArray, or } from "drizzle-orm";
import {
  getDomainDigest,
  purgeRedisClientCredentials,
  rotateAccessToken,
} from "./util.js";
import { decryptString } from "./crypter.js";
import * as constants from "./constants.js";
import {
  addEditBlock,
  addUnknownDomain,
  clearMatchedUnknownDomains,
  deauthorizeServer,
  getAccount,
  getBlockByID,
  getBlocksByThreshold,
  getLastAccountTokenVerifyDate,
  getLastServerRefreshDate,
  getServerConfig,
  getServerID,
  getTotalTrustedSources,
  getTrustedFileSources,
  getTrustedSourceBlocks,
  queryDomainBlocks,
  queryServerSafeList,
  removeAllBlocks,
  removeBlock,
} from "./queries.js";
import { USER_AGENT, getActualServerPushLimit } from "./fetcher.js";

import { DomainsRepository } from "./repositories/DomainsRepository.js";
import { ServerRepository } from "./repositories/ServerRepository.js";

import punycode from "punycode/punycode.js";
import {
  parseDomainBlocksCsv,
  writeDomainBlocksCsv,
  writeAuditCsv,
} from "./csv.js";
import { flash } from "@prairielearn/flash";
import { captureException } from '@sentry/node';

const isObfuscated = /(\*)+/;

function isObfuscatedDomain(input) {
  let domain = input;

  if (domain.startsWith("*.")) {
    // remove the leading wildcard
    domain = domain.slice(2);
  }

  return isObfuscated.test(domain);
}

async function deObfuscateDomain(domain, digest) {
  domain = standardizeDomain(domain);
  if (isObfuscated.test(domain)) {
    // check if we have an existing digest for this domain
    if (digest) {
      const domain_by_digest = await DomainsRepository.findByDigest(digest);

      // we found a matching digest
      // de-obfuscate the domain
      if (domain_by_digest && !isObfuscated.test(domain_by_digest.domain)) {
        // console.debug(`Deobfuscated ${domain} by using the digest. Domain is now ${domain_by_digest.domain}`);
        return domain_by_digest.domain;
      } else {
        // console.warn(`Could not find a domain match for digest: ${digest}, obfuscated domain ${domain}`);
        await addUnknownDomain(domain, digest);
        return;
      }
    } else {
      return;
    }
  }

  return domain;
}

function standardizeDomain(domain) {
  if (domain.startsWith("*.")) {
    domain = domain.slice(2);
  }
  if (domain.startsWith(".")) {
    domain = domain.slice(1);
  }
  return domain;
}

function normalizeDomain(domain) {
  return punycode.toASCII(standardizeDomain(domain));
}

function getDomainVariants(domain) {
  const host = new URL("https://" + standardizeDomain(domain)).host;
  const segments = host.split(".");
  return segments.map((_value, index) => segments.slice(index).join("."));
}

export async function findExistingDomainBlock(server_id, domain) {
  const variants = getDomainVariants(domain);
  const existing = await DomainsRepository.findByServerAndVariants(
    server_id,
    variants,
  );

  if (existing?.length > 0) {
    return existing[0].domain;
  }
  return;
}

// note: we don't care if this block already exists, only if it's covered by a higher order block
export async function isServerBlockRedundant(server_id, domain) {
  const existing = await findExistingDomainBlock(server_id, domain);
  return existing && existing !== domain;
}

/**
 * Creates a CSV of the denylist, and return it as a string
 * Used by the "/download/:denylist_type" endpoint (currently disabled)
 * @param {Array} blocks to write to the file
 * @returns {Promise<string>}
 */
export async function writeBlocklistToBuffer(blocks) {
  return writeDomainBlocksCsv(blocks);
}

/**
 * Creates a CSV of the domain audits and returns it as a string
 * @param {Array} domain_audits to write to the file
 * @returns {Promise<string>}
 */
export async function writeDomainAuditsToBuffer(domain_audits) {
  return writeAuditCsv(domain_audits);
}

export async function getBlocklist(blocklist_type) {
  let threshold = constants.MAX_TIER_PERCENT;

  if (blocklist_type === "minimum") {
    threshold = constants.MIN_TIER_PERCENT;
  } else if (blocklist_type === "moderate") {
    threshold = constants.MOD_TIER_PERCENT;
  }

  const blocks = await getBlocksByThreshold(threshold);

  return blocks.filter((block) => {
    // unapproved domains are never included in the blocklist
    if (!block.approved_at) {
      return false;
    }

    return true;
  });
}

/**
 * Checks if a domain's blocklist needs to be updated based on the last refresh date.
 *
 * @async
 * @param {string} domain - The domain to check for updates.
 * @returns {Promise<boolean>} Returns true if the domain needs an update, false otherwise.
 *
 * @description
 * This function checks the last refresh date for a given domain and compares it
 * with the current time. If the time difference is less than the defined LATENCY
 * constant, it returns false, indicating no update is needed. Otherwise, it
 * returns true, suggesting an update should be performed.
 *
 * The LATENCY constant is used to prevent too frequent updates, optimizing
 * performance and reducing unnecessary API calls.
 */
export async function checkNeedsUpdate(domain) {
  // check latency on the last refresh date
  // if it was updated recently, no sense doing too much
  // work.
  const now = new Date();
  const last_update = await getLastServerRefreshDate(domain);

  if (last_update.length > 0) {
    const last_refresh = last_update[0].last_server_refresh_date;

    if (last_refresh) {
      const checkDate = new Date(last_refresh);

      // To calculate the time difference of two dates
      const Difference_In_Time = now.getTime() - checkDate.getTime();

      // To calculate the no. of minutes between two dates
      const Difference_In_Minutes = Difference_In_Time / (1000 * 60);

      if (Difference_In_Minutes < constants.LATENCY) {
        // console.debug(`Last update for ${domain} was within ${constants.LATENCY} minutes. Skipped.`);
        return false;
      }
    }
  }

  return true;
}

export async function checkNeedsTokenVerifyUpdate(server_id, account_id) {
  // check latency on the last refresh date
  // if it was updated recently, no sense doing too much
  // work.
  const now = new Date();
  const last_update = await getLastAccountTokenVerifyDate(
    server_id,
    account_id,
  );

  if (last_update.length > 0) {
    const last_refresh = last_update[0].last_token_verify_date;

    if (last_refresh) {
      const checkDate = new Date(last_refresh);

      // To calculate the time difference of two dates
      const Difference_In_Time = now.getTime() - checkDate.getTime();

      // To calculate the no. of minutes between two dates
      const Difference_In_Minutes = Difference_In_Time / (1000 * 60);

      if (Difference_In_Minutes < constants.TOKEN_VERIFY_LATENCY) {
        return false;
      }
    }
  }

  return true;
}

async function handleDomainMapRow(blocks, domainMap, domains_list) {
  const serverDomainIndex = new Map();

  for (let i = 0; i < blocks.length; i++) {
    const block = blocks[i];

    // skip domains that are obfuscated
    const deobfuscatedDomain = await deObfuscateDomain(
      block.domains.domain,
      block.domains.digest,
    );
    if (deobfuscatedDomain) {
      block.domains.domain = deobfuscatedDomain;
    } else {
      continue;
    }

    // normalize the domain for save
    block.domains.domain = normalizeDomain(block.domains.domain);

    // skip any domain on the hardcoded safelist
    if (constants.SAFE_DOMAINS.includes(block.domains.domain)) {
      continue;
    }

    // skip garbage domains
    if (constants.IGNORED_DOMAINS.includes(block.domains.domain)) {
      continue;
    }

    // due to de-obfuscation and normalization of domains, we might end up with
    // multiple domains from the same server, like a server stores both
    // *.activitypub-troll.cf and activitypub-troll.cf
    // use this to ensure that the same server and domain combo is only counted once
    if (
      !serverDomainIndex.has(`${block.servers.server}:${block.domains.domain}`)
    ) {
      serverDomainIndex.set(
        `${block.servers.server}:${block.domains.domain}`,
        true,
      );
    } else {
      continue;
    }

    if (!domainMap[block.domains.domain]) {
      const domain = await DomainsRepository.findByDomain(
        block.domains.domain,
        {
          columns: { id: true },
        },
      );

      // We can only calculate audits for domains we already know about:
      if (!domain) {
        continue;
      }

      domains_list.push(block.domains.domain);
      domainMap[block.domains.domain] = {
        domain_id: domain.id,
        count: 0,
        percent: 0,
        public_comment: "",
        severity: block.blocks.severity,
        count_suspend: 0,
        count_silence: 0,
        count_noop: 0,
        authoritative: false,
      };
    }

    // concatenate public comments for this block
    if (block.blocks.public_comment && block.blocks.public_comment.length) {
      if (domainMap[block.domains.domain].public_comment.length) {
        domainMap[block.domains.domain].public_comment =
          domainMap[block.domains.domain].public_comment + "; ";
      }

      domainMap[block.domains.domain].public_comment =
        domainMap[block.domains.domain].public_comment +
        block.blocks.public_comment;
    }

    // console.log('block.servers', block.servers);

    // if this domain is authoritative, add it and continue
    // do not count this authoritative source in the counts or percentages
    if (block.servers.authoritative) {
      domainMap[block.domains.domain].severity = block.blocks.severity;
      domainMap[block.domains.domain].authoritative = true;
      continue;
    }

    domainMap[block.domains.domain].count++;

    // check severity, use the majority vote, tiebreaker is the least severe
    switch (block.blocks.severity) {
      case "suspend":
        domainMap[block.domains.domain].count_suspend++;
        break;
      case "silence":
        domainMap[block.domains.domain].count_silence++;
        break;
      case "noop":
        domainMap[block.domains.domain].count_noop++;
        break;
    }

    // if this entry has previously been designated as authoritative,
    // skip further modifications of the assigned severity and
    // treat the authoritative block recommendation as an override to all other changes
    if (domainMap[block.domains.domain].authoritative) {
      continue;
    }

    const current_block = domainMap[block.domains.domain];

    const current_severity_recommendation =
      current_block.count_suspend > current_block.count_silence &&
      current_block.count_suspend > current_block.count_noop
        ? "suspend"
        : // equal number of suspends and silences means a silence recommendation
          current_block.count_suspend > 0 &&
            current_block.count_suspend > current_block.count_noop &&
            current_block.count_suspend === current_block.count_silence
          ? "silence"
          : current_block.count_silence > current_block.count_noop
            ? "silence"
            : "noop";

    current_block.severity = current_severity_recommendation;
  }
}

// create high, medium, and low tier blocklists,
// stored in the blocklist table
export async function compileBlocklists() {
  // get all the blocks from all trusted sources
  // also includes authoritative blocks
  const domain_blocks = await getTrustedSourceBlocks();

  /**
   * @type {Record<string, DomainAudit>}
   */
  const domainMap = {};
  const domains_list = [];

  await handleDomainMapRow(domain_blocks, domainMap, domains_list);

  const total_sources = await getTotalTrustedSources();

  for (let i = 0; i < domains_list.length; i++) {
    const domain = domains_list[i];
    domainMap[domain]["percent"] =
      (domainMap[domain].count / total_sources) * 100;
  }

  // save the updated domain map to the db and disk
  await saveDomainMap(domainMap);
}

async function retrieveBlocklistFromAPI(server, server_config, account) {
  let domain_blocks = await api.get_domain_blocks(
    server,
    decryptString(server_config.access_token),
    true,
    undefined,
    undefined,
    account,
  );

  if (domain_blocks?.error) {
    if (domain_blocks.error === "The access token was revoked") {
      // rotate the access token
      const rotated = await rotateAccessToken(server);
      if (rotated) {
        const new_config = await getServerConfig(server);
        domain_blocks = await api.get_domain_blocks(
          server,
          decryptString(new_config.access_token),
          true,
          undefined,
          undefined,
          account,
        );
      } else {
        console.error(`Could not rotate access token for ${server}.`);
        deauthorizeServer(server);
        purgeRedisClientCredentials(server);
      }
    }
  }
  // FIXME @thisismissem pass correct logger
  await saveServerBlocklist(
    server_config.server_id,
    domain_blocks,
    true,
    console,
  );
}
export async function getDomainBlocks(server, noCache) {
  const server_config = await getServerConfig(server);
  const account = await getAccount(
    server_config.server_id,
    server_config.api_account_id,
  );

  
  if (noCache || (await checkNeedsUpdate(server))) {
    await retrieveBlocklistFromAPI(server, server_config, account);
  }

  return queryDomainBlocks(server);
}

export async function getDomainBlocksAsDict(server, blocks) {
  const domain_blocks = blocks ? blocks : await getDomainBlocks(server);
  return domain_blocks.length > 0
    ? Object.fromEntries(domain_blocks.map((block) => [block.domain, block]))
    : {};
}

export async function getServerSafelistAsDict(server, safelist) {
  const server_safelist = safelist
    ? safelist
    : await queryServerSafeList(server);
  return server_safelist.length > 0
    ? Object.fromEntries(server_safelist.map((block) => [block.domain, true]))
    : {};
}

// complicated logic: when editing, we don't care about a change in comments (ignore_comments === true)
// unless the user has no comments at all, in which case we'll push an update adding comments *once*
export function blockWillChange(
  origin_block,
  blocklist_block,
  ignore_comments,
) {
  return (
    blocklist_block.severity !== origin_block.severity ||
    ((!ignore_comments || (origin_block.public_comment?.length ?? 0 === 0)) &&
      constants.ENABLE_PUBLIC_COMMENTS &&
      blocklist_block.public_comment !== origin_block.public_comment)
  );
}

/*
    Returns a list of adds, edits, and deletes to be applied to a given server, up to the limit

    Allows us to handle updates in batches without smashing the server's rate limit.
*/
export async function getManagedBlocks(server, serverConfig, limit, logger) {
  // we're not spending energy getting instance follows
  // for servers that aren't receiving blocks
  if (!serverConfig.enabled) {
    return [];
  }

  const managed_blocks = [];
  let not_managed_total = 0;
  // This is the source list that will be pushed
  const blocklist = await getBlocklist(serverConfig.list_type);

  // the server's existing blocks
  const server_blocks = await getDomainBlocks(server.server);
  const server_blocks_dictionary = await getDomainBlocksAsDict(
    server.server,
    server_blocks,
  );

  // the server's safe list, blocks that will be ignored in future blocklist updates
  const safelist_dictionary = await getServerSafelistAsDict(server.server);

  // reverse dictionary: retractions
  // when we have blocks on our server that aren't in the blocklist from which we're pulling
  const blocklist_dictionary =
    blocklist.length > 0
      ? Object.fromEntries(blocklist.map((block) => [block.domain, true]))
      : [];

  for (let i = 0; i < blocklist.length; i++) {
    // the new block
    const block = blocklist[i];

    // skip any domain on the hardcoded safelist
    if (constants.SAFE_DOMAINS.includes(block.domain)) {
      continue;
    }

    // the origin server block
    const origin_block = server_blocks_dictionary[block.domain];

    // make sure this domain isn't on the safe list for this server
    const is_safe = safelist_dictionary[block.domain];
    if (is_safe) {
      // console.debug(`${block.domain} is on the safelist for ${server}. Skipping update.`);
      continue;
    }

    // do not EVER let the server defederate themselves by appearing on someone else's blocklist
    if (server.server === block.domain) {
      continue;
    }

    // break after we have hit our managed block push limit per batch for this server
    if (limit && managed_blocks.length >= limit) {
      break;
    }

    // nonexistent entry on our server, this is a new entry
    if (!origin_block) {
      // make sure this block is not already covered by a more generic block
      if (await isServerBlockRedundant(server.id, block.domain)) {
        logger.info(
          `skipping ${block.domain} for ${server.server} as a redundant block was found`,
        );
        continue;
      }

      // add to our managed blocks
      managed_blocks.push({
        ...block,
        option: "add",
      });
      continue;
    }

    // Don't update any blocks not managed by this server
    if (!origin_block.is_managed) {
      not_managed_total++;
      // console.debug(`${block.domain} is not managed by this server. Skipping update.`);
      continue;
    }

    // edit
    // we don't send changes for edits, simply because the user may want to hand-edit the comment
    // after it's added. Also because comments might be compiled in a different order and trigger
    // "fake changes"
    // because we're possibly/augmenting the block severity, it's possible we're not changing anything now
    // so check again
    if (blockWillChange(origin_block, block, true)) {
      // we need an existing ID reference. Luckily, we should have one from the origin
      managed_blocks.push({
        ...block,
        id: origin_block.block_ref,
        option: "edit",
      });
      continue;
    }
  }

  // removals are the reverse: items in the server_blocks which are managed by us
  // but are no longer in our target blocklist.
  const removals = [];
  for (let j = 0; j < server_blocks.length; j++) {
    // break after we have hit our managed block push limit per batch for this server
    if (limit && managed_blocks.length + removals.length >= limit) {
      break;
    }

    const server_block = server_blocks[j];

    // if the block still exists in our subscribed blocklist, this is not a removal
    if (blocklist_dictionary[server_block.domain]) {
      continue;
    }

    // Don't update any blocks not managed by this server
    if (!server_block.is_managed) {
      not_managed_total++;
      continue;
    }

    // make sure this domain isn't on the safe list for this server
    if (safelist_dictionary[server_block.domain]) {
      continue;
    }

    // a little decoration
    removals.push({
      ...server_block,
      option: "delete",
    });
  }

  managed_blocks.push(...removals);

  return managed_blocks;
}

// hit the public api for all source servers
// unless we have configuration for the sources
// in which case we'll favor the admin api
// returns true if fresh data was pulled
export async function downloadSourceBlocks(refresh, logger) {
  const source_domains = constants.SOURCE_DOMAINS.sort();
  let updates = false;

  for (let i = 0; i < source_domains.length; i++) {
    const domain = source_domains[i];

    // check latency on the last refresh date
    // if it was updated recently, no sense doing too much
    // work.
    if (!refresh && !(await checkNeedsUpdate(domain))) {
      continue;
    }

    logger.info(`Fetching instance blocklist from ${domain}...`);

    try {
      const config = await getServerConfig(domain);
      const isAdmin = config?.access_token ? true : false;
      const access_token = config?.access_token
        ? decryptString(config.access_token)
        : "";
      const domain_blocks = await api.get_domain_blocks(
        domain,
        access_token,
        isAdmin,
      );
      const server_id = await getServerID(domain);

      if (!domain_blocks?.error && domain_blocks.length > 0) {
        updates = true;
        await saveServerBlocklist(server_id, domain_blocks, false, logger);
        logger.info(`Saved blocklist for source server ${domain}.`);
      }
    } catch (err) {
      logger.error(
        { error: err.toString(), stack: err.stack, domain },
        `Error pulling down source for ${domain}. Skipping...`,
      );
    }
  }

  return updates;
}

export async function importFileSources(refresh, logger) {
  const file_sources = await getTrustedFileSources();
  const downloads = file_sources.map(async (source) => {
    // check latency on the last refresh date
    // if it was updated recently, no sense doing too much
    // work.
    if (!refresh && !(await checkNeedsUpdate(source.server))) {
      return;
    }

    const file = constants.SOURCE_FILES[source.server];
    if (!file) {
      logger.info(`Missing file blocklist for ${source.server}`);
      return;
    }

    try {
      await importSourceFile(source, file, logger);
    } catch (err) {
      logger.error(
        { error: err.toString(), stack: err.stack, file },
        `Error pulling down file for ${file.name}. Skipping...`,
      );
    }
  });

  await Promise.all(downloads);
}

export async function importSourceFile(source, file, logger) {
  if (!source.id) {
    return false;
  }

  logger.info(`Fetching file blocklist from ${file.name}...`);

  const contents = await fetch(file.url, {
    headers: {
      Accept: "text/csv,text/plain",
      "User-Agent": USER_AGENT,
    },
  }).then((res) => {
    if (!res.ok) {
      throw new Error(`Failed to fetch file ${URL}`);
    }

    return res.text();
  });

  if (!contents) {
    return false;
  }

  const domain_blocks = await parseDomainBlocksCsv(contents, file.format);

  if (domain_blocks.length === 0) {
    logger.info(
      { file },
      `Unable to parse domain blocks from ${source.server} file`,
    );
    return false;
  }

  const domains = domain_blocks.map((block) => block.domain);
  const domainMap = new Map();

  await Promise.all(
    domains.map(async (domain) => {
      const result = await DomainsRepository.upsert({ domain });
      domainMap.set(domain, result.id);
    }),
  );

  await db.transaction(async (tx) => {
    await tx
      .update(blocks)
      .set({ updated: false })
      .where(eq(blocks.server_id, source.id));

    await Promise.all(
      domain_blocks.map((domain_block) => {
        const domain_id = domainMap.get(domain_block.domain);

        // This will never be the case, but keeps types happy:
        if (!domain_id) {
          return;
        }

        return tx
          .insert(blocks)
          .values({
            server_id: source.id,
            domain_id: domain_id,
            severity: domain_block.severity,
            public_comment: domain_block.public_comment,
            reject_media: domain_block.reject_media,
            last_updated_date: sql`NOW()`,
            updated: false,
          })
          .onConflictDoUpdate({
            target: [blocks.server_id, blocks.domain_id],
            set: {
              severity: domain_block.severity,
              public_comment: domain_block.public_comment,
              reject_media: domain_block.reject_media,
              last_updated_date: sql`NOW()`,
              updated: true,
              removed: false,
            },
          })
          .returning({
            domain_id: blocks.domain_id,
            severity: blocks.severity,
            public_comment: blocks.public_comment,
            reject_media: blocks.reject_media,
          });
      }),
    );

    await tx
      .update(blocks)
      .set({ removed: true, removed_date: sql`NOW()` })
      .where(
        and(
          eq(blocks.server_id, source.id),
          eq(blocks.updated, false),
          eq(blocks.removed, false),
        ),
      );

    // update server last_server_refresh_date
    await tx
      .update(servers)
      .set({ last_server_refresh_date: sql`NOW()` })
      .where(eq(servers.id, source.id));
  });
}

export async function pushServerBlock(server_id, serverConfig, block, logger) {
  const account = await getAccount(server_id, serverConfig.api_account_id);
  const domain = await DomainsRepository.findByDomain(block.domain, {
    columns: { id: true },
  });
  const accessToken = decryptString(serverConfig.access_token);

  if (!domain) {
    flash("error", `Cannot add block ${block.domain} for server ${server_id} because block domain wasn't found`);
    logger.error(
      { block },
      `Cannot add block ${block.domain} for server ${server_id} because block domain wasn't found`,
    );
    return;
  }

  // console.debug(`Pushing ${block.option === 'edit' ? 'an update to a ' :''}block for ${block.domain} on behalf of ${config.server}...`);
  if (block.option === "add") {
    try {
      const add_block = await api.add_domain_block(
        serverConfig.server,
        accessToken,
        block.domain,
        block.severity,
        block.public_comment,
        account,
      );

      if (add_block.error) {
        logger.error(
          { error: add_block.error },
          `Error adding block for ${block.domain} to ${serverConfig.server}`,
        );
        next(add_block.error);
        return;
      }

      // add to the db as well
      await addEditBlock({
        ...add_block,
        server_id: server_id,
        domain_id: domain.id,
        is_admin: true,
        is_managed: true,
        block_ref: add_block.id,
        private_comment: constants.CLIENT_NAME,
      });
    } catch (err) {
      if (err instanceof DuplicateDomainBlockError) {
        // We upsert here because the DuplicateDomainBlock may be for a domain
        // we don't yet know about:
        const existingDomain = await DomainsRepository.upsert({
          domain: err.existingBlock.domain,
          // The admin API only starts returning the digest as of 4.3:
          digest: err.existingBlock.digest,
        });

        await addEditBlock({
          server_id: account.server_id,
          domain_id: existingDomain.id,
          is_admin: true,
          // if this is a surprise to us, it's almost certainly not managed by us
          is_managed: false,
          // The domain block properties, we take these from the API error
          // instead of the `add_block`, since the existing block may be for a
          // domain variant
          block_ref: err.existingBlock.id,
          severity: err.existingBlock.severity,
          reject_media: err.existingBlock.reject_media,
          reject_reports: err.existingBlock.reject_reports,
          public_comment: err.existingBlock.public_comment ?? "",
          private_comment: err.existingBlock.private_comment ?? "",
          obfuscate: err.existingBlock.obfuscate,
        });
      }
    }
  } else if (block.option === "edit") {
    if (!block.id) {
      logger.error(
        `Error editing block for ${block.domain} to ${serverConfig.server}. Missing block.id`,
      );
      return;
    }
    const update = await api.update_domain_block(
      serverConfig.server,
      accessToken,
      block.id,
      block.severity,
      block.public_comment,
      account,
    );

    if (update.error) {
      logger.error(
        { error: update.error },
        `Error editing block for ${block.domain} to ${serverConfig.server}`,
      );
      return;
    }

    // add to the db as well
    await addEditBlock({
      ...update,
      server_id: server_id,
      domain_id: domain.id,
      is_admin: true,
      is_managed: true,
      block_ref: block.id,
    });
  } else {
    logger.error(`unknown block.option '${block.option}'`);
  }
}

export async function deleteServerBlock(
  server_id,
  serverConfig,
  block_id,
  safelist,
  logger,
) {
  const block = await getBlockByID(server_id, block_id);

  let result = true;
  if (!block) {
    logger.error(
      { block },
      `Invalid internal block ID ${block_id} for server_id ${server_id}.`,
    );
    return false;
  }

  const domain = await DomainsRepository.find(block.domain_id);
  const accessToken = decryptString(serverConfig.access_token);
  const account = await getAccount(server_id, serverConfig.api_account_id);

  // DomainsRepository.find can return undefined:
  if (!domain) {
    const error = new Error(`Cannot remove block ${block_id} for server ${server_id} because block domain wasn't found`);
    error.status = 404;
    
    // Keep existing logging
    logger.error(
      { block },
      `Cannot remove block ${block_id} for server ${server_id} because block domain wasn't found`,
    );

    // Log to database and Sentry
    logErrorToDatabase(error, {
      userId: account?.id,
      serverId: server_id,
      path: 'deleteServerBlock',
      context: {
        blockId: block_id,
        type: 'block_removal_error'
      }
    });

    return false;
  }

  const response = await api.remove_domain_block(
    block.block_ref,
    serverConfig.server,
    accessToken,
    account,
  );

  if (response.error) {
    result = false;
    const error = new Error(`Error deleting block for domain ${domain.domain} for ${serverConfig.server}`);
    error.status = 400;
    
    // Keep existing logging
    logger.error(
      { error: response.error },
      `Error deleting block for domain ${domain.domain} for ${serverConfig.server}`,
    );

    // Log to database and Sentry
    logErrorToDatabase(error, {
      userId: account?.id,
      serverId: server_id,
      path: 'deleteServerBlock',
      context: {
        blockId: block_id,
        domain: domain.domain,
        responseError: response.error,
        type: 'block_deletion_error'
      }
    });
    
  } else {
    // update 'remove' flag in db, so long as we actually removed the block and didn't soft remove
    // with a noop entry, otherwise update severity
    await removeBlock(server_id, block_id, safelist);
  }

  return result;
}

export async function deleteServerBlocklist(server_id, hard_wipe) {
  return await removeAllBlocks(server_id, hard_wipe);
}

// runs in stages, returns false when there are no more blocks available
export async function wipeAllBlockData(server, limit) {
  // FIXME: Handle missing server config:
  const server_config = await getServerConfig(server);
  const access_token = decryptString(server_config.access_token);
  const account = await getAccount(
    server_config.server_id,
    server_config.api_account_id,
  );

  // do not rely on the db cache for this operation
  const domain_blocks = await getDomainBlocks(server, true);

  let x = 0;
  let removals = 0;
  const block_removals = domain_blocks.map(async (block) => {
    if (x > limit) {
      return;
    }

    // check the push limit and abort if necessary
    const push_limit = await getActualServerPushLimit(server);
    if (push_limit === 0) {
      return;
    }

    // skip unmanaged blocks or blocks on the locked list
    if (!block.is_managed || block.safelist) {
      return;
    }

    const response = await api.remove_domain_block(
      block.block_ref,
      server,
      access_token,
      account,
    );
    if (!response.error && response.error !== "Record not found") {
      removals++;
    }
    // remove the entry from the db either way
    await removeBlock(server_config.id, block.block_ref);
    x++;
  });

  await Promise.all(block_removals);

  // if there are *no* more domain blocks, set the server wipe flag OFF
  if (removals === 0) {
    const server_id = await getServerID(server);
    await ServerRepository.finishWipe(server_id);
    return true;
  }

  return false;
}

// saves updated blocks to the db
// domain_blocks returned from api.get_domain_blocks
export async function saveServerBlocklist(
  server_id,
  domain_blocks,
  isAdmin,
  logger,
) {
  if (domain_blocks?.error) {
    const error = new Error(`Error pulling blocks, aborting save. Error: ${domain_blocks.error}`);
    error.status = 500;
    
    // Keep existing logging
    logger.error(
      { error: domain_blocks?.error },
      "error pulling blocks, aborting save",
    );

    // Capture with Sentry
    captureException(error, {
      level: 'error',
      tags: {
        type: 'block_pull_error',
        server_id: server_id
      },
      extra: {
        domain_blocks_error: domain_blocks.error
      }
    });
    
    return;
  }

  const domains = await Promise.all(
    domain_blocks.map(async (block) => {
      let domain = block.domain;
      let digest = block.digest;

      if (digest && isObfuscatedDomain(domain)) {
        const deobfuscatedDomain = await deObfuscateDomain(domain, digest);
        if (deobfuscatedDomain) {
          domain = deobfuscatedDomain;
          digest = getDomainDigest(deobfuscatedDomain);
        } else {
          // If we can't deobfuscate the domain, we skip it as we only store
          // deobfuscated domains:
          return;
        }
      } else {
        digest = digest ?? getDomainDigest(domain);
      }

      const result = await DomainsRepository.upsert({
        domain,
        digest,
      });

      return { originalDomain: block.domain, ...result };
    }),
  );

  const domainMap = new Map();
  domains.forEach((domain) => {
    if (!domain) {
      return;
    }
    domainMap.set(domain.originalDomain, domain.id);
  });

  const rows = domain_blocks.reduce((rows, block) => {
    const domain_id = domainMap.get(block.domain);

    if (domain_id) {
      rows.push({
        server_id,
        domain_id,
        block_ref: isAdmin ? block.id.toString() : undefined,
        severity: block.severity,
        is_admin: isAdmin,
        private_comment: isAdmin ? block.private_comment : "",
        public_comment: isAdmin ? block.public_comment : block.comment,
        reject_media: isAdmin ? block.reject_media : undefined,
        created_at: isAdmin ? new Date(block.created_at) : new Date(),
        updated: true,
        removed: false,
      });
    }

    return rows;
  }, []);

  await db.transaction(async (tx) => {
    // Locks the rows in `blocks` where server_id matches, ensuring no one else
    // writes to them, transactions will be queued:
    // See more: https://www.cockroachlabs.com/blog/select-for-update/
    //
    // select from "blocks" where "blocks"."server_id" = $1 for update
    await tx
      .select({})
      .from(blocks)
      .where(eq(blocks.server_id, server_id))
      .for("update");

    // set all existing blocks for this server to updated = 0
    await tx
      .update(blocks)
      .set({ updated: false })
      .where(eq(blocks.server_id, server_id));

    await Promise.all(
      rows.map((row) => {
        return tx
          .insert(blocks)
          .values(row)
          .onConflictDoUpdate({
            target: [blocks.server_id, blocks.domain_id],
            set: {
              block_ref: row.block_ref,
              severity: row.severity,
              public_comment: row.public_comment,
              private_comment: row.private_comment,
              reject_media: row.reject_media,
              last_updated_date: sql`NOW()`,
              updated: true,
              removed: false,
              removed_date: null,
            },
          })
          .returning({
            block_ref: blocks.block_ref,
            domain_id: blocks.domain_id,
            severity: blocks.severity,
            public_comment: blocks.public_comment,
            reject_media: blocks.reject_media,
          });
      }),
    );

    // we now know the retractions from our own list: everything with updated=0 that never got pushed
    await tx
      .update(blocks)
      .set({ is_managed: false, removed: true, removed_date: sql`NOW()` })
      .where(
        and(
          eq(blocks.server_id, server_id),
          eq(blocks.updated, false),
          eq(blocks.removed, false),
        ),
      );

    // update server last_server_refresh_date
    await tx
      .update(servers)
      .set({ last_server_refresh_date: sql`NOW()` })
      .where(eq(servers.id, server_id));
  });
}

export async function clearDomainAudits() {
  await db.delete(domain_audits);
}

export async function expireDomainAudits() {
  // delete anything that wasn't updated in the last day
  const now = new Date();
  const cutoff = new Date(now.setDate(now.getDate() - 1));
  await db.delete(domain_audits).where(lt(domain_audits.updated_at, cutoff));
}

/**
 * @typedef DomainAudit
 * @type {object}
 * @property {string} domain
 * @property {number} domain_id
 * @property {number} count
 * @property {number} percent
 * @property {string} public_comment
 * @property {string} severity
 * @property {number} count_suspend
 * @property {number} count_silence
 * @property {number} count_noop
 * @property {boolean} authoritative
 */

/**
 * Saves or updates domain audit information in bulk.
 *
 * @async
 * @function saveDomainMap
 * @param {Record<string, DomainAudit>} domainMap - An object where keys are domain strings and values are DomainAudit objects.
 * @returns {Promise<void>} - A promise that resolves when all updates are complete.
 * @throws {Error} Will throw an error if the database transaction fails.
 *
 * @description
 * This function performs a database transaction to insert or update domain audit information.
 * If a record with the same domain_id already exists, it will be updated instead of inserted.
 * The function uses a single transaction for all operations to ensure data consistency.
 */
export async function saveDomainMap(domainMap) {
  await db.transaction(async (tx) => {
    for (const audit of Object.values(domainMap)) {
      await tx
        .insert(domain_audits)
        .values({
          domain_id: audit.domain_id,
          count: audit.count,
          percent: audit.percent,
          public_comment: audit.public_comment,
          severity: audit.severity,
          count_suspend: audit.count_suspend,
          count_silence: audit.count_silence,
          count_noop: audit.count_noop,
          authoritative: audit.authoritative,
        })
        .onConflictDoUpdate({
          target: domain_audits.domain_id,
          set: {
            count: audit.count,
            percent: audit.percent,
            public_comment: audit.public_comment,
            severity: audit.severity,
            count_suspend: audit.count_suspend,
            count_silence: audit.count_silence,
            count_noop: audit.count_noop,
            authoritative: audit.authoritative,
            updated_at: sql`NOW()`,
          },
        });
    }
  });
}

// function to check if trusted sources are stale
export async function areTrustedSourcesStale() {
  try {
    // Get all trusted sources
    const file_sources = await getTrustedFileSources();
    const source_domains = constants.SOURCE_DOMAINS;
    const all_sources = [...file_sources, ...source_domains.map(domain => ({server: domain}))];
    
    if (all_sources.length === 0) {
      const noSourcesError = new Error('No trusted sources configured');
      noSourcesError.status = 400;
      captureException(noSourcesError, {
        level: 'warning',
        tags: { type: 'no_trusted_sources' }
      });
      flash('warning', 'No trusted sources are configured');
      return true;
    }

    // Check if any source is older than 24 hours
    const now = new Date();
    const cutoff = new Date(new Date().setHours(new Date().getHours() - 24));

    const stale_sources = await db
      .select({ 
        server: servers.server,
        last_refresh: servers.last_server_refresh_date 
      })
      .from(servers)
      .where(
        and(
          inArray(servers.server, all_sources.map(s => s.server)),
          or(
            eq(servers.last_server_refresh_date, null),
            lt(servers.last_server_refresh_date, cutoff)
          )
        )
      );

    if (stale_sources.length > 0) {
      const staleServers = stale_sources.map(s => s.server).join(', ');
      const staleError = new Error(`Trusted sources not updated in 24 hours: ${staleServers}`);
      staleError.status = 400;
      
      captureException(staleError, {
        level: 'warning',
        tags: {
          type: 'stale_trusted_sources',
          affected_servers: staleServers
        }
      });

      flash('warning', `The following trusted sources need updating: ${staleServers}`);
    }

    return stale_sources.length > 0;

  } catch (error) {
    const contextError = new Error(`Failed to check trusted sources status: ${error.message}`);
    contextError.originalError = error;
    contextError.status = 500;

    captureException(contextError, {
      level: 'error',
      tags: { type: 'trusted_sources_check_failure' }
    });

    flash('error', 'Unable to verify trusted sources status. Our team has been notified.');
    throw contextError; // Will be caught by errorMiddleware
  }
}

// initialize by downloading blocks for all source domains
// then compile the domain audit table which will generate our
// blocklist files and set thresholds for each domain
export async function updateSourceBlocklists(refresh, logger) {
  // Check if sources are stale when refresh is false
  // if (!refresh) {
  //   const isStale = await areTrustedSourcesStale();
  //   if (!isStale) {
  //     logger.info("Trusted sources are up to date (updated within last 24 hours)");
  //     return false;
  //   }
  // }

  
  const updates = await downloadSourceBlocks(refresh, logger);
  await importFileSources(refresh, logger);
  await compileBlocklists();
  await clearMatchedUnknownDomains();
  return updates;
}

// write block to user server as well as the db
export async function migrateBlockToServer(
  server,
  serverConfig,
  block,
  logger,
) {
  if (block.option === "add" || block.option === "edit") {
    return await pushServerBlock(server.id, serverConfig, block, logger);
  } else if (block.option === "delete") {
    return await deleteServerBlock(
      server.id,
      serverConfig,
      block.block_ref,
      undefined,
      logger,
    );
  }

  logger.error(`Unknown block option '${block.option}`);
}

export async function pushBlocklist(serverName, limit, logger) {
  // verify this server has push enabled
  const server = await ServerRepository.findByName(serverName);
  const serverConfig = await getServerConfig(serverName);

  if (!serverConfig || !serverConfig.enabled) {
    logger.info(`Push isn't enabled for ${server.server}`);
    return;
  }

  // get the list of changes we manage to be added, edited, or deleted
  const blocklist = await getManagedBlocks(server, serverConfig, limit, logger);

  if (blocklist.length === 0) {
    // Whilst we didn't need to push, we did calculate that, so we can update
    // the last pushed at timestamp. This affected servers for whom there were
    // no changes never getting a last pushed at timestamp.
    await ServerRepository.updateLastPushedAt(server.id);

    logger.info(`Nothing to do for ${server.server}`);
  }

  if (blocklist.length > 0) {
    logger.info(`Beginning block updates for ${server.server}...`);
  }

  const block_threads = blocklist.map(async (block) => {
    // no more pushing when we're out of remaining requests for this server
    const push_limit = await getActualServerPushLimit(server.server);
    if (push_limit === 0) {
      return;
    }
    await migrateBlockToServer(server, serverConfig, block, logger);
  });

  await Promise.all(block_threads);

  if (blocklist.length > 0) {
    logger.info(`Block updates complete this batch for ${server.server}.`);

    await ServerRepository.updateLastPushedAt(server.id);
  }
}

export async function verifyCredentials(
  server,
  access_token,
  noRotate,
  logger,
) {
  try {
    const verified = await api.verify_app_credentials(
      server,
      access_token,
      null, // no account
      logger,
    );

    if (!verified) {
      if (!noRotate) {
        // if apptoken is invalid, rotate it, if possible
        const rotated = await rotateAccessToken(server);

        // we cannot rotate to a working access_token so it's time to abandon ship
        if (!rotated) {
          return false;
        }
      } else {
        return false;
      }
    }

    return true;
  } catch (err) {
    logger.error(
      { error: err.toString(), stack: err.stack },
      `Error verifying credentials for push`,
    );

    // Capture with Sentry
    captureException(err, {
      level: 'error',
      tags: {
        type: 'credential_verification_error',
        server: server
      },
      extra: {
        error_message: err.toString(),
        stack: err.stack
      }
    });

    return false;
  }
}

