import convict from "convict";
import toml from "@iarna/toml";

import { fileURLToPath } from "url";
import path from "path";

const __filename = fileURLToPath(import.meta.url);
const __dirname = path.dirname(__filename);

// TOML config schema for this app
const config = convict({
  app: {
    admin_accounts: {
      doc: "An array of superadmin accounts that will have access to the admin panel",
      format: Array,
      default: [],
    },
  },
  sources: {
    source_domains: {
      doc: "Trusted sources for this application",
      format: Array,
      default: [],
    },
    safe_domains: {
      doc: "Server-level safe domains",
      format: Array,
      default: [],
    },
    authoritative_domains: {
      doc: "Server-level authoritative domains",
      format: Array,
      default: [],
    },
  },
  file_sources: {
    source_files: {
      doc: "Trusted source files for this application",
      format: Array,
      default: ["DNI"],
    },
    authoritative_files: {
      doc: "Server-level authoritative files",
      format: Array,
      default: ["DNI"],
    },
  },
  settings: {
    show_sources: {
      doc: "Whether or not to show trusted sources",
      format: "Boolean",
      default: true,
    },
    refresh_sources_on_startup: {
      doc: "Always pull fresh data from all sources on restart of the app.",
      format: "Boolean",
      default: false,
    },
    enable_public_comments: {
      doc: "Will push public comments to subscriber servers when true",
      format: "Boolean",
      default: false,
    },
    min_tier_percent: {
      doc: "The percentage threshold for the highest tier of consensus required.",
      format: "nat",
      default: 80,
    },
    mod_tier_percent: {
      doc: "The percentage threshold for the moderate tier of consensus required.",
      format: "nat",
      default: 66,
    },
    max_tier_percent: {
      doc: "The percentage threshold for the lowest tier of consensus required.",
      format: "nat",
      default: 51,
    },
    block_refresh_latency: {
      doc: "How often we refresh data pulled from trusted sources",
      format: "nat",
      default: 60,
    },
    token_verify_latency: {
      doc: "How often we verify access tokens",
      format: "nat",
      default: 10,
    },
    runner_schedule_in_minutes: {
      doc: "The runner handles all bulk operations against the rate limit and should run frequently but can be tuned to your application. Only one runner thread will run at a time.",
      format: "nat",
      default: 2,
    },
    server_push_limit: {
      doc: "The maximum number of requests allowed to be sent to any server at once",
      format: "nat",
      default: 200,
    },
    server_push_fuzz: {
      doc: "The amount of reserve we must leave in the tank of available requests for a given server. Will prevent maxing out the rate limit.",
      format: "nat",
      default: 50,
    },
  },
  restrictions: {
    allowed_servers: {
      doc: "An array of servers allowed to access this application. Takes precedence over denied servers.",
      format: Array,
      default: [],
    },
    denied_servers: {
      doc: "An array of servers denied access to this application.",
      format: Array,
      default: [],
    },
  },
});

convict.addParser({ extension: "toml", parse: toml.parse });

config.loadFile(path.join(__dirname, "../config.toml"));

// Perform validation
config.validate({ allowed: "strict" });

export default config;
