/**
 * Admin Controller
 * 
 * This controller handles all admin-related operations and now includes enhanced error handling.
 * Errors are passed down the middleware chain using the `next()` function, which maintains
 * the original behavior of the application (no further operations after an error).
 * 
 * For more information on the error handling process, see `errorHandler.js`.
 * 
 * @module controllers/AdminController
 * @requires errorHandler
 */


import { getDomainDigest } from "../util.js";
import { writeDomainAuditsToBuffer } from "../blocklist.js";
import * as queries from "../queries.js";
import { flash } from "@prairielearn/flash";
import { eq } from "drizzle-orm";
import { db } from "../db.js";
import * as Schema from "../db/schema.js";
import { errorLogs } from "../db/schema.js";

export async function pendingDomains(req, res, next) {
  try {
    const pendingDomains = await queries.getDomainsList(false);
    res.render("admin/pending.nj", {
      pendingDomains,
    });
  } catch (err) {
    next(err);
  }
}

export async function approvedDomains(req, res, next) {
  try {
    const approvedDomains = await queries.getDomainsList(true);
    res.render("admin/approved.nj", {
      approvedDomains,
    });
  } catch (err) {
    next(err);
  }
}

export async function inactiveDomains(req, res, next) {
  try {
    const inactiveDomains = await queries.getInactiveDomains();
    res.render("admin/inactive.nj", {
      inactiveDomains,
    });
  } catch (err) {
    next(err);
  }
}

export async function unknownDomains(req, res, next) {
  try {
    const unknownDomains = await queries.getUnknownDomains();
    res.render("admin/unknown.nj", {
      unknownDomains,
    });
  } catch (err) {
    next(err);
  }
}

export async function auditDomains(req, res, next) {
  try {
    res.render("admin/audit.nj", {});
  } catch (err) {
    next(err);
  }
}

export async function accounts(req, res, next) {
  try {
    const accounts = await queries.getAccounts();
    res.render("admin/accounts.nj", {
      accounts,
    });
  } catch (err) {
    next(err);
  }
}

export async function unapproveDomain(req, res, next) {
  try {
    const [domain] = await queries.unapproveDomain(req.params.domain_id);
    flash("success", `Domain have been unapproved: ${domain.domain}!`);
    res.redirect("/admin/approved");
  } catch (err) {
    next(err);
  }
}

export async function approveDomain(req, res, next) {
  try {
    const [domain] = await queries.approveDomain(req.params.domain_id);
    flash("success", `Domain have been approved: ${domain.domain}!`);
    res.redirect("/admin/pending");
  } catch (err) {
    next(err);
  }
}

export async function deactivateDomain(req, res, next) {
  try {
    const [domain] = await queries.deactivateDomain(req.params.domain_id);
    flash("success", `Domain have been marked as inactive: ${domain.domain}!`);
    res.redirect("back");
  } catch (err) {
    next(err);
  }
}

export async function activateDomain(req, res, next) {
  try {
    const [domain] = await queries.activateDomain(req.params.domain_id);
    flash("success", `Domain have been marked as active: ${domain.domain}!`);
    res.redirect("/admin/inactive");
  } catch (err) {
    next(err);
  }
}

export async function approveAllPendingDomains(_req, res, next) {
  try {
    await queries.approveAllPendingDomains();
    flash("success", "All pending domains have been approved!");
    res.redirect("/admin/approved");
  } catch (err) {
    next(err);
  }
}

export async function unapproveAllDomains(_req, res, next) {
  try {
    await queries.unapproveAllDomains();
    flash("success", "All domains have been unapproved!");
    res.redirect("/admin/pending");
  } catch (err) {
    next(err);
  }
}

export async function deactivateAllPendingDomains(_req, res, next) {
  try {
    await queries.deactivateAllPendingDomains();
    flash("success", "All domains have been marked inactive!");
    res.redirect("/admin/inactive");
  } catch (err) {
    next(err);
  }
}

// basic api endpoint for matching an unknown domain
export async function guessDomainForDigest(req, res, next) {
  const domain = req.body.domain;
  const digest = req.params.digest;

  if (!domain || !digest) {
    res.status(400).send({ error: "Bad Request" });
    return;
  }

  try {
    const unknown = await queries.getUnknownDomain(digest);

    if (!unknown) {
      res.status(404).send({ error: "Not Unknown" });
      return;
    }

    if (getDomainDigest(domain) !== digest) {
      res.status(404).send({ error: "Digest did not match" });
      return;
    }

    const results = await db.transaction(async (tx) => {
      await tx
        .delete(Schema.unknown_domains)
        .where(eq(Schema.unknown_domains.digest, digest));

      return await tx
        .insert(Schema.domains)
        .values({
          domain,
          digest: getDomainDigest(domain),
        })
        .onConflictDoNothing()
        .returning({ id: Schema.domains.id, domain: Schema.domains.domain });
    });

    res.status(200).send({ result: results[0] });
  } catch (err) {
    next(err);
  }
}

export async function downloadDomainAudit(_req, res, next) {
  try {
    const audits = await queries.getDomainAudits();
    const csv = await writeDomainAuditsToBuffer(audits);
    res.header("Content-Type", "text/csv");
    res.attachment("domain_audit_file.csv");
    return res.send(csv);
  } catch (err) {
    next(err);
  }
}

/**
 * Fetches error logs and renders them on the admin dashboard.
 *
 * @param {Object} req - The Express request object.
 * @param {Object} res - The Express response object.
 */
export async function errorLogsPage(req, res, next) {
  try {
    // Fetch error logs from the database
    const logs = await db.select().from(errorLogs).orderBy('timestamp', 'desc').limit(100);

    // Fetch account and server details for each log
    const logsWithDetails = await Promise.all(logs.map(async (log) => {
      const accountId = log.user_id;
      const serverId = log.context.serverId;

      let account = null;
      let server = null;

      if (accountId) {
        account = await queries.getAccountById(accountId);
        
      }

      if (serverId) {
        server = await queries.getServerById(serverId);
      }

      return {
        ...log,
        account,
        server,
      };
    }));

    // Group logs by account
    const groupedLogs = logsWithDetails.reduce((acc, log) => {
      const accountKey = log.user_id ? log.user_id : 'Unknown';

      if (!acc[accountKey]) {
        acc[accountKey] = {
          account: log.account,
          server: log.server,
          logs: [],
        };
      }
      acc[accountKey].logs.push(log);
      return acc;
    }, {});

    console.log(groupedLogs); // Log the grouped data to verify

    // Render the error logs page with grouped account and server details
    res.render("admin/error_logs.nj", { groupedLogs });
  } catch (err) {
    console.error("Failed to fetch error logs:", err);
    res.status(500).send("Failed to load error logs.");
  }
}

export async function domainChanges(req, res) {
  try {
    const page = parseInt(req.query.page) || 1;
    const limit = parseInt(req.query.limit) || 40;

    const searchTerm = req.query.q?.trim();
    
    const { changes, pagination } = await queries.getRecentDomainChanges(page, limit, searchTerm);
    
    const formattedChanges = changes.map(change => ({
      ...change,
      datetime: change.datetime ? new Date(change.datetime).toISOString().replace('T', ', ').slice(0, 20) : '',
      percentChange: Number((Number(change.new_percent) - Number(change.previous_percent)).toFixed(1)),
      absoluteChange: Math.abs(Number(change.new_percent) - Number(change.previous_percent)).toFixed(1),
      threshold: change.new_percent >= 80 ? 80 : 
                change.new_percent >= 66 ? 66 : 
                change.new_percent >= 51 ? 51 : null,
      previousThreshold: change.previous_percent >= 80 ? 80 : 
                        change.previous_percent >= 66 ? 66 : 
                        change.previous_percent >= 51 ? 51 : null,
      authoritative: change.authoritative || false,
      is_active: change.is_active
    }));

    // Debug log to verify data
    console.log('First change:', formattedChanges[0]);

    // Calculate pagination display values
    const paginationDisplay = {
      ...pagination,
      startItem: (pagination.page - 1) * pagination.pageSize + 1,
      endItem: Math.min(pagination.page * pagination.pageSize, pagination.total)
    };

    return res.render("admin/domain-changes.nj", {
      recentChanges: formattedChanges,
      pagination: paginationDisplay,
      searchTerm,
      title: "Domain Changes History"
    });
  } catch (error) {
    req.log.error("Error fetching domain changes:", error);
    flash("error", "Failed to load domain changes history");
    return res.redirect("/admin");
  }
}

export async function searchDomains(req, res, next) {
  try {
    const searchTerm = req.query.q?.trim();
    
    if (!searchTerm) {
      return res.status(400).json({ 
        error: 'Search term is required'
      });
    }
    const results = await queries.searchDomains(searchTerm);
    
    // Format the results for display
    const formattedResults = results.map(domain => ({
      ...domain,
      approved_at: domain.approved_at ? new Date(domain.approved_at).toLocaleString() : null,
      percent: domain.percent ? Number(domain.percent).toFixed(2) : null,
      status: domain.active ? 
        (domain.approved_at ? 'Approved' : 'Pending') : 
        'Inactive'
    }));
    // If it's an AJAX request, return JSON
    if (req.xhr || req.headers.accept?.includes('application/json')) {
      return res.json({ results: formattedResults });
    }
    // Otherwise render the search results page
    res.render("admin/search.nj", {
      searchTerm,
      searchResults: formattedResults
    });
  } catch (err) {
    next(err);
  }
}

export async function getAccountById(req, res, next) {
  try {
    const accountId = req.params.account_id;

    // Remove UUID validation if accountId is not a UUID
    const account = await queries.getAccountById(accountId);

    if (!account) {
      res.status(404).send({ error: "Account not found" });
      return;
    }

    res.status(200).send({ account });
  } catch (err) {
    next(err);
  }
}

export async function clearErrorLogs(req, res, next) {
  try {
    // Use the correct method to delete all entries from the error_logs table
    await db.delete(errorLogs).execute(); // Adjust based on your ORM's syntax
    // Redirect back to the error logs page
    res.redirect('/admin/error_logs');
  } catch (error) {
    next(error); // Pass the error to the next middleware
  }
}

export async function getDomainHistory(req, res) {
  try {
    const { domain } = req.params;
    const history = await queries.getDomainAuditHistory(domain);
    
    const formattedHistory = history.map(change => ({
      ...change,
      percentChange: Number((Number(change.new_percent) - Number(change.previous_percent)).toFixed(1)),
    }));
    
    res.json(formattedHistory);
  } catch (error) {
    req.log.error("Error fetching domain history:", error);
    res.status(500).json({ error: "Failed to load domain history" });
  }
}
