import * as queries from "../queries.js";
import { flash } from "@prairielearn/flash";

/**
 * Domain Change Controller
 * 
 * This controller handles the domain changes history page,
 * showing the 10 most recent domain changes.
 * 
 * @module controllers/DomainChangeController
 * @requires errorHandler
 */
export async function domainChanges(req, res) {
    try {
      // Fetch more items initially to ensure we get 10 unique domains
      const limit = 50; // Increased limit to get more items
      const page = 1; // Always show first page since we're only showing most recent
      
      const { changes, pagination } = await queries.getRecentDomainChanges(page, limit);
      
      const formattedChanges = changes.map(change => ({
        ...change,
        datetime: change.datetime ? new Date(change.datetime).toISOString().replace('T', ', ').slice(0, 20) : '',
        percentChange: Number((Number(change.new_percent) - Number(change.previous_percent)).toFixed(1)),
        absoluteChange: Math.abs(Number(change.new_percent) - Number(change.previous_percent)),
        threshold: change.new_percent >= 80 ? 80 : 
                  change.new_percent >= 66 ? 66 : 
                  change.new_percent >= 51 ? 51 : null,
        previousThreshold: change.previous_percent >= 80 ? 80 : 
                          change.previous_percent >= 66 ? 66 : 
                          change.previous_percent >= 51 ? 51 : null,
        authoritative: change.authoritative || false,
        is_active: change.is_active
      }));

      // Group changes by domain and take only the most recent 10 domains
      const domainGroups = new Map();
      formattedChanges.forEach(change => {
        if (!domainGroups.has(change.domainname) && change.is_active) {
          domainGroups.set(change.domainname, change);
        }
      });

      // Convert map back to array and take first 10
      const recentDomainChanges = Array.from(domainGroups.values()).slice(0, 10);

      // Debug log to check the number of items
      console.log('Number of unique domains:', domainGroups.size);
      console.log('Number of changes shown:', recentDomainChanges.length);
  
      // Calculate display values
      const paginationDisplay = {
        page: 1,
        pageSize: 10,
        total: recentDomainChanges.length,
        startItem: 1,
        endItem: recentDomainChanges.length
      };
  
      return res.render("domain-changes/index.nj", {
        recentChanges: recentDomainChanges,
        pagination: paginationDisplay,
        title: "Recent Domain Changes"
      });
    } catch (error) {
      req.log.error("Error fetching domain changes:", error);
      flash("error", "Failed to load domain changes history");
      return res.redirect("/");
    }
  }