import { flash } from "@prairielearn/flash";
import crypto from "node:crypto";

import { api } from "../api.js";
import {
  getClientCredentials,
  sanitizeServerInput,
  purgeRedisClientCredentials,
  checkServerPermissions,
  setClientCredentials,
} from "../util.js";
import { ServerRepository } from "../repositories/ServerRepository.js";
import { AccountRepository } from "../repositories/AccountRepository.js";
import env from "../env.js";
import { errorMiddleware } from "../middleware/errorHandler.js";

const REMEMBER_SERVER_URL_COOKIE = "rememberedServerUrl";

/**
 * Initiates the login process by validating server input and generating OAuth credentials.
 * Handles errors using errorMiddleware for consistent error management.
 */
export function login(req, res, next) {
  // Redirect logged in users to the homepage:
  try {
    if (req.session.account_id && req.session.server_id) {
      res.redirect("/dashboard");
      return;
    }

    // get sanitized serverUrl from cookie if it exists
    if (req.cookies[REMEMBER_SERVER_URL_COOKIE]) {
      const serverUrl = sanitizeServerInput(
        req.cookies[REMEMBER_SERVER_URL_COOKIE],
      );
      if (serverUrl) {
        const serverName = new URL(serverUrl).hostname;
        res.locals.serverName = serverName;
      }
    }

    res.render("login.nj");
  } catch (err) {
    next(err);
  }
}

export function logout(req, res, next) {
  try {
    res.clearCookie(REMEMBER_SERVER_URL_COOKIE);

    req.session.destroy((err) => {
      if (err) {
        req.log.error(
          { error: err.toString(), stack: err.stack },
          "Error destroying session",
        );
      }

      res.redirect("/login");
    });
  } catch (err) {
    next(err);
  }
}

/**
 * Starts the OAuth process by validating server input and generating OAuth credentials.
 * Handles errors using errorMiddleware for consistent error management.
 */
export async function start(req, res, next) {
    // Validate server name input
    if (!req.body.serverName) {
      flash("error", "Missing server name");
      return res.redirect("back");
    }

    // Ensure server uses HTTPS
    if (req.body.serverName.startsWith("http://")) {
      flash("error", "FediCheck only supports servers over HTTPS");
      return res.redirect("back");
    }

    // Sanitize and validate server URL
    let serverUrl = sanitizeServerInput(req.body.serverName);
    if (!serverUrl) {
      flash("error", "Invalid server name entered");
      return res.redirect("back");
    }

    const serverName = new URL(serverUrl).hostname;

    // Check server permissions
    if (!checkServerPermissions(serverName)) {
      flash("error", "You cannot log in at this time");
      return res.redirect("back");
    }

    let credentials;

    try {
      // Retrieve or register OAuth credentials
      credentials = await getClientCredentials(serverName).then(
        async (existingCredentials) => {
          if (
            existingCredentials &&
            existingCredentials.client_id &&
            existingCredentials.client_secret
          ) {
            return existingCredentials;
          } else {
            const newCredentials = await api.registerApplication(
              serverName,
              req.log,
            );

            await setClientCredentials(serverName, newCredentials);

            return newCredentials;
          }
        },
      );
    } catch (err) {
      // Handle errors with errorMiddleware
      errorMiddleware(err, req, res, () => {
        flash("error", "An unexpected error occurred: unable to create client");
        res.redirect("back");
      });
      return;
    }

    // Manage "remember me" cookie
    if (req.body.rememberMe) {
      res.cookie(REMEMBER_SERVER_URL_COOKIE, serverUrl, {
        maxAge: 1000 * 60 * 60 * 24 * 30, // 30 days
        httpOnly: true,
        secure: env.isProduction,
      });
    } else {
      res.clearCookie(REMEMBER_SERVER_URL_COOKIE);
    }

    // In order to prevent session replay attacks, we generate a unique state value
    // which is checked when handling the OAuth Callback against the value stored
    // in the Session
    const oauth_state = crypto.randomBytes(64).toString("base64url");

    // Construct the OAuth authorization URL
    const redirectUrl = api.authorizeUrl(
      serverUrl,
      credentials.client_id,
      oauth_state,
      false, // Disable authorize prompt for existing users
    );

    // Regenerate session to prevent session fixation
    req.session.regenerate((err) => {
      if (err) {
        errorMiddleware(err, req, res, () => {
          flash("error", "An unexpected error occurred");
          res.redirect("back");
        });
        return;
      }

      // Store session data
      req.session.server_url = serverUrl;
      req.session.server_name = serverName;
      req.session.oauth_state = oauth_state;

      res.redirect(redirectUrl);
    });
}

/**
 * Handles the OAuth callback, verifying credentials and managing session data.
 */
export async function callback(req, res, next) {
  try {
    const sessionOAuthState = req.session.oauth_state;
    const serverName = req.session.server_name;
    const serverUrl = req.session.server_url;

    // Handle OAuth errors
    if (req.query.error) {
      if (req.query.error === "access_denied") {
        flash("notice", "You denied FediCheck access");
        return res.redirect("/login");
      } else {
        const error = new Error("Unknown OAuth Error returned from server");
        error.details = {
          error: req.query.error,
          error_description: req.query.error_description,
        };
        errorMiddleware(error, req, res, () => {
          flash("error", "Failed to authenticate");
          res.redirect("/login");
        });
        return;
      }
    }

    // Validate OAuth query parameters
    if (!req.query.code || !req.query.state) {
      const error = new Error("OAuth: Missing code or state query parameters");
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    }

    // Validate session state
    if (!sessionOAuthState) {
      const error = new Error("OAuth: Missing req.session.oauth_state");
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    }

    // Check for state mismatch
    if (sessionOAuthState !== req.query.state) {
      const error = new Error("OAuth: Mismatch between req.session.oauth_state and req.query.state");
      error.details = {
        oauth_state: sessionOAuthState,
        state: req.query.state,
      };
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    }

    // Validate session server information
    if (!serverName) {
      const error = new Error("Missing req.session.server_name");
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    }

    if (!serverUrl) {
      const error = new Error("Missing req.session.server_url");
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    }

    // Retrieve OAuth credentials
    const credentials = await getClientCredentials(serverName).catch((err) => {
      const error = new Error("Failed to retrieve oauth credentials for the server");
      error.details = {
        error: err.toString(),
        stack: err.stack,
        serverName,
      };
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.redirect("/login");
      });
      return;
    });

    // Check for missing credentials
    if (!credentials || !credentials.client_id || !credentials.client_secret) {
      const error = new Error("Missing oauth credentials for the server");
      error.details = { serverName };
      errorMiddleware(error, req, res, () => {
        flash("error", "Failed to authenticate");
        res.status(401).send("Authentication failed due to missing credentials.");
      });
      return;
    }

    try {
      // Exchange code for access token
      const data = await api.get_access_token(
        credentials.client_id,
        credentials.client_secret,
        serverName,
        req.query.code,
      );

      // Handle token exchange errors
      if (data.error) {
        const error = new Error("OAuth: Error exchanging code for access token");
        error.details = { error: data };
        errorMiddleware(error, req, res, () => {
          flash("error", data.error_description ?? "An unknown error occurred");
          res.redirect("/login");
        });
        return;
      }

      // Verify application credentials
      const verifiedCredentials = await api.verify_app_credentials(
        serverName,
        data.access_token,
        null,
        res.log,
      );

      if (!verifiedCredentials) {
        await purgeRedisClientCredentials(serverName);

        flash("error", "Your credentials have expired, please login again");
        return res.redirect("/login");
      }

      // Verify account credentials
      const account = await api
        .verify_account_credentials(serverName, data.access_token, res.log)
        .catch((err) => {
          const error = new Error("OAuth: Error verifying account credentials");
          error.details = { error: err.toString(), stack: err.stack };
          errorMiddleware(error, req, res, () => {
            flash("error", "Failed to verify account credentials");
            res.status(401).send("Account verification failed.");
          });
          return;
        });

      if (!account || account?.error) {
        await purgeRedisClientCredentials(serverName);

        const error = new Error("Failed to verify account credentials");
        errorMiddleware(error, req, res, () => {
          flash("error", "Failed to verify account credentials");
          res.status(401).send("Account verification failed.");
        });
        return;
      }

      // Test domain block permissions
      const canReadDomainBlocks = await api.test_domain_blocks(
        serverName,
        data.access_token,
      );

      if (!canReadDomainBlocks) {
        await purgeRedisClientCredentials(serverName);
        await AccountRepository.deleteAccount(serverName, account.id);

        flash("error", "Your credentials have insufficient privileges");
        return res.redirect("/login");
      }

      // Create or find server and account records
      const serverId = await ServerRepository.findOrCreate({
        name: serverName,
        url: serverUrl,
        accountId: account.id,
        credentials,
      });

      const accountId = await AccountRepository.findOrCreate({
        id: account.id,
        username: account.username,
        display_name: account.display_name,
        access_token: data.access_token,
        server: { id: serverId },
      });

      // Regenerate session and store account information
      req.session.regenerate((err) => {
        if (err) {
          errorMiddleware(err, req, res, () => {
            flash("error", "An unexpected error occurred");
            res.redirect("back");
          });
          return;
        }

        req.session.server_id = serverId;
        req.session.account_id = accountId;

        return res.redirect("/dashboard");
      });
    } catch (err) {
      errorMiddleware(err, req, res, () => {
        flash("error", "An unknown error occurred");
        res.redirect("back");
      });
    }
  } catch (err) {
    errorMiddleware(err, req, res, () => {
      flash("error", "An unknown error occurred");
      res.redirect("back");
    });
  }
}
