import { flash } from "@prairielearn/flash";
import { ServerRepository } from "../repositories/ServerRepository.js";
import { getServerConfigById } from "../queries.js";
import { purgeRedisClientCredentials, revokeAccessTokens } from "../util.js";
import { errorMiddleware } from "../middleware/errorHandler.js";

export function reset(req, res) {
  res.render("settings/reset.nj");
}

export async function updatePushStatus(req, res) {
  // @ts-ignore
  const serverId = req.session.server_id;

  if (
    !req.body.push_status ||
    !["enable", "disable"].includes(req.body.push_status)
  ) {
    flash("error", "An unknown error occurred");
    res.redirect("back");
    return;
  }

  let pushStatus = req.body.push_status === "enable";

  if (res.locals.block_wipe_in_progress && pushStatus === true) {
    flash("error", "Cannot enable push during a block wipe.");
    res.redirect("back");
    return;
  }

  try {
    await ServerRepository.updatePushStatus(serverId, pushStatus);

    flash(
      "success",
      pushStatus
        ? `Push enabled! We will start pushing changes to your server soon.`
        : `Push disabled! You will no longer see updates from FediCheck on your server`,
    );
    res.redirect("back");
  } catch (err) {
    errorMiddleware(err, req, res, () => {
      flash("error", "Failed to update push status");
      res.redirect("back");
    });
  }
}

export async function updatePushAccount(req, res) {
  // @ts-ignore
  const serverId = req.session.server_id;
  const accountId = req.body.api_account_id;

  if (!accountId) {
    flash("error", "Missing Push API Account ID");
    res.redirect("back");
    return;
  }

  try {
    await ServerRepository.updateApiAccount(serverId, accountId);
    flash("success", "Push API Account updated successfully!");
  } catch (err) {
    if (err && err.code === "INVALID_ACCOUNT") {
      flash(
        "error",
        "Failed to update Push API Account, the selected account does not have a valid access token!",
      );
      res.redirect("back");
      return;
    } else {
      // Handle other errors with errorMiddleware to send to Sentry
      errorMiddleware(err, req, res, () => {
        flash("error", "An unexpected error occurred while updating the Push API Account");
        res.redirect("back");
      });
    }
  }
}

export async function requestWipe(req, res, next) {
  // @ts-ignore
  const serverId = req.session.server_id;

  // We can't trigger two wipes at once.
  if (res.locals.block_wipe_in_progress) {
    flash("error", "Block wipe already in progress");
    res.redirect("/");
    return;
  }

  try {
    await ServerRepository.requestWipe(serverId);
    // Note: we don't display a flash message on success here, since there's a
    // message displayed based on `res.locals.block_wipe_in_progress`
    res.redirect("back");
  } catch (err) {
    // Pass error to errorMiddleware
    return errorMiddleware(err, req, res, () => {
      flash("error", "Failed to request block wipe");
      res.redirect("back");
    });
  }
}
// TODO: Refactor to use an async job:
export async function revoke(req, res) {
  // @ts-ignore
  const serverId = req.session.server_id;
  const server_config = await getServerConfigById(serverId);

  try {
    // Revoke all access tokens & credentials
    await revokeAccessTokens(server_config);
    await purgeRedisClientCredentials(server_config.server);

    // Delete all data associated with the server
    await ServerRepository.deleteServer(serverId);

    req.session.destroy((err) => {
      if (err) {
        // Handle session destruction error with errorMiddleware
        errorMiddleware(err, req, res, () => {
          flash("error", "An error occurred while logging out");
          res.redirect("/login");
        });
        return;
      }

      res.redirect("/login");
    });
  } catch (err) {
    // Handle errors with errorMiddleware to send to Sentry
    errorMiddleware(err, req, res, () => {
      flash("error", "An unexpected error occurred during revocation");
      res.redirect("back");
    });
  }
}
