import crypto from "crypto";
import env from "./env.js";

const ENCRYPTION_ALGORITHM = "aes-256-gcm";
const ENCRYPTION_KEY_LENGTH = 32;
const ENCRYPTION_IV_LENGTH = 12;
const ENCRYPTION_AUTHTAG_LENGTH = 16;

const DATABASE_ENCRYPTION_SALT = "database_encryption";
const DATABASE_ENCRYPTION_ENCODING = "base64url";

const SECRET_CACHE = new Map();
const SECRET_GENERATOR_ITERATIONS = 2 ** 16;

/**
 * Generates a derived secret key for varying purposes, using the
 * SECRET_KEY_BASE as the input key. The generated secrets are cached, such that
 * each usage for the same salt and keylen doesn't have to block
 * @param {string} salt The secret key salt
 * @param {number} [keylen=64] The key length in bytes e.g., for a AES-256-CBC
 * algorithm key, the keylen should be 256 / 8 = 32
 * @returns {Buffer} The generated secret key
 */
export function generateSecretKey(salt, keylen = 64) {
  const cache_key = `${salt}:${keylen}`;
  if (SECRET_CACHE.has(cache_key)) {
    return SECRET_CACHE.get(cache_key);
  }

  const secret_key = crypto.pbkdf2Sync(
    // env.SECRET_KEY_BASE is validated by envalid as being at least 128 characters
    env.SECRET_KEY_BASE,
    salt,
    SECRET_GENERATOR_ITERATIONS,
    keylen,
    "SHA1",
  );

  SECRET_CACHE.set(cache_key, secret_key);

  return secret_key;
}

export function encryptString(text) {
  const key = generateSecretKey(
    DATABASE_ENCRYPTION_SALT,
    ENCRYPTION_KEY_LENGTH,
  );

  return encrypt(text, key);
}

export function decryptString(value) {
  const key = generateSecretKey(
    DATABASE_ENCRYPTION_SALT,
    ENCRYPTION_KEY_LENGTH,
  );

  return decrypt(value, key);
}

export function encrypt(text, key) {
  // For AES-256-GCM, the key length must be 32 bytes:
  if (key.byteLength !== ENCRYPTION_KEY_LENGTH) {
    throw new TypeError("Invalid key-size passed to encrypt()");
  }

  // generate a random iv for each encryption
  const iv = crypto.randomBytes(ENCRYPTION_IV_LENGTH);
  const cipher = crypto.createCipheriv(ENCRYPTION_ALGORITHM, key, iv, {
    authTagLength: ENCRYPTION_AUTHTAG_LENGTH,
  });

  let encrypted = cipher.update(text, "utf8", DATABASE_ENCRYPTION_ENCODING);
  encrypted += cipher.final(DATABASE_ENCRYPTION_ENCODING);

  return [
    encrypted,
    iv.toString(DATABASE_ENCRYPTION_ENCODING),
    cipher.getAuthTag().toString(DATABASE_ENCRYPTION_ENCODING),
  ].join(":");
}

export function decrypt(value, key) {
  // For AES-256-GCM, the key length must be 32 bytes:
  if (key.byteLength !== ENCRYPTION_KEY_LENGTH) {
    throw new TypeError("Invalid key-size passed to decrypt()");
  }

  const parts = value.split(":", 3);
  if (parts.length !== 3) {
    throw new TypeError("Invalid value supplied to decrypt()");
  }

  const [encrypted, iv, authTag] = parts;

  const ivBuffer = Buffer.from(iv, DATABASE_ENCRYPTION_ENCODING);
  if (ivBuffer.byteLength != ENCRYPTION_IV_LENGTH) {
    throw new TypeError(
      "Invalid iv byte length in value passed to decryptString",
    );
  }

  const authTagBuffer = Buffer.from(authTag, DATABASE_ENCRYPTION_ENCODING);
  if (authTagBuffer.byteLength != ENCRYPTION_AUTHTAG_LENGTH) {
    throw new TypeError(
      "Invalid authTag byte length in value passed to decryptString",
    );
  }

  const decipher = crypto.createDecipheriv(
    ENCRYPTION_ALGORITHM,
    key,
    ivBuffer,
    {
      authTagLength: ENCRYPTION_AUTHTAG_LENGTH,
    },
  );

  decipher.setAuthTag(authTagBuffer);

  let decrypted = decipher.update(
    encrypted,
    DATABASE_ENCRYPTION_ENCODING,
    "utf8",
  );

  decrypted += decipher.final("utf8");

  return decrypted;
}

export const generateAuthToken = () => {
  return crypto.randomUUID();
};
