// reads remote CSV file and imports domains
import { parse } from "csv-parse";
import { stringify } from "csv-stringify";
import * as constants from "./constants.js";

/**
 * This is the maximum number of rows allowed in Mastodon Blocklist CSV files:
 */
const MAX_CSV_RECORDS = 20_000;

/**
 * Converts the Mastodon style CSV header row into standard headers
 * @param {string[]} header
 * @returns {string[]}
 */
const CONVERT_HEADER_ROW = function (header) {
  // Strip the leading # from Mastodon CSV column names:
  return header.map((column) => {
    if (column.startsWith("#")) {
      return column.slice(1);
    } else {
      return column;
    }
  });
};

/**
 * @param {string|Buffer} contents
 * @param {csvFormat} format
 * @returns {Promise<DomainBlock[]>}
 */
export async function parseDomainBlocksCsv(contents, format) {
  return new Promise((resolve, reject) => {
    parse(
      contents,
      {
        to: MAX_CSV_RECORDS + 1,
        delimiter: ",",
        columns: format === "domain_only_csv" ? ["domain"] : CONVERT_HEADER_ROW,
      },
      /**
       *
       * @param {import('csv-parse').CsvError | undefined} err
       * @param {Array<Object | undefined>} records
       * @param {import('csv-parse').Info} info
       */
      (err, records, info) => {
        if (err) {
          reject(err);
          return;
        }

        // This is a check that Mastodon enforces, so we should never see more
        // than 20,000 rows in a CSV file:
        if (info.records > MAX_CSV_RECORDS) {
          reject(
            new Error(
              "CSV file contains more than 20,000 records which is the maximum supported by Mastodon",
            ),
          );
          return;
        }

        /**
         * @param {Record<string, string>} record
         * @returns {DomainBlock|undefined}
         */
        let parser = (record) => {
          return;
        };

        switch (format) {
          case "domain_only_csv":
            parser = domainOnlyParser;
            break;
          case "fediblockhole":
            parser = fediblockholeParser;
            break;
          case "mastodon_csv":
          default:
            parser = mastodonParser;
            break;
        }

        const parsed_records = records.map((record) => parser(record));
        const domain_blocks = parsed_records.filter(
          (block) => block !== undefined,
        );

        // @ts-ignore
        resolve(domain_blocks);
      },
    );
  });
}

/**
 * @param {Record<string, any>[]} rows
 * @param {string[]} columns
 * @param {Record<string, string>|true} headers
 * @returns {Promise<string>}
 */
export async function writeCsv(rows, columns, headers) {
  return new Promise((resolve, reject) => {
    const csv = stringify({
      header: typeof headers !== "object" && headers,
      columns: columns,
    });

    const chunks = [];

    csv.on("readable", () => {
      let chunk;
      while (null !== (chunk = csv.read())) {
        chunks.push(chunk);
      }
    });

    csv.on("end", () => {
      resolve(chunks.join(""));
    });

    // Write the CSV headers:
    if (typeof headers === "object") {
      csv.write(headers);
    }

    rows.forEach((row) => {
      csv.write(row);
    });

    csv.end();
  });
}

/**
 * Generates a CSV string from domain block data, optionally including public comments.
 * @param {DomainBlock[]} domain_blocks - An array of domain block objects.
 *
 * @returns {Promise<string>} A promise that resolves to a CSV string.
 * #domain,#severity[,#public_comment]
 * example.com,suspend[,"Optional public comment"]
 * badactor.net,silence
 *
 */
export async function writeDomainBlocksCsv(domain_blocks) {
  const columns = ["domain", "severity"];

  // add public comments to the export file if enabled
  if (constants.ENABLE_PUBLIC_COMMENTS) {
    columns.push("public_comment");
  }

  return writeCsv(domain_blocks, columns, {
    domain: "#domain",
    severity: "#severity",
    public_comment: "#public_comment",
  });
}

export async function writeAuditCsv(domain_audits) {
  const columns = [
    "domain",
    "percent",
    "severity",
    "count",
    "count_suspend",
    "count_silence",
    "count_noop",
    "authoritative",
    "active",
  ];

  // add public comments to the export file if enabled
  if (constants.ENABLE_PUBLIC_COMMENTS) {
    columns.push("public_comment");
  }

  return writeCsv(domain_audits, columns, true);
}

// The value is converted to lowercase to handle Python True values:
const castBoolean = (value) => (value.toLowerCase() === "true" ? true : false);

const BLOCK_SEVERITIES = ["suspend", "silence", "noop"];

/**
 * @param {Record<string, string>} record
 * @returns {DomainBlock | undefined}
 */
function mastodonParser(record) {
  if (
    !record.domain ||
    !record.severity ||
    !BLOCK_SEVERITIES.includes(record.severity)
  ) {
    // TODO: Log skipped record due to malformed data?
    return;
  }

  return {
    domain: record.domain,
    severity: record.severity,
    reject_media: castBoolean(record.reject_media),
    reject_reports: castBoolean(record.reject_reports),
    public_comment: record.public_comment ?? "",
    obfuscate: castBoolean(record.obfuscate),
  };
}

/**
 * Note: This implementation is currently the same as mastodonParser, so may be
 * redundant.
 * @param {Record<string, string>} record
 * @returns {DomainBlock | undefined}
 */
function fediblockholeParser(record) {
  if (
    !record.domain ||
    !record.severity ||
    !BLOCK_SEVERITIES.includes(record.severity)
  ) {
    // TODO: Log skipped record due to malformed data?
    return;
  }

  return {
    domain: record.domain,
    severity: record.severity,
    reject_media: castBoolean(record.reject_media),
    reject_reports: castBoolean(record.reject_reports),
    public_comment: record.public_comment ?? "",
    obfuscate: castBoolean(record.obfuscate),
  };
}

/**
 * @param {Record<string, string>} record
 * @returns {DomainBlock | undefined}
 */
function domainOnlyParser(record) {
  if (!record.domain) {
    // TODO: Log skipped record due to malformed data?
    return;
  }

  return {
    domain: record.domain,
    severity: "suspend",
    reject_media: false,
    reject_reports: false,
    public_comment: "",
    obfuscate: false,
  };
}
