  import { sql } from "drizzle-orm";
  import {
    serial,
    boolean,
    pgTable,
    pgEnum,
    varchar,
    primaryKey,
    integer,
    timestamp,
    doublePrecision,
    text,
    uniqueIndex,
    numeric,
    uuid,
    jsonb,
    bigint,
  } from "drizzle-orm/pg-core";

  // declaring enum in database
  export const blocklistTypesEnum = pgEnum("blocklist_types", [
    "minimum",
    "moderate",
    "maximum",
  ]);
  export const blocklistOptionEnum = pgEnum("blocklist_option", [
    "add",
    "edit",
    "delete",
  ]);
  export const followStatusEnum = pgEnum("follow_status", [
    "unchecked",
    "pending",
    "updating_followers",
    "ready",
  ]);

  export const blockFileFormatEnum = pgEnum("block_file_format", [
    "mastodon",
    "fediblockhole",
  ]);

  // servers are subscribers who have logged in
  // or are trusted sources with public apis or files for download
  export const servers = pgTable(
    "servers",
    {
      id: serial("id").primaryKey(),
      server: varchar("server", { length: 400 }).unique().notNull(),
      url: text("url").unique().notNull(),
      created_date: timestamp("created_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      source: boolean("source").default(false).notNull(),
      verified: boolean("verified").default(false).notNull(),
      is_file: boolean("is_file").default(false).notNull(), // if this "server" is in fact pulled from a CSV file
      authoritative: boolean("authoritative").default(false).notNull(), // a authoritative source includes its blocklist verbatim without consensus
      last_server_push_date: timestamp("last_update_date"),
      last_server_refresh_date: timestamp("last_server_refresh_date"),
      block_wipe_in_progress: boolean("block_wipe_in_progress")
        .default(false)
        .notNull(), // when true, we will ignore 'enabled' of true and wipe blocks
    },
    (servers) => {
      return {
        serverIdx: uniqueIndex("server").on(servers.server),
        urlIdx: uniqueIndex("server_url").on(servers.url),
      };
    },
  );

  export const accounts = pgTable(
    "accounts",
    {
      id: uuid("id").defaultRandom().notNull().primaryKey(),
      server_id: integer("server_id")
        .notNull()
        .references(() => servers.id, { onDelete: "cascade" }),
      account_id: numeric("account_id").notNull(),
      username: text("username").notNull(),
      display_name: text("display_name"),
      access_token: text("access_token"),
      created_date: timestamp("created_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      last_login_date: timestamp("last_login_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      last_token_verify_date: timestamp("last_token_verify_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
    },
    (account) => {
      return {
        remoteAccountIndex: uniqueIndex("accounts_remote_account_idx").on(
          account.server_id,
          account.account_id,
        ),
      };
    },
  );

  // config table holds encrypted secrets for subscriber servers
  export const server_config = pgTable("server_config", {
    server_id: integer("server_id")
      .notNull()
      .primaryKey()
      .references(() => servers.id, { onDelete: "cascade" }),
    client_id: varchar("client_id", { length: 256 }).notNull(),
    client_secret: varchar("client_secret", { length: 256 }).notNull(),
    api_account_id: numeric("api_account_id"), // primary account to be used for this server
    enabled: boolean("enabled").default(false).notNull(), // if push is enabled
    list_type: blocklistTypesEnum("list_type").default("minimum").notNull(),
  });

  // distinct list of all domains blocked across the app
  // both within source blocklists, as well as blocked by users
  export const domains = pgTable(
    "domains",
    {
      id: serial("id").primaryKey(),
      domain: text("domain").unique().notNull(),
      digest: text("digest").unique().notNull(),
      created_date: timestamp("created_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      active: boolean("active").default(true).notNull(),
      approved_at: timestamp("approved_at"),
    },
    (domains) => {
      return {
        domainIdx: uniqueIndex("domain_idx").on(domains.domain),
        digestIdx: uniqueIndex("digest_idx").on(domains.digest),
      };
    },
  );

  // obfuscated domains that have no known match
  // we save them for auditing and manual matches later
  // if necessary. A cleanup job will regularly check to delete
  // anything with a match in the domains table
  export const unknown_domains = pgTable(
    "unknown_domains",
    {
      id: uuid("id").defaultRandom().notNull().primaryKey(),
      digest: text("digest").unique().notNull(),
      domain: text("domain").unique().notNull(),
      created_date: timestamp("created_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
    },
    (domains) => {
      return {
        unknownDomainIdx: uniqueIndex("unknown_domain_idx").on(domains.domain),
      };
    },
  );

  // blocks of domains across the application, associated to a given server
  // this is pulled directly from the api for a given server
  export const blocks = pgTable(
    "blocks",
    {
      server_id: integer("server_id")
        .notNull()
        .references(() => servers.id, { onDelete: "cascade" }),
      domain_id: integer("domain_id")
        .notNull()
        .references(() => domains.id),
      block_ref: varchar("block_ref"),
      severity: varchar("severity").notNull(),
      // whether or not this is managed by our service
      // whether we created this block ourselves
      is_managed: boolean("is_managed").default(false).notNull(),
      // this field exists because some blocks might be pulled
      // from the public api (for sources)
      is_admin: boolean("is_admin").default(false).notNull(),
      safelist: boolean("safelist").default(false).notNull(),
      public_comment: text("public_comment"),
      private_comment: text("private_comment"),
      reject_media: boolean("reject_media").default(false).notNull(),
      created_date: timestamp("created_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      last_updated_date: timestamp("last_updated_date")
        .default(sql`CURRENT_TIMESTAMP`)
        .notNull(),
      removed: boolean("removed").default(false).notNull(),
      removed_date: timestamp("removed_date"),
      updated: boolean("updated").default(true).notNull(),
      option: blocklistOptionEnum("blocklist_option"), // generally never actually persisted, but assigned during managed block compilation
    },
    (blocks) => {
      return {
        pk: primaryKey({ columns: [blocks.server_id, blocks.domain_id] }),
        blockIndex: uniqueIndex("blockref_idx").on(
          blocks.block_ref,
          blocks.server_id,
        ),
      };
    },
  );

  export const domain_audits = pgTable("domain_audits", {
    domain_id: integer("domain_id")
      .primaryKey()
      .references(() => domains.id),
    count: integer("count").default(0).notNull(),
    percent: doublePrecision("percent").default(0).notNull(),
    public_comment: text("public_comment"),
    severity: varchar("severity").notNull(),
    count_suspend: integer("count_suspend").default(0).notNull(),
    count_silence: integer("count_silence").default(0).notNull(),
    count_noop: integer("count_noop").default(0).notNull(),
    authoritative: boolean("authoritative").default(false).notNull(),
    created_date: timestamp("created_date")
      .default(sql`CURRENT_TIMESTAMP`)
      .notNull(),
    updated_at: timestamp("updated_at")
      .default(sql`CURRENT_TIMESTAMP`)
      .notNull(),
  });

  // Define the error_logs table
  export const errorLogs = pgTable("error_logs", {
    id: serial("id").primaryKey(),
    message: text("message").notNull(),
    severity: varchar("severity", { length: 50 }).notNull(),
    context: jsonb("context"),
    user_id: bigint("user_id", { mode: "bigint" }),
    timestamp: timestamp("timestamp").default(sql`CURRENT_TIMESTAMP`).notNull(),
  });

  export const domain_audit_history = pgTable("domain_audit_history", {
    id: serial("id").primaryKey(),
    domain_id: integer("domain_id")
      .notNull()
      .references(() => domains.id),
    old_percent: doublePrecision("old_percent").notNull(),
    new_percent: doublePrecision("new_percent").notNull(),
    old_severity: varchar("old_severity"),
    new_severity: varchar("new_severity"),
    change_reason: varchar("change_reason"), // Add this field to track why the change occurred
    created_at: timestamp("created_at")
      .default(sql`CURRENT_TIMESTAMP`)
      .notNull(),
  });

  // Define the domain_audits_changelog table
  export const domain_audits_changelog = pgTable("domain_audits_changelog", {
      id: serial("id").primaryKey(),
      domain_id: integer("domain_id")
      .notNull()
      .references(() => domain_audits.domain_id),
      domainname: varchar("domainname", { length: 255 }).notNull(),
      threshold_crossed: varchar("threshold_crossed", { length: 50 }).notNull(), // Store the threshold crossed
      previous_percent: doublePrecision("previous_percent").notNull(), // Store the previous percent
      new_percent: doublePrecision("new_percent").notNull(), // Store the new percent
      datetime: timestamp("datetime").default(sql`CURRENT_TIMESTAMP`).notNull(),
      authoritative: boolean("authoritative").default(false).notNull(),
  });

  export const api_keys = pgTable('api_keys', {
    id: serial('id').primaryKey(),
    key: text('key').notNull().unique(),
    description: text('description'),
    created_at: timestamp('created_at').defaultNow(),
    last_used_at: timestamp('last_used_at'),
    is_active: boolean('is_active').default(true),
    is_admin: boolean('is_admin').default(false)
  }, (table) => {
    return {
      singleActiveAdminKey: uniqueIndex('idx_single_active_admin')
        .on(table.id)
        .where(sql`is_admin AND is_active`)
    }
  });

  export const api_request_logs = pgTable('api_request_logs', {
    id: serial('id').primaryKey(),
    api_key_id: integer('api_key_id').references(() => api_keys.id),
    endpoint: text('endpoint').notNull(),
    method: text('method').notNull(),
    timestamp: timestamp('timestamp').defaultNow(),
    ip_address: text('ip_address'),
    user_agent: text('user_agent')
  });

  export const admin_sessions = pgTable('admin_sessions', {
    id: text('id').primaryKey(),
    admin_id: integer('admin_id').notNull(),
    token: text('token').notNull(),
    created_at: timestamp('created_at').defaultNow(),
    expires_at: timestamp('expires_at').notNull()
  });

  export const api_key_permissions = pgTable('api_key_permissions', {
    id: serial('id').primaryKey(),
    api_key_id: integer('api_key_id')
      .notNull()
      .references(() => api_keys.id),
    permission: varchar('permission', { length: 50 }).notNull(), // e.g., 'threshold_51', 'threshold_66', 'threshold_80'
    created_at: timestamp('created_at').defaultNow()
  });

  export const api_threshold_requests = pgTable('api_threshold_requests', {
    id: serial('id').primaryKey(),
    api_key_id: integer('api_key_id')
      .notNull()
      .references(() => api_keys.id),
    threshold: varchar('threshold', { length: 10 }).notNull(), // '51', '66', '80'
    request_count: integer('request_count').notNull().default(0),
    last_used: timestamp('last_used').defaultNow(),
    created_at: timestamp('created_at').defaultNow()
  });

