import { captureException } from '@sentry/node';
import { nextTick } from "process";
import { purgeRedisClientCredentials, sleep } from "./util.js";
import { redis } from "./redis.js";
import {
  countAccessTokens,
  deauthorizeServer,
  getAccessTokens,
  getServerConfig,
  updateServerAccountID,
  wipeAccountAccessToken,
} from "./queries.js";
import {
  RATE_LIMIT,
  SERVER_PUSH_FUZZ,
  SERVER_PUSH_LIMIT,
  VERSION,
} from "./constants.js";
import env from "./env.js";

export const USER_AGENT = `FediCheck/${VERSION} +${env.PUBLIC_URL}`;

/**
 *
 * @param {number} rateLimitReset
 * @returns {number}
 */
function getMillisToSleep(rateLimitReset) {
  return rateLimitReset - Date.now();
}

/**
 *
 * @param {string} server
 * @param {string|URL} location
 * @param {RequestInit} options
 * @param {string|null} account
 * @returns {Promise<Response>}
 */
export async function serverFetch(server, location, options, account = null) {
  const ratelimiter = new RedisRateLimiter(server);
  return await fetchAndRetryIfNecessary(
    async () =>
      ratelimiter.acquireToken(async () => {
        return await fetch(location, options);
      }),
    server,
    account,
  );
}

function isRateLimitExpired(ratelimits) {
  const limit_date = ratelimits.reset;
  const now = Date.now();

  return limit_date < now;
}

export async function getActualServerPushLimit(server) {
  const ratelimits = await getRateLimitsForServer(server);
  const actual_limit = ratelimits.remaining;

  // compare dates, if the reset date is in the past
  // we can use the full limit
  if (isRateLimitExpired(ratelimits)) {
    await deleteRatelimitInRedis(server);
    return SERVER_PUSH_LIMIT;
  }

  // if we're below our safe zone of server requests, abort immediately
  if (actual_limit <= SERVER_PUSH_FUZZ) {
    return 0;
  }
  // we use the actual limit when it's less than the push limit+reserve
  // in other words, if our reserve (fuzz) is 50 requests, and our push limit is 200
  // if we have 234 requests remaining, that's less than the total of 250 so using
  // 200 requests would bring us below our reserve of 50 requests.
  // to ensure we never run past the reserve, we should only rely on the hardcoded
  // server push limit when we have an abundant amount of requests remainining
  const using_actual_limit =
    SERVER_PUSH_LIMIT + SERVER_PUSH_FUZZ > actual_limit;
  const limit = using_actual_limit
    ? actual_limit - SERVER_PUSH_FUZZ
    : SERVER_PUSH_LIMIT;

  // if the resulting limit is larger tha the SERVER_PUSH_LIMIT, use the configured SERVER_PUSH_LIMIT
  return limit > SERVER_PUSH_LIMIT ? SERVER_PUSH_LIMIT : limit;
}

export async function getRateLimitsForServer(server) {
  return await getRatelimitFromRedis(server);
}

async function deleteRatelimitInRedis(server) {
  return await redis.del(`ratelimit:${server}`);
}

async function setRatelimitInRedis(server, responseRateLimits, millisToSleep) {
  return await redis.hset(`ratelimit:${server}`, {
    remaining: responseRateLimits.remaining,
    limit: responseRateLimits.limit,
    reset: responseRateLimits.reset,
    millis: millisToSleep,
  });
}

async function getRatelimitFromRedis(server) {
  const data = await redis.hgetall(`ratelimit:${server}`);

  if (!data || !data.limit) {
    return {
      limit: RATE_LIMIT,
      remaining: RATE_LIMIT,
      reset: Date.now(),
      millis: 0,
    };
  }

  // We could be defensive here over bad data stored in Redis, but we'd probably
  // notice that issue pretty quickly
  return {
    limit: parseInt(data.limit, 10),
    remaining: parseInt(data.remaining, 10),
    reset: parseInt(data.reset, 10),
    millis: parseInt(data.millis, 10),
  };
}

function parseRateLimitHeaders(response) {
  const rateLimitReset = response.headers.get("x-ratelimit-reset");
  if (rateLimitReset) {
    const limit = parseInt(response.headers.get("x-ratelimit-limit"), 10);
    const remaining = parseInt(response.headers.get("x-ratelimit-remaining"));
    const resetDate = Date.parse(rateLimitReset);

    if (isNaN(limit) || isNaN(remaining) || isNaN(resetDate)) {
      return {
        limit: RATE_LIMIT,
        remaining: RATE_LIMIT,
        reset: Date.now(),
      };
    } else {
      return {
        limit,
        remaining,
        reset: resetDate,
      };
    }
  } else {
    return null;
  }
}

/**
 * @callback fetchCallback
 * @returns {Promise<Response>}
 */

/**
 *
 * @param {fetchCallback} callAPIFn
 * @param {string} server
 * @param {Record<string, any>|null} [account = null]
 * @returns {Promise<Response>}
 */
export async function fetchAndRetryIfNecessary(
  callAPIFn,
  server,
  account = null,
) {
  try {
    const response = await callAPIFn();

    if (response) {
      // "The Access Token Has Been Revoked"
      if (response.status === 401) {
        if (account) {
          const config = await getServerConfig(server);
          await wipeAccountAccessToken(config.server_id, account.account_id);

           // if the token being used for this server is the same as the account we're using
        // we need to rotate the token if possible
          if (config.api_account_id === account.account_id) {
            const tokens = await getAccessTokens(server, true);

            if (tokens.length > 0) {
              await updateServerAccountID(config.server_id, tokens[0].account_id);
            } else {
              console.error(
                `Access token was externally revoked for ${server}. No available tokens for refresh. Purging all login and credential data for this server.`,
              );
              await deauthorizeServer(server);
              await purgeRedisClientCredentials(server);
            }
          }

          return response;
        }

        const total_tokens = await countAccessTokens(server);

        if (total_tokens === 0) {
          console.error(
            `Access token was externally revoked for ${server}. Purging all login and credential data for this server.`,
          );
          await deauthorizeServer(server);
          // kill all redis credentials, too
          await purgeRedisClientCredentials(server);
        }

        return response;
      }

      // because everything is asynchronous, a long-running request can
      // inadvertently set the remaining requests *higher*, moving from 100
      // back up to 111 available requests just because a certain request
      // took longer
      // so we both check the existing rate limits and timeouts to see if
      // the timeout has expired or not
      const limits = await getRatelimitFromRedis(server);
      const responseRateLimits = parseRateLimitHeaders(response);
      if (responseRateLimits) {
        const millisToSleep = getMillisToSleep(responseRateLimits.reset);
        const limits = await getRatelimitFromRedis(server);

        // only set a new remaining value in redis if the rate limit in redis has expired
      // or if the remaining value pulled from the header is actually lower
      // than what's stored in redis at this time.
      // this protects against out-of-sequence requests in large batches
      // setting the remaining requests *higher* within the same rate limit window
        if (
          isRateLimitExpired(limits) ||
          responseRateLimits.remaining < limits.remaining
        ) {
          await setRatelimitInRedis(server, responseRateLimits, millisToSleep);
        }

        if (response.status === 429) {
          console.debug(
            `429 rate limit hit, sleeping for ${millisToSleep / 1000} seconds...`,
          );

          await sleep(millisToSleep);
          return await fetchAndRetryIfNecessary(callAPIFn, server, account);
        }
      }
    }

    return response;
  } catch (err) {
    // Log the error to Sentry
    captureException(err);
    console.error(`Error during fetch operation for server ${server}:`, err);
    throw err; // Re-throw the error to ensure it propagates
  }
}

export class RedisRateLimiter {
  constructor(server) {
    this.server = server;
  }

  async acquireToken(fn) {
    const { remaining, millis } = await getRatelimitFromRedis(this.server);

    // this almost represents a failure state
    // as the application has been written
    if (remaining === 0) {
      await sleep(millis);
      return this.acquireToken(fn);
    }

    await new Promise((resolve) => nextTick(resolve));
    return fn();
  }
}
