import { BaseJob } from "../base_job.js";
import { db } from "../db.js";
import { domain_audits, domains, domain_audits_changelog } from "../db/schema.js";
import { recordDomainAuditChange } from "../queries.js";
import { eq, desc } from "drizzle-orm";
import { redis } from "../redis.js";

class MonitorDomainChanges extends BaseJob {
  thresholds = [80, 66, 51];

  checkThresholdsCrossed(oldPercent, newPercent) {
    for (const threshold of this.thresholds) {
      if ((oldPercent < threshold && newPercent >= threshold) || 
          (oldPercent >= threshold && newPercent < threshold)) {
        this.logger.info(`Threshold ${threshold}% crossed: ${oldPercent} -> ${newPercent}`);
        return true;
      }
    }
    return false;
  }

  async handle() {
    try {
      this.logger.info("Starting domain audit monitoring job");

      await this.acquireLock(300000);
      if (!this.hasLock) {
        this.logger.warn("Could not acquire lock - another instance may be running");
        return;
      }

      const currentAudits = await db
        .select({
          domain_id: domain_audits.domain_id,
          domain: domains.domain,
          percent: domain_audits.percent,
          authoritative: domain_audits.authoritative
        })
        .from(domain_audits)
        .innerJoin(domains, eq(domains.id, domain_audits.domain_id))
        .where(eq(domains.active, true));

      this.logger.info(`Processing ${currentAudits.length} domain audits`);

      for (const audit of currentAudits) {
        if (!this.isRunning) break;

        // Check if domain exists in changelog
        const previousAudit = await db
          .select({
            previous_percent: domain_audits_changelog.previous_percent,
            new_percent: domain_audits_changelog.new_percent,
          })
          .from(domain_audits_changelog)
          .where(eq(domain_audits_changelog.domain_id, audit.domain_id))
          .orderBy(desc(domain_audits_changelog.datetime))
          .limit(1);

        // If domain doesn't exist in changelog, record it as new
        if (previousAudit.length === 0) {
          this.logger.info({
            domain: audit.domain,
            percent: audit.percent,
            authoritative: audit.authoritative
          }, "New domain detected, recording initial entry");

          await recordDomainAuditChange(
            audit.domain_id,
            audit.domain,
            audit.percent,
            audit.percent, // we don't have a previous percent for a new domain so we use the current percent to mark as new
            audit.authoritative
          );
          continue;
        }

        // For existing domains, check for threshold crossings
        const previousPercent = previousAudit[0].new_percent;
        
        if (previousPercent !== audit.percent && 
            this.checkThresholdsCrossed(previousPercent, audit.percent)) {
          
          this.logger.info({
            domain: audit.domain,
            old_percent: previousPercent,
            new_percent: audit.percent,
            authoritative: audit.authoritative
          }, "Threshold crossed, recording change");

          await recordDomainAuditChange(
            audit.domain_id,
            audit.domain,
            previousPercent,
            audit.percent,
            audit.authoritative
          );
        }
      }

      this.logger.info("Domain audit monitoring completed successfully");

    } catch (error) {
      this.logger.error({
        error: error.toString(),
        stack: error.stack
      }, "Failed to process domain audits");
      throw error;
    } finally {
      if (this.hasLock) {
        await redis.del(this.lockName);
      }
    }
  }
}

// Create and run the job
const job = new MonitorDomainChanges();
await job.run();