import { BaseJob } from "../base_job.js";
import {
  pushBlocklist,
  verifyCredentials,
  wipeAllBlockData,
} from "../blocklist.js";
import { getActualServerPushLimit } from "../fetcher.js";
import { decryptString } from "../crypter.js";
import { ServerRepository } from "../repositories/ServerRepository.js";
import { AccountRepository } from "../repositories/AccountRepository.js";
import * as Sentry from "@sentry/node";

class PushBlocksJob extends BaseJob {
  async handle() {
    // FIXME: Move to delayed job, as soon as the block wipe is requested we should enqueue this work:
    // Priority one: block wipes. servers that need block wipes will be
    // omitted from the 'pushEnabledServers' automatically
    const block_wipe_servers = await ServerRepository.findAllByWipeRequested();
    const wipe_threads = block_wipe_servers.map(async (server) => {
      try {
        const push_limit = await getActualServerPushLimit(server.server);
        if (push_limit > 0) {
          this.logger.info(
            `Wiping all blocks for ${server.server} using a push limit of ${push_limit}`,
          );
          await wipeAllBlockData(server.server, push_limit);
        }
      } catch (err) {
        this.logger.error(
          { error: err.toString(), stack: err.stack, server },
          `Error wiping blocks from server ${server.server}`,
        );
        Sentry.captureException(err);
      }
    });

    await Promise.all(wipe_threads);

    // get all servers that have enabled push
    const push_servers = await ServerRepository.findAllPushEnabled();

    // each server can run its own thread with its own server's ratelimit
    const server_threads = push_servers.map(async (server) => {
      const serverLogger = this.logger.child({
        server: {
          id: server.server_id,
          server: server.server,
        },
      });
      try {
        // We must find the push account using the combination of api_account_id
        // and server_id, as Account's don't currently have a unique ID
        const push_account = await AccountRepository.findById(
          server.api_account_id,
          server.server_id,
        );

        if (!push_account || !push_account.access_token) {
          serverLogger.warn(
            {
              push_account_id: push_account?.account_id,
              has_access_token: !!push_account?.access_token,
            },
            "Server push credentials could not be found, missing push account or access token",
          );

          // We don't have an account to push with, therefore disable pushes:
          await ServerRepository.updatePushStatus(server.server_id, false);
          return;
        }

        const verified = await verifyCredentials(
          server.server,
          decryptString(push_account.access_token),
          false,
          serverLogger,
        );

        if (!verified) {
          serverLogger.error(
            { verified },
            "Server push credentials could not be verified",
          );
          // We don't have an account to push with because the credentials
          // couldn't be verified, therefore disable pushes:
          await ServerRepository.updatePushStatus(server.server_id, false);
          return;
        }

        const push_limit = await getActualServerPushLimit(server.server);
        serverLogger.info({ push_limit }, `Push limits for ${server.server}`);

        if (push_limit > 0) {
          await pushBlocklist(server.server, push_limit, serverLogger);
        }
      } catch (err) {
        // TODO: Implement circuit breaker to disable push if failed_pushes > x
        serverLogger.error(
          { error: err.toString(), stack: err.stack },
          `Error pushing blocks to server: ${server.server}`,
        );
        Sentry.captureException(err);
      }
    });

    const results = await Promise.allSettled(server_threads);
    results.forEach((result) => {
      if (result.status === "rejected") {
        Sentry.captureException(result.reason);
      }
    });
  }
}

new PushBlocksJob().run();
