import { 
  getApiKey, 
  logApiRequest, 
  hasPermission, 
  validateApiKey
} from '../queries.js';
import { getClientIP } from '../util.js';

export const apiKeyAuth = async (req, res, next) => {
  try {
    const apiKey = req.headers['x-api-key'] || req.query.api_key;
    
    if (!apiKey) {
      return res.status(401).json({
        success: false,
        error: 'API key is required'
      });
    }

    const keyDetails = await validateApiKey(apiKey);
    
    if (!keyDetails) {
      return res.status(401).json({
        success: false,
        error: 'Invalid or inactive API key'
      });
    }

    if (req.path === '/domains/by-threshold') {
      const { threshold } = req.query;
      if (threshold) {
        const requiredPermission = `threshold_${threshold}`;
        const hasAccess = await hasPermission(keyDetails.id, requiredPermission);
        
        if (!hasAccess) {
          return res.status(403).json({
            success: false,
            error: `API key does not have permission for ${threshold}% threshold`
          });
        }
      }
    }

    await logApiRequest(
      keyDetails.id,
      req.path,
      req.method,
      getClientIP(req),
      req.headers['user-agent'],
      req.query
    );
    
    req.apiKeyId = keyDetails.id;
    next();
  } catch (error) {
    console.error('API key auth error:', error);
    res.status(500).json({
      success: false,
      error: 'Authentication error'
    });
  }
};

export async function adminApiKeyAuth(req, res, next) {
  const apiKey = req.headers['x-admin-key'] || req.query.admin_key;
  
  if (!apiKey) {
    return res.status(401).json({
      success: false,
      error: 'Admin API key is required'
    });
  }

  try {
    const keyDetails = await getApiKey(apiKey);
    
    if (!keyDetails || !keyDetails.is_admin) {
      return res.status(401).json({
        success: false,
        error: 'Invalid admin API key'
      });
    }

    req.apiKeyId = keyDetails.id;
    
    await logApiRequest(
      keyDetails.id,
      req.path,
      req.method,
      req.headers['do-connecting-ip'],
      req.get('user-agent')
    );

    next();
  } catch (error) {
    console.error('Admin API key authentication error:', error);
    res.status(500).json({
      success: false,
      error: 'Internal server error'
    });
  }
} 