import { errorMiddleware } from "./errorHandler.js";
import { flash } from "@prairielearn/flash";
import { checkServerPermissions } from "../util.js";
import * as constants from "../constants.js";
import * as queries from "../queries.js";

const resetSession = (req, res, next) => {
  req.session.regenerate((err) => {
    if (err) {
      errorMiddleware(err, req, res, () => {
        flash("error", "An unexpected error occurred while regenerating the session");
        res.redirect("/login");
      });
      return;
    }

    flash("error", "Please login to access that page");
    res.redirect("/login");
  });
};

export async function authenticationMiddleware(req, res, next) {
  // @ts-ignore
  const account_id = req.session.account_id;
  // @ts-ignore
  const server_id = req.session.server_id;

  if (!account_id || !server_id) {
    // Wipe the session and redirect to login: if account_id or server_id is not set
    resetSession(req, res, next);
    return;
  }

  Promise.all([
    queries.getServerConfigBy("id", server_id),
    queries.getAccount(server_id, account_id),
  ])
    .then((results) => {
      const [server, account] = results;
      if (!server || !account || !checkServerPermissions(server?.server)) {
        // Wipe the session and redirect to login: if server or account is not found or if the server is not authorized
        resetSession(req, res, next);
        return;
      }

      const accountHandle = `${account.username}@${server.server}`;

      // add all pertinent info to the template locals
      res.locals.server_id = server.server_id;
      res.locals.server = server.server;

      res.locals.list_type = server.list_type;
      res.locals.push_enabled = server.enabled;
      res.locals.api_account_id = server.api_account_id;

      res.locals.block_wipe_in_progress = server.block_wipe_in_progress;

      res.locals.account = account;
      res.locals.handle = accountHandle;

      // check if this account has superadmin privileges
      if (constants.ADMIN_ACCOUNTS.includes(accountHandle)) {
        res.locals.is_admin = true;
      }

      return next();
    })
    .catch((err) => {
      next(err);
      req.log.error(
        { error: err.toString(), stack: err.stack },
        "Error validating session",
      );
      resetSession(req, res, next);
    });
}
