import { captureException, setUser } from '@sentry/node';
import { logErrorToDatabase } from "../utils/errorLogger.js";

/**
 * Middleware to handle errors in the application.
 * Logs the error, captures it with Sentry, and sends an appropriate response.
 *
 * @param {Error} err - The error object.
 * @param {Object} req - The Express request object.
 * @param {Object} res - The Express response object.
 * @param {Function} next - The next middleware function.
 *
 * Usage:
 * The `errorMiddleware` is used to handle errors that occur during request processing.
 * It provides a centralized way to log errors, capture them with Sentry for monitoring,
 * and send a response to the client. This middleware is particularly useful for:
 * 
 * - Logging detailed error information using the request's logger.
 * - Capturing exceptions with Sentry for alerting and monitoring.
 * - Sending a consistent error response to the client.
 * 
 * By default, calling `next(err)` in your route handlers will trigger this middleware,
 * allowing it to handle the error. However, if you need to perform additional actions
 * after an error occurs, you can call `errorMiddleware` directly and provide a callback
 * function to execute custom logic, such as flashing messages or redirecting the user.
 *
 * Example:
 * In a route handler, you might use `errorMiddleware` like this:
 *
 * try {
 *   // Some operation that might throw an error
 * } catch (err) {
 *   errorMiddleware(err, req, res, () => {
 *     flash("error", "An unexpected error occurred");
 *     res.redirect("back");
 *   });
 *   return;
 * }
 *
 * This approach ensures that errors are handled consistently and that the client
 * receives appropriate feedback, such as flashing an error message and redirecting
 * to a specific page.
 */
export function errorMiddleware(err, req, res, next) {
  // Log the error using the request's logger, if available
  if (req.log) {
    req.log.error({ err }, err.toString());
  }

  // Set user context for Sentry
  if (req.session && req.session.account_id) {
    setUser({
      id: req.session.account_id,
      server_id: req.session.server_id,
    });
  }

  // Capture the exception with Sentry for monitoring and alerting
  captureException(err);

  logErrorToDatabase(err, {
    userId: req.session.account_id,
    serverId: req.session.server_id,
    path: req.path,
  });

  // If headers have already been sent, delegate to the default Express error handler
  if (res.headersSent) {
    next(err);
    return;
  }

  // Determine the status code to send in the response
  const hasStatusCode = Object.hasOwnProperty.call(err, "status");
  const statusCode = hasStatusCode ? err.status : 500;

  // Determine the error message to send in the response
  const errorMessage = hasStatusCode
    ? err.toString()
    : "An unexpected error occurred";

  // Set the response status code
  res.status(statusCode);

  // Use a callback to handle custom redirection logic if provided
  if (typeof next === 'function') {
    next();
  } else {
    // Send the error message as the response
    res.send(errorMessage);
  }
}

// function determineSeverity(err) {
//   if (err.message.includes("fatal")) {
//     return "critical";
//   } else if (err.message.includes("high-risk")) {
//     return "high-risk";
//   }
//   return "normal";
// }
