import {
  AUTHORITATIVE_DOMAINS,
  AUTHORITATIVE_FILE_SOURCES,
  INCLUDED_FILE_SOURCES,
  LATENCY,
  MAX_TIER_PERCENT,
  SOURCE_DOMAINS,
  SOURCE_FILES,
} from "./constants.js";
import { db } from "./db.js";
import {
  blocks,
  domain_audits,
  domains,
  servers,
  server_config,
  unknown_domains,
  accounts,
  domain_audit_history,
  domain_audits_changelog,
  api_keys,
  api_request_logs,
  admin_sessions,
  api_key_permissions,
  api_threshold_requests,
} from "./db/schema.js";
import {
  and,
  or,
  eq,
  gte,
  asc,
  sql,
  isNotNull,
  isNull,
  inArray,
  count,
  desc,
  lt,
  like,
} from "drizzle-orm";
import { randomBytes } from 'crypto';

const MS_PER_MINUTE = 60000;

export function getBlockInactiveDate() {
  let compareDate = Date.now();
  let minutes = LATENCY + 2; // add 2 minutes of latency, just in case
  return new Date(compareDate - minutes * MS_PER_MINUTE);
}

export async function addUnknownDomain(domain, digest) {
  if (!digest) {
    return;
  }
  return await db
    .insert(unknown_domains)
    .values({
      domain,
      digest,
    })
    .onConflictDoNothing();
}

// This is called when a 401 is received from the server indication the
// access_token itself is invalid, and has been *externally* revoked and an
// attempt to rotate to a new token has failed. At that time we must immediately
// delete all stored configuration data along with access tokens. this leaves
// current blocks and accounts data stored in the database it's assumed the user
// can log back in for their server later and resubscribe to blocks to hard wipe
// their data later
export async function deauthorizeServer(server) {
  const server_id = await getServerID(server);
  await db.delete(server_config).where(eq(server_config.server_id, server_id));
  await wipeAccessTokens(server);
}

/**
 * Counts the number of accounts with valid access tokens for a given server.
 *
 * This function is used to determine if a server still has any valid OAuth connections.
 *
 * @param {string} server - The domain of the Mastodon server to check.
 * @returns {Promise<number>} The count of accounts with non-null access tokens.
 */
export async function countAccessTokens(server) {
  const server_id = await getServerID(server);
  const result = await db
    .select({ count: count() })
    .from(accounts)
    .where(
      and(eq(accounts.server_id, server_id), isNotNull(accounts.access_token)),
    );

  return result[0].count;
}

export async function wipeAccessTokens(server) {
  const server_id = await getServerID(server);
  await db
    .update(accounts)
    .set({
      access_token: null,
    })
    .where(eq(accounts.server_id, server_id));
}

export async function wipeAccountAccessToken(server_id, account_id) {
  await db
    .update(accounts)
    .set({
      access_token: null,
    })
    .where(
      and(
        eq(accounts.server_id, server_id),
        eq(accounts.account_id, account_id),
      ),
    );
}

export async function getAccessTokens(server, notNull) {
  const server_id = await getServerID(server);
  return await db
    .select({
      account_id: accounts.account_id,
      access_token: accounts.access_token,
    })
    .from(accounts)
    .where(
      and(
        eq(accounts.server_id, server_id),
        notNull ? isNotNull(accounts.access_token) : undefined,
      ),
    );
}

/**
 * Retrieves domain audits from the database, including domain details and audit data.
 *
 * @returns {Promise<Array<Object>>} A promise that resolves to an array of domain audit objects.
 */
export async function getDomainAudits() {
  const audits = await db
    .select({
      domain: domains.domain,
      count: domain_audits.count,
      percent: domain_audits.percent,
      public_comment: domain_audits.public_comment,
      severity: domain_audits.severity,
      count_suspend: domain_audits.count_suspend,
      count_silence: domain_audits.count_silence,
      count_noop: domain_audits.count_noop,
      active: sql`CASE WHEN ${domains.active} IS true THEN 'true' ELSE 'false' END`,
      authoritative: sql`CASE WHEN ${domain_audits.authoritative} IS true THEN 'true' ELSE 'false' END`,
      created_date: domain_audits.created_date,
      updated_at: domain_audits.updated_at,
    })
    .from(domain_audits)
    .innerJoin(domains, eq(domain_audits.domain_id, domains.id))
    .orderBy(asc(domains.domain));

  return audits;
}

export async function getBlockByID(server_id, block_id) {
  const response = await db
    .select()
    .from(blocks)
    .where(
      and(eq(blocks.block_ref, block_id), eq(blocks.server_id, server_id)),
    );

  return response[0];
}

// get blocks that meet the threshold, or authoritative blocks
export async function getBlocksByThreshold(threshold) {
  // query the domain_audits table to get the blocklist
  return await db
    .select({
      domain: domains.domain,
      severity: domain_audits.severity,
      public_comment: domain_audits.public_comment,
      approved_at: domains.approved_at,
    })
    .from(domain_audits)
    .innerJoin(domains, eq(domains.id, domain_audits.domain_id))
    .where(
      and(
        or(
          gte(domain_audits.percent, threshold),
          eq(domain_audits.authoritative, true),
        ),
        eq(domains.active, true),
        gte(domain_audits.updated_at, getBlockInactiveDate()),
      ),
    )
    .orderBy(asc(domains.domain));
}

export async function getLastServerRefreshDate(domain) {
  return await db
    .select({ last_server_refresh_date: servers.last_server_refresh_date })
    .from(servers)
    .where(eq(servers.server, domain));
}

export async function getLastAccountTokenVerifyDate(server_id, account_id) {
  return await db
    .select({ last_token_verify_date: accounts.last_token_verify_date })
    .from(accounts)
    .where(
      and(
        eq(accounts.server_id, server_id),
        eq(accounts.account_id, account_id),
      ),
    );
}

export async function getServerID(server) {
  // There's a refactor in progress to not call this as much, and there's been
  // some cases where this has received an object instead of the server name, so
  // if we have any more cases of that, we want the function to blow up and give
  // a stacktrace.
  if (typeof server !== "string") {
    throw new Error("getServerID received objected expected string");
  }

  const result = await db.query.servers.findFirst({
    where: eq(servers.server, server),
  });

  if (!result) {
    throw new Error(`getServerID failed to find server: ${server}`);
  }

  return result?.id;
}

const ServerConfigFields = {
  id: server_config.server_id,
  server_id: server_config.server_id,
  server: servers.server,
  client_id: server_config.client_id,
  client_secret: server_config.client_secret,
  access_token: accounts.access_token,
  enabled: server_config.enabled,
  list_type: server_config.list_type,
  block_wipe_in_progress: servers.block_wipe_in_progress,
  api_account_id: server_config.api_account_id,
};

export async function getServerConfigById(serverId) {
  // TODO: refactor to using relationships db.query.servers.findFirst
  const results = await db
    .select(ServerConfigFields)
    .from(servers)
    .innerJoin(server_config, eq(server_config.server_id, servers.id))
    .innerJoin(accounts, eq(server_config.api_account_id, accounts.account_id))
    .where(eq(servers.id, serverId))
    .limit(1);

  if (results.length !== 1) {
    throw new Error(
      `Failed to retrieve ServerConfig by server ID: ${serverId}`,
    );
  }

  return results[0];
}

export async function getServerConfigByServerName(serverName) {
  // TODO: refactor to using relationships db.query.servers.findFirst
  const results = await db
    .select(ServerConfigFields)
    .from(servers)
    .innerJoin(server_config, eq(server_config.server_id, servers.id))
    .innerJoin(accounts, eq(server_config.api_account_id, accounts.account_id))
    .where(eq(servers.server, serverName))
    .limit(1);

  if (results.length !== 1) {
    throw new Error(
      `Failed to retrieve ServerConfig by server name: ${serverName}`,
    );
  }

  return results[0];
}

/**
 * Note: this method is unsafe due to it potentially returning undefined:
 */
export async function getServerConfigBy(by, value) {
  if (by === "id") {
    return await getServerConfigById(value).catch(() => undefined);
  } else if (by === "server") {
    return await getServerConfigByServerName(value).catch(() => undefined);
  } else {
    throw new TypeError("Invalid `by` passed to `getServerConfigBy");
  }
}

/**
 * Note: this method is unsafe due to it potentially returning undefined:
 */

export async function getServerConfig(server) {
  return await getServerConfigByServerName(server).catch(() => undefined);
}

const account_select_fields = {
  server_id: accounts.server_id,
  account_id: accounts.account_id,
  username: accounts.username,
  display_name: accounts.display_name,
  access_token: accounts.access_token,
  created_date: accounts.created_date,
  last_login_date: accounts.last_login_date,
  last_token_verify_date: accounts.last_token_verify_date,
};
export async function getAccount(server_id, account_id) {
  const response = await db
    .select(account_select_fields)
    .from(accounts)
    .where(
      and(
        eq(accounts.server_id, server_id),
        eq(accounts.account_id, account_id),
      ),
    );

  return response[0];
}

export async function getTrustedSourceBlocks() {
  return await db
    .select()
    .from(servers)
    .innerJoin(blocks, eq(servers.id, blocks.server_id))
    .innerJoin(domains, eq(blocks.domain_id, domains.id))
    .where(and(eq(servers.source, true), eq(blocks.removed, false)));
}

export async function getTrustedFileSources() {
  return await db
    .select()
    .from(servers)
    .where(and(eq(servers.source, true), eq(servers.is_file, true)));
}

export async function getTotalTrustedSources() {
  const response = await db
    .select({ count: count() })
    .from(servers)
    .where(and(eq(servers.source, true), eq(servers.authoritative, false)));
  return response[0].count;
}

export async function initializeTrustedSources() {
  await db.transaction(async (tx) => {
    // clear source servers
    await tx
      .update(servers)
      .set({ source: false, authoritative: false })
      .where(and(eq(servers.source, true), eq(servers.is_file, false)));

    // verify on startup that we have a servers table entry for all of our
    // trusted sources
    for (const source of SOURCE_DOMAINS) {
      // manual upsert
      await tx
        .insert(servers)
        .values({
          server: source,
          url: `https://${source}/`,
          source: true,
          authoritative: AUTHORITATIVE_DOMAINS.includes(source),
        })
        .onConflictDoUpdate({
          target: servers.server,
          set: {
            source: true,
            url: `https://${source}/`,
            authoritative: AUTHORITATIVE_DOMAINS.includes(source),
          },
        });
    }
  });
}

export async function preloadServerFileSources() {
  await db.transaction(async (tx) => {
    // Disable existing source files:
    await tx
      .update(servers)
      .set({
        source: false,
        authoritative: false,
      })
      .where(and(eq(servers.source, true), eq(servers.is_file, true)));

    // Upsert any new source files:
    for (const source of Object.keys(SOURCE_FILES)) {
      if (!INCLUDED_FILE_SOURCES.includes(source)) {
        continue;
      }

      const authoritative = AUTHORITATIVE_FILE_SOURCES.includes(source);
      const sourceUrl = SOURCE_FILES[source].url;
      await tx
        .insert(servers)
        .values({
          server: source,
          url: sourceUrl,
          source: true,
          is_file: true,
          authoritative,
        })
        .onConflictDoUpdate({
          target: servers.server,
          set: {
            source: true,
            url: sourceUrl,
            authoritative,
          },
        });
    }
  });
}

/**
 * @typedef BlockRecord
 * @type {object}
 * @property {number} server_id
 * @property {number} domain_id
 * @property {string} severity
 * @property {boolean} reject_media
 * @property {boolean} reject_reports
 * @property {string} public_comment
 * @property {string} private_comment
 * @property {boolean} obfuscate
 * @property {string} block_ref
 * @property {boolean} is_managed
 * @property {boolean} is_admin
 */

/**
 * @param {BlockRecord} block
 * @returns {Promise<void>}
 */
export async function addEditBlock(block) {
  // update 'remove' flag in db
  await db
    .insert(blocks)
    .values(block)
    .onConflictDoUpdate({
      target: [blocks.server_id, blocks.domain_id],
      set: {
        severity: block.severity,
        public_comment: block.public_comment,
        private_comment: block.private_comment,
        block_ref: block.block_ref,
        is_managed: block.is_managed,
        is_admin: block.is_admin,
        removed: false,
        last_updated_date: sql`NOW()`,
      },
    });
}

export async function queryServerSafeList(server) {
  const server_id = await getServerID(server);
  const safelist = await db
    .select({
      block_ref: blocks.block_ref,
      domain_id: domains.id,
      server_id: blocks.server_id,
      domain: domains.domain,
      severity: blocks.severity,
      is_admin: blocks.is_admin,
      is_managed: blocks.is_managed,
      safelist: blocks.safelist,
      public_comment: blocks.public_comment,
      private_comment: blocks.private_comment,
      reject_media: blocks.reject_media,
      created_date: blocks.created_date,
      last_updated_date: blocks.last_updated_date,
      removed: blocks.removed,
    })
    .from(blocks)
    .innerJoin(domains, eq(domains.id, blocks.domain_id))
    .where(
      and(
        eq(blocks.server_id, server_id),
        or(eq(blocks.safelist, true), eq(blocks.is_managed, false)),
      ),
    )
    .orderBy(domains.domain);

  return safelist;
}

export async function queryDomainBlocks(server) {
  const server_id = await getServerID(server);
  const domain_blocks = await db
    .select({
      block_ref: blocks.block_ref,
      domain_id: domains.id,
      server_id: blocks.server_id,
      domain: domains.domain,
      severity: blocks.severity,
      is_admin: blocks.is_admin,
      is_managed: blocks.is_managed,
      safelist: blocks.safelist,
      public_comment: blocks.public_comment,
      private_comment: blocks.private_comment,
      reject_media: blocks.reject_media,
      created_date: blocks.created_date,
      last_updated_date: blocks.last_updated_date,
      removed: blocks.removed,
    })
    .from(blocks)
    .innerJoin(domains, eq(domains.id, blocks.domain_id))
    .where(eq(blocks.server_id, server_id))
    .orderBy(domains.domain);

  return domain_blocks;
}

export async function getDomainsMeta(domainIds) {
  if (domainIds.length === 0) {
    return [];
  }

  return await db
    .select({
      domain_id: domains.id,
      domain: domains.domain,
      severity: domain_audits.severity,
      public_comment: domain_audits.public_comment,
    })
    .from(domain_audits)
    .innerJoin(domains, eq(domains.id, domain_audits.domain_id))
    .where(inArray(domains.id, domainIds));
}

// when safelisting, we manage the block, when removing from safelist, we don't
// when 'unlocking' a domain, flip back to unmanaged. This prevents accidental removal.
// important safety tip, thanks, Egon.
export async function updateServerBlockSafelist(
  server_id,
  domain_id,
  safelist,
) {
  return await db
    .update(blocks)
    .set({
      safelist,
      is_managed: safelist,
    })
    .where(
      and(eq(blocks.server_id, server_id), eq(blocks.domain_id, domain_id)),
    );
}

// ensures that a block in our system is managed by us
export async function manageServerBlock(server_id, block_id, value) {
  // update 'remove' flag in db and reset the block_ref
  return await db
    .update(blocks)
    .set({
      is_managed: value,
    })
    .where(
      and(eq(blocks.server_id, server_id), eq(blocks.block_ref, block_id)),
    );
}

export async function manageAllServerBlocks(server_id, locked) {
  return await db
    .update(blocks)
    .set({
      is_managed: true,
      safelist: locked,
    })
    .where(eq(blocks.server_id, server_id));
}

export async function unlockAllDomains(server_id) {
  return await db
    .update(blocks)
    .set({
      is_managed: false,
      safelist: false,
    })
    .where(eq(blocks.server_id, server_id));
}

export async function removeBlock(server_id, block_id, safelist) {
  // update 'remove' flag in db and reset the block_ref
  return await db
    .update(blocks)
    .set({
      block_ref: null,
      removed: true,
      removed_date: new Date(),
      safelist: !!safelist, // any removed block also becomes allowlisted and won't be pushed until user removes from safelist
      updated: false,
    })
    .where(and(eq(blocks.server_id, server_id), eq(blocks.block_ref, block_id)))
    .returning();
}

export async function removeAllBlocks(server_id, hard_wipe) {
  if (hard_wipe) {
    await db.delete(blocks).where(eq(blocks.server_id, server_id));
  } else {
    // set all existing blocks for this server to removed = true and remove the block_ref
    await db
      .update(blocks)
      .set({
        block_ref: null,
        removed: true,
        removed_date: new Date(),
      })
      .where(eq(blocks.server_id, server_id));
  }
}

// return a list of domains, either approved or not approved
// note we only include domains that match domain_audits minimum threshold,
// meaning it is on at least one denylist
const domainslist_select = {
  id: domains.id,
  domain: domains.domain,
  approved_at: domains.approved_at,
  active: domains.active,
  severity: domain_audits.severity,
  public_comment: domain_audits.public_comment,
};

export async function getDomainsList(approved) {
  return await db
    .select(domainslist_select)
    .from(domains)
    .innerJoin(domain_audits, eq(domains.id, domain_audits.domain_id))
    .where(
      and(
        eq(domains.active, true),
        approved ? isNotNull(domains.approved_at) : isNull(domains.approved_at),
        or(
          gte(domain_audits.percent, MAX_TIER_PERCENT),
          eq(domain_audits.authoritative, true),
        ),
        gte(domain_audits.updated_at, getBlockInactiveDate()),
      ),
    )
    .orderBy(asc(domains.domain));
}

export async function getInactiveDomains() {
  return await db
    .select(domainslist_select)
    .from(domains)
    .innerJoin(domain_audits, eq(domains.id, domain_audits.domain_id))
    .where(eq(domains.active, false))
    .orderBy(asc(domains.domain));
}

export async function approveDomain(domain_id) {
  return await db
    .update(domains)
    .set({ approved_at: sql`NOW()` })
    .where(eq(domains.id, domain_id))
    .returning({ domain: domains.domain });
}

export async function unapproveDomain(domain_id) {
  return await db
    .update(domains)
    .set({ approved_at: null })
    .where(eq(domains.id, domain_id))
    .returning({ domain: domains.domain });
}

export async function approveAllPendingDomains() {
  const domainsList = await getDomainsList(false);
  const ids = domainsList.map((domain) => domain.id);

  if (ids.length === 0) {
    return;
  }

  return await db
    .update(domains)
    .set({ approved_at: sql`NOW()` })
    .where(inArray(domains.id, ids));
}

export async function unapproveAllDomains() {
  const domainsList = await getDomainsList(true);
  const ids = domainsList.map((domain) => domain.id);

  if (ids.length === 0) {
    return;
  }

  return await db
    .update(domains)
    .set({ approved_at: null })
    .where(inArray(domains.id, ids));
}

export async function deactivateAllPendingDomains() {
  const domainsList = await getDomainsList(false);
  const ids = domainsList.map((domain) => domain.id);

  if (ids.length === 0) {
    return;
  }

  return await db
    .update(domains)
    .set({ active: false })
    .where(inArray(domains.id, ids));
}

export async function activateDomain(domain_id) {
  return await db
    .update(domains)
    .set({ active: true })
    .where(eq(domains.id, domain_id))
    .returning({ domain: domains.domain });
}

export async function deactivateDomain(domain_id) {
  return await db
    .update(domains)
    .set({ active: false })
    .where(eq(domains.id, domain_id))
    .returning({ domain: domains.domain });
}

export async function getAccounts(serverId) {
  const query = db
    .select({
      server_id: accounts.server_id,
      server: servers.server,
      account_id: accounts.account_id,
      username: accounts.username,
      display_name: accounts.display_name,
      access_token: accounts.access_token,
      created_date: accounts.created_date,
      last_login_date: accounts.last_login_date,
      enabled: server_config.enabled,
      list_type: server_config.list_type,
      is_api_user: sql`CASE WHEN ${accounts.account_id} = ${server_config.api_account_id} THEN true ELSE false END`,
    })
    .from(accounts)
    .innerJoin(servers, eq(servers.id, accounts.server_id))
    .innerJoin(server_config, eq(server_config.server_id, servers.id));

  if (serverId) {
    query.where(
      and(eq(servers.id, serverId), isNotNull(accounts.access_token)),
    );
    return await query.orderBy(asc(accounts.username));
  } else {
    return await query.orderBy(
      asc(servers.server),
      asc(accounts.display_name),
      asc(accounts.username),
    );
  }
}

export async function updateClientCredentials(
  server_id,
  client_id,
  client_secret,
) {
  return await db
    .update(server_config)
    .set({
      client_id,
      client_secret,
    })
    .where(eq(server_config.server_id, server_id));
}

export async function updateServerAccountID(server_id, account_id) {
  return await db
    .update(server_config)
    .set({
      api_account_id: account_id,
    })
    .where(eq(server_config.server_id, server_id));
}

export async function updateLastAccountTokenVerifyDate(server_id, account_id) {
  return await db
    .update(accounts)
    .set({
      last_token_verify_date: new Date(),
    })
    .where(
      and(
        eq(accounts.server_id, server_id),
        eq(accounts.account_id, account_id),
      ),
    );
}

// we throw unknown domains into a table along with their digest so we can expose this to superadmins
// and they can take some guesses at the domain name. We can compare their guess with the digest and if correct,
// we add the domain to the domains table along with the digest, and remove it from unknown_domains.
export async function clearMatchedUnknownDomains() {
  const matched_domains = await db
    .select({ digest: unknown_domains.digest })
    .from(unknown_domains)
    .innerJoin(domains, eq(unknown_domains.digest, domains.digest));

  if (matched_domains.length > 0) {
    await db.delete(unknown_domains).where(
      inArray(
        unknown_domains.digest,
        matched_domains.map((domain) => domain.digest),
      ),
    );
  }
}

// get all unknown domains
export async function getUnknownDomains() {
  return await db
    .select()
    .from(unknown_domains)
    .orderBy(asc(unknown_domains.domain));
}

// get a single unknown domain by digest
export async function getUnknownDomain(digest) {
  const response = await db
    .select()
    .from(unknown_domains)
    .where(eq(unknown_domains.digest, digest));

  return response[0];
}

export async function getAccountById(accountId) {
  try {
    // Ensure accountId is treated as a string or number, not a UUID
    const account = await db
      .select()
      .from(accounts)
      .where(eq(accounts.account_id, accountId)); // Ensure this matches the actual data type

    console.log(`Fetched account for ID ${accountId}:`, account[0]); // Log the fetched account

    // Return the first account if it exists
    return account[0] || null;
  } catch (err) {
    console.error("Error fetching account by ID:", err);
    throw err;
  }
}

export async function getServerById(serverId) {
  try {
    const server = await db
      .select()
      .from(servers) // Ensure this matches your server table
      .where(eq(servers.id, serverId));

    console.log(`Fetched server for ID ${serverId}:`, server[0]); // Log the fetched server

    // Return the first server if it exists
    return server[0] || null;
  } catch (err) {
    console.error("Error fetching server by ID:", err);
    throw err;
  }
}

export const deleteAllErrorLogs = 'DELETE FROM error_logs'; // Adjust the table name as necessary

// export async function recordDomainAuditChange(
//   domain_id,
//   old_percent,
//   new_percent,
//   old_severity,
//   new_severity,
//   change_reason = 'automatic_update'
// ) {
//   try {
//     await db.insert(domain_audit_history).values({
//       domain_id,
//       old_percent,
//       new_percent,
//       old_severity,
//       new_severity,
//       change_reason,
//     });
//   } catch (error) {
//     console.error('Failed to record domain audit change:', {
//       error: error.toString(),
//       domain_id,
//       old_percent,
//       new_percent,
//       change_reason
//     });
//     throw error;
//   }
// }

export async function updateDomainAuditWithHistory(domain_id, updateData) {
  // First get the current values
  const currentAudit = await db
    .select({
      domain_id: domain_audits.domain_id,
      domain: domains.domain,
      percent: domain_audits.percent,
    })
    .from(domain_audits)
    .innerJoin(domains, eq(domains.id, domain_audits.domain_id))
    .where(eq(domain_audits.domain_id, domain_id))
    .limit(1);

  // If no current audit exists, this is a new domain
  const isNewDomain = currentAudit.length === 0;

  if (isNewDomain) {
    // Get domain info
    const domainInfo = await db
      .select()
      .from(domains)
      .where(eq(domains.id, domain_id))
      .limit(1);

    if (!domainInfo[0]) {
      throw new Error(`No domain found for domain_id: ${domain_id}`);
    }

    // Record the initial entry in changelog
    await recordDomainAuditChange(
      domain_id,
      domainInfo[0].domain,
      updateData.percent, // Use the same value for both previous and new
      updateData.percent  // This will mark it as 'new_domain'
    );
  }

  // Update the domain audit
  await db
    .update(domain_audits)
    .set({
      ...updateData,
      updated_at: new Date(),
    })
    .where(eq(domain_audits.domain_id, domain_id));

  // If not a new domain and percent changed, record the change
  if (!isNewDomain && currentAudit[0].percent !== updateData.percent) {
    await recordDomainAuditChange(
      domain_id,
      currentAudit[0].domain,
      currentAudit[0].percent,
      updateData.percent
    );
  }
}

export async function getRecentDomainChanges(page = 1, limit = 20, searchTerm = null) {
  // Calculate offset
  const offset = (page - 1) * limit;

  // Build the where clause
  let whereClause = sql`1=1`; // Default true condition
  if (searchTerm) {
    whereClause = sql`${domain_audits_changelog.domainname} ILIKE ${`%${searchTerm}%`}`;
  }

  // Get total count for pagination
  const [{ count }] = await db
    .select({ count: sql`count(*)::int` })
    .from(domain_audits_changelog)
    .where(whereClause);

  // Get paginated results with authoritative flag and active status
  const changes = await db
    .select({
      domainname: domain_audits_changelog.domainname,
      threshold_crossed: domain_audits_changelog.threshold_crossed,
      previous_percent: sql`ROUND(CAST(${domain_audits_changelog.previous_percent} AS NUMERIC), 1)`,
      new_percent: sql`ROUND(CAST(${domain_audits_changelog.new_percent} AS NUMERIC), 1)`,
      datetime: domain_audits_changelog.datetime,
      authoritative: domain_audits.authoritative,
      is_active: domains.active // Add active status from domains table
    })
    .from(domain_audits_changelog)
    .leftJoin(
      domain_audits,
      eq(domain_audits_changelog.domain_id, domain_audits.domain_id)
    )
    .leftJoin( // Add join with domains table
      domains,
      eq(domain_audits_changelog.domain_id, domains.id)
    )
    .where(whereClause)
    .orderBy(desc(domain_audits_changelog.datetime))
    .limit(limit)
    .offset(offset);

  return {
    changes,
    pagination: {
      total: count,
      page,
      pageSize: limit,
      totalPages: Math.ceil(count / limit)
    }
  };
}

export async function getDomainAuditHistory(domainname) {
  return await db
    .select({
      threshold_crossed: domain_audits_changelog.threshold_crossed,
      previous_percent: domain_audits_changelog.previous_percent,
      new_percent: domain_audits_changelog.new_percent,
      datetime: domain_audits_changelog.datetime,
    })
    .from(domain_audits_changelog)
    .where(eq(domain_audits_changelog.domainname, domainname))
    .orderBy(desc(domain_audits_changelog.datetime));
}

export async function recordDomainAuditChange(domain_id, domainname, previous_percent, new_percent, authoritative = false) {
  let threshold_crossed;

  if (previous_percent === new_percent) {
    threshold_crossed = 'new_domain';
  } else {
    threshold_crossed = previous_percent < new_percent ? 'increased' : 'decreased';
  }

  await db.insert(domain_audits_changelog).values({
    domain_id,
    domainname,
    threshold_crossed,
    previous_percent,
    new_percent,
    datetime: new Date(),
    authoritative
  });
}

// Add this test function
export async function createTestDomainAudit(domain = 'test.example.com', initialPercent = 45) {
  try {
    // First create a domain if it doesn't exist
    const [domainRecord] = await db
      .insert(domains)
      .values({
        domain,
        digest: 'test-digest',
        active: true
      })
      .onConflictDoUpdate({
        target: domains.domain,
        set: { active: true }
      })
      .returning();

    // Then create or update the domain audit
    const [auditRecord] = await db
      .insert(domain_audits)
      .values({
        domain_id: domainRecord.id,
        percent: initialPercent,
        severity: 'suspend',
        count: 1
      })
      .onConflictDoUpdate({
        target: domain_audits.domain_id,
        set: { percent: initialPercent }
      })
      .returning();

    return { domainRecord, auditRecord };
  } catch (error) {
    console.error('Error creating test domain audit:', error);
    throw error;
  }
}

// Add this test function to update the percent
export async function updateTestDomainAudit(domain = 'test.example.com', newPercent) {
  try {
    const domainRecord = await db
      .select()
      .from(domains)
      .where(eq(domains.domain, domain))
      .limit(1);

    if (!domainRecord[0]) {
      throw new Error('Test domain not found');
    }

    await db
      .update(domain_audits)
      .set({
        percent: newPercent,
        updated_at: new Date()
      })
      .where(eq(domain_audits.domain_id, domainRecord[0].id));

    return { domain: domainRecord[0], newPercent };
  } catch (error) {
    console.error('Error updating test domain audit:', error);
    throw error;
  }
}

export async function searchDomains(searchTerm) {
  try {
    // Search across domains table with optional join to domain_audits
    const results = await db
      .select({
        id: domains.id,
        domain: domains.domain,
        active: domains.active,
        approved_at: domains.approved_at,
        severity: domain_audits.severity,
        percent: domain_audits.percent,
        public_comment: domain_audits.public_comment,
        authoritative: domain_audits.authoritative
      })
      .from(domains)
      .leftJoin(domain_audits, eq(domains.id, domain_audits.domain_id))
      .where(sql`${domains.domain} ILIKE ${`%${searchTerm}%`}`)
      .orderBy(asc(domains.domain))
      .limit(50);

    return results;
  } catch (error) {
    console.error('Error searching domains:', error);
    throw error;
  }
}

/**
 * Gets domains that are at or above specified percentage threshold
 * @param {number} minPercent - Minimum percentage threshold (inclusive)
 * @returns {Promise<Array>} Array of domain audit records at or above the threshold
 */
export async function getDomainsByMinimumThreshold(minPercent) {
  console.log('Querying domains with minimum threshold:', { minPercent });

  const query = db
    .select({
      domain: sql`DISTINCT ${domains.domain}`,
      severity: domain_audits.severity,
    })
    .from(domain_audits)
    .innerJoin(domains, eq(domains.id, domain_audits.domain_id))
    .where(
      and(
        or(
          gte(domain_audits.percent, minPercent),
          eq(domain_audits.authoritative, true)
        ),
        eq(domains.active, true)
      )
    )
    .orderBy(asc(domains.domain));

  const results = await query;
  console.log('Query results:', results);

  return results;
}

/**
 * Gets domains with percentage at least 51%
 */
export async function getDomainsAtLeast51Percent() {
  return await getDomainsByMinimumThreshold(51);
}

/**
 * Gets domains with percentage at least 66%
 */
export async function getDomainsAtLeast66Percent() {
  return await getDomainsByMinimumThreshold(66);
}

/**
 * Gets domains with percentage at least 80%
 */
export async function getDomainsAtLeast80Percent() {
  return await getDomainsByMinimumThreshold(80);
}

// Generate a new API key
function generateApiKey() {
  return randomBytes(32).toString('hex');
}

// Create a new API key
export async function createApiKey(description) {
  const key = generateApiKey();
  const [apiKey] = await db
    .insert(api_keys)
    .values({
      key,
      description,
      is_active: true
    })
    .returning();
  return apiKey;
}

// Get API key by key string
export async function getApiKey(key) {
  return await db
    .select()
    .from(api_keys)
    .where(and(
      eq(api_keys.key, key),
      eq(api_keys.is_active, true)
    ))
    .limit(1)
    .then(rows => rows[0]);
}


// TODO: Use Redis
// Update the logApiRequest function
export async function logApiRequest(apiKeyId, endpoint, method, ipAddress, userAgent, queryParams) {
  // Construct endpoint with query parameters for threshold requests
  let fullEndpoint = endpoint;
  if (endpoint === '/domains/by-threshold' && queryParams?.threshold) {
    fullEndpoint = `${endpoint}?threshold=${queryParams.threshold}`;
  }

  // Log the request
  await db.insert(api_request_logs).values({
    api_key_id: apiKeyId,
    endpoint: fullEndpoint,
    method,
    ip_address: ipAddress,
    user_agent: userAgent
  });

  // Update last_used_at timestamp
  await db
    .update(api_keys)
    .set({ last_used_at: new Date() })
    .where(eq(api_keys.id, apiKeyId));
}

// Update getApiUsageStats to include permissions
export async function getApiUsageStats() {
  const apiKeys = await db
    .select({
      key_id: api_keys.id,
      description: api_keys.description,
      key: api_keys.key,
      created_at: api_keys.created_at,
      last_used_at: api_keys.last_used_at,
      total_requests: sql`count(${api_request_logs.id})::int`,
      is_active: api_keys.is_active,
      is_admin: api_keys.is_admin
    })
    .from(api_keys)
    .leftJoin(api_request_logs, eq(api_keys.id, api_request_logs.api_key_id))
    .groupBy(api_keys.id)
    .orderBy(desc(api_keys.created_at));

  // Get permissions and threshold stats for each API key
  const apiKeysWithDetails = await Promise.all(
    apiKeys.map(async (key) => {
      const [permissions, thresholdStats] = await Promise.all([
        getApiKeyPermissions(key.key_id),
        getApiKeyThresholdStats(key.key_id)
      ]);

      return {
        ...key,
        permissions: permissions.map(p => p.permission),
        threshold_stats: thresholdStats.reduce((acc, stat) => {
          acc[`threshold_${stat.threshold}`] = {
            requests: stat.request_count,
            last_used: stat.last_used
          };
          return acc;
        }, {})
      };
    })
  );

  return apiKeysWithDetails;
}

// Add this function to verify admin secret key
export async function verifyAdminSecret(secretKey) {
  // Compare with environment variable or a stored secure value
  return secretKey === process.env.ADMIN_API_SECRET;
}

export async function createAdminSession(adminId) {
  const sessionToken = randomBytes(32).toString('hex');
  const sessionId = randomBytes(16).toString('hex');

  await db.insert(admin_sessions).values({
    id: sessionId,
    admin_id: adminId,
    token: sessionToken,
    created_at: new Date(),
    expires_at: new Date(Date.now() + 24 * 60 * 60 * 1000) // 24 hours
  });

  return {
    sessionId,
    sessionToken
  };
}

export async function verifyAdminSession(sessionId, token) {
  const session = await db
    .select()
    .from(admin_sessions)
    .where(
      and(
        eq(admin_sessions.id, sessionId),
        eq(admin_sessions.token, token),
        gte(admin_sessions.expires_at, new Date())
      )
    )
    .limit(1);

  return session.length > 0;
}

export async function verifyAdminApiKey(key) {
  const adminKey = await db
    .select()
    .from(api_keys)
    .where(
      and(
        eq(api_keys.key, key),
        eq(api_keys.is_active, true),
        eq(api_keys.is_admin, true)
      )
    )
    .limit(1);

  return adminKey.length > 0;
}

// Get overall API usage statistics
export async function getApiRequestStats() {
  const result = await db
    .select({
      total_requests: sql`count(${api_request_logs.id})::int`,
      unique_keys: sql`count(distinct ${api_request_logs.api_key_id})::int`,
      active_keys: sql`count(distinct case when ${api_keys.is_active} then ${api_keys.id} end)::int`,
      most_used_endpoint: sql`mode() within group (order by ${api_request_logs.endpoint})`
    })
    .from(api_request_logs)
    .leftJoin(api_keys, eq(api_keys.id, api_request_logs.api_key_id));

  return result[0];
}

// Get stats for a specific API key
export async function getApiKeyStats(keyId) {
  const result = await db
    .select({
      total_requests: sql`count(${api_request_logs.id})::int`,
      unique_endpoints: sql`count(distinct ${api_request_logs.endpoint})::int`,
      first_used: sql`min(${api_request_logs.timestamp})`,
      last_used: sql`max(${api_request_logs.timestamp})`,
      most_used_endpoint: sql`mode() within group (order by ${api_request_logs.endpoint})`
    })
    .from(api_request_logs)
    .where(eq(api_request_logs.api_key_id, keyId));

  return result[0];
}

// Get stats by endpoint
export async function getEndpointStats() {
  return await db
    .select({
      endpoint: api_request_logs.endpoint,
      total_requests: sql`count(${api_request_logs.id})::int`,
      unique_keys: sql`count(distinct ${api_request_logs.api_key_id})::int`,
      last_used: sql`max(${api_request_logs.timestamp})`
    })
    .from(api_request_logs)
    .groupBy(api_request_logs.endpoint)
    .orderBy(desc(sql`count(${api_request_logs.id})`));
}

// Get daily usage stats
export async function getDailyStats(days = 7) {
  return await db
    .select({
      date: sql`date_trunc('day', ${api_request_logs.timestamp})`,
      total_requests: sql`count(${api_request_logs.id})::int`,
      unique_keys: sql`count(distinct ${api_request_logs.api_key_id})::int`
    })
    .from(api_request_logs)
    .where(gte(api_request_logs.timestamp, sql`now() - interval '${days} days'`))
    .groupBy(sql`date_trunc('day', ${api_request_logs.timestamp})`)
    .orderBy(sql`date_trunc('day', ${api_request_logs.timestamp})`);
}

// Create API key with permissions
export async function createApiKeyWithPermissions(description, permissions = []) {
  const key = generateApiKey();

  // Create API key
  const [apiKey] = await db
    .insert(api_keys)
    .values({
      key,
      description,
      is_admin: false,
      is_active: true
    })
    .returning();

  // Add permissions
  if (permissions.length > 0) {
    await db.insert(api_key_permissions)
      .values(permissions.map(permission => ({
        api_key_id: apiKey.id,
        permission
      })));
  }

  return apiKey;
}

// Get API key permissions
export async function getApiKeyPermissions(keyId) {
  return await db
    .select({
      permission: api_key_permissions.permission
    })
    .from(api_key_permissions)
    .where(eq(api_key_permissions.api_key_id, keyId));
}

// Update API key permissions with better error handling
export async function updateApiKeyPermissions(keyId, permissions) {
  try {
    // First verify the API key exists
    const apiKey = await db
      .select()
      .from(api_keys)
      .where(eq(api_keys.id, keyId))
      .limit(1);

    if (!apiKey.length) {
      throw new Error(`API key with ID ${keyId} does not exist`);
    }

    await db.transaction(async (tx) => {
      // Remove existing permissions
      await tx
        .delete(api_key_permissions)
        .where(eq(api_key_permissions.api_key_id, keyId));

      // Add new permissions if any
      if (permissions && permissions.length > 0) {
        await tx
          .insert(api_key_permissions)
          .values(permissions.map(permission => ({
            api_key_id: keyId,
            permission
          })));
      }
    });

    return true;
  } catch (error) {
    if (error.code === '23503') { // Foreign key violation
      throw new Error(`Invalid API key ID: ${keyId}`);
    }
    throw error;
  }
}

// Check if API key has specific permission
export async function hasPermission(keyId, permission) {
  const result = await db
    .select()
    .from(api_key_permissions)
    .where(
      and(
        eq(api_key_permissions.api_key_id, keyId),
        eq(api_key_permissions.permission, permission)
      )
    )
    .limit(1);

  return result.length > 0;
}

// Add this new function to get request stats by method type
export async function getRequestStatsByType() {
  return await db
    .select({
      method: api_request_logs.method,
      total_requests: sql`count(${api_request_logs.id})::int`,
      unique_keys: sql`count(distinct ${api_request_logs.api_key_id})::int`,
      last_used: sql`max(${api_request_logs.timestamp})`,
      endpoints: sql`array_agg(distinct ${api_request_logs.endpoint})`,
      success_rate: sql`round((count(case when ${api_request_logs.method} = ${api_request_logs.method} then 1 end)::decimal / count(*) * 100), 2)`
    })
    .from(api_request_logs)
    .groupBy(api_request_logs.method)
    .orderBy(desc(sql`count(${api_request_logs.id})`));
}
// Update the getThresholdUsageStats function
export async function getThresholdUsageStats() {
  return await db
    .select({
      threshold: sql`CASE
        WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=51%' OR ${api_request_logs.endpoint} LIKE '%&threshold=51%' THEN '51%'
        WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=66%' OR ${api_request_logs.endpoint} LIKE '%&threshold=66%' THEN '66%'
        WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=80%' OR ${api_request_logs.endpoint} LIKE '%&threshold=80%' THEN '80%'
        ELSE 'other'
      END`,
      total_requests: sql`count(*)::int`,
      unique_keys: sql`count(distinct ${api_request_logs.api_key_id})::int`,
      last_used: sql`max(${api_request_logs.timestamp})`,
      percentage: sql`ROUND(CAST((count(*)::float / NULLIF((SELECT count(*) FROM ${api_request_logs} WHERE ${api_request_logs.endpoint} LIKE '/domains/by-threshold%'), 0) * 100) as numeric), 2)`
    })
    .from(api_request_logs)
    .where(like(api_request_logs.endpoint, '/domains/by-threshold%'))
    .groupBy(sql`CASE
      WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=51%' OR ${api_request_logs.endpoint} LIKE '%&threshold=51%' THEN '51%'
      WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=66%' OR ${api_request_logs.endpoint} LIKE '%&threshold=66%' THEN '66%'
      WHEN ${api_request_logs.endpoint} LIKE '/domains/by-threshold?threshold=80%' OR ${api_request_logs.endpoint} LIKE '%&threshold=80%' THEN '80%'
      ELSE 'other'
    END`)
    .orderBy(desc(sql`count(*)`));
}

// Add this function to toggle API key active status
export async function toggleApiKeyStatus(keyId) {
  try {
    // First verify the API key exists and get current status
    const apiKey = await db
      .select({
        id: api_keys.id,
        is_active: api_keys.is_active,
        is_admin: api_keys.is_admin
      })
      .from(api_keys)
      .where(eq(api_keys.id, keyId))
      .limit(1);

    if (!apiKey.length) {
      throw new Error(`API key with ID ${keyId} does not exist`);
    }

    // Don't allow deactivating admin keys
    if (apiKey[0].is_admin) {
      throw new Error('Cannot toggle admin API key status');
    }

    // Toggle the status
    const [updatedKey] = await db
      .update(api_keys)
      .set({
        is_active: !apiKey[0].is_active,
        last_used_at: new Date() // Update timestamp
      })
      .where(eq(api_keys.id, keyId))
      .returning();

    return updatedKey;
  } catch (error) {
    if (error.code === '23503') { // Foreign key violation
      throw new Error(`Invalid API key ID: ${keyId}`);
    }
    throw error;
  }
}

// TODO:Handle this with Redis
// Add this function to increment threshold request count
export async function incrementThresholdRequest(apiKeyId, threshold) {
  // First try to update existing record
  const result = await db
    .update(api_threshold_requests)
    .set({
      request_count: sql`${api_threshold_requests.request_count} + 1`,
      last_used: new Date()
    })
    .where(
      and(
        eq(api_threshold_requests.api_key_id, apiKeyId),
        eq(api_threshold_requests.threshold, threshold)
      )
    )
    .returning();

  // If no record was updated, create a new one
  if (result.length === 0) {
    await db.insert(api_threshold_requests).values({
      api_key_id: apiKeyId,
      threshold,
      request_count: 1,
      last_used: new Date()
    });
  }
}

// Add this function to get threshold usage for an API key
export async function getApiKeyThresholdStats(apiKeyId) {
  return await db
    .select({
      threshold: api_threshold_requests.threshold,
      request_count: api_threshold_requests.request_count,
      last_used: api_threshold_requests.last_used
    })
    .from(api_threshold_requests)
    .where(eq(api_threshold_requests.api_key_id, apiKeyId))
    .orderBy(desc(api_threshold_requests.request_count));
}

// Add this new function
export async function validateApiKey(key) {
  const keyData = await db
    .select()
    .from(api_keys)
    .where(and(
      eq(api_keys.key, key),
      eq(api_keys.is_active, true)
    ))
    .limit(1);

  return keyData.length > 0 ? keyData[0] : null;
}

// Add this new function to get detailed API key info
export async function getApiKeyDetails(keyId) {
  // Get basic API key info
  const apiKey = await db
    .select({
      id: api_keys.id,
      description: api_keys.description,
      key: api_keys.key,
      created_at: api_keys.created_at,
      last_used_at: api_keys.last_used_at,
      is_active: api_keys.is_active,
      is_admin: api_keys.is_admin
    })
    .from(api_keys)
    .where(eq(api_keys.id, keyId))
    .limit(1)
    .then(rows => rows[0]);

  if (!apiKey) {
    return null;
  }

  return apiKey;
}

export async function getApiKeyRequestHistory(keyId, limit = 100) {
  return await db
    .select({
      endpoint: api_request_logs.endpoint,
      method: api_request_logs.method,
      timestamp: api_request_logs.timestamp,
      ip_address: api_request_logs.ip_address,
      user_agent: api_request_logs.user_agent
    })
    .from(api_request_logs)
    .where(eq(api_request_logs.api_key_id, keyId))
    .orderBy(desc(api_request_logs.timestamp))
    .limit(limit);
}

export async function getApiKeyEndpointStats(keyId) {
  return await db
    .select({
      endpoint: api_request_logs.endpoint,
      total_requests: sql`count(*)::int`,
      last_used: sql`max(${api_request_logs.timestamp})`,
      unique_ips: sql`count(distinct ${api_request_logs.ip_address})::int`
    })
    .from(api_request_logs)
    .where(eq(api_request_logs.api_key_id, keyId))
    .groupBy(api_request_logs.endpoint)
    .orderBy(desc(sql`count(*)`));
}

export async function getApiKeyDailyStats(keyId, days = 30) {
  return await db
    .select({
      date: sql`date_trunc('day', ${api_request_logs.timestamp})`,
      total_requests: sql`count(*)::int`,
      unique_endpoints: sql`count(distinct ${api_request_logs.endpoint})::int`
    })
    .from(api_request_logs)
    .where(
      and(
        eq(api_request_logs.api_key_id, keyId),
        // Fix: Use raw SQL for the interval comparison
        sql`${api_request_logs.timestamp} >= now() - interval '30 days'`
      )
    )
    .groupBy(sql`date_trunc('day', ${api_request_logs.timestamp})`)
    .orderBy(sql`date_trunc('day', ${api_request_logs.timestamp})`);
}

export async function getApiKeyHourlyDistribution(keyId) {
  return await db
    .select({
      hour: sql`extract(hour from ${api_request_logs.timestamp})::int`,
      total_requests: sql`count(*)::int`
    })
    .from(api_request_logs)
    .where(eq(api_request_logs.api_key_id, keyId))
    .groupBy(sql`extract(hour from ${api_request_logs.timestamp})`)
    .orderBy(sql`extract(hour from ${api_request_logs.timestamp})`);
}

// Add this new function with the other API key related functions
export async function deleteApiKey(keyId) {
  try {
    // First verify the API key exists and is not an admin key
    const apiKey = await db
      .select()
      .from(api_keys)
      .where(eq(api_keys.id, keyId))
      .limit(1)
      .then(rows => rows[0]);

    if (!apiKey) {
      throw new Error('API key not found');
    }

    if (apiKey.is_admin) {
      throw new Error('Cannot delete admin API key');
    }

    // Use a transaction to delete related records first
    return await db.transaction(async (tx) => {
      // Delete from api_threshold_requests
      await tx
        .delete(api_threshold_requests)
        .where(eq(api_threshold_requests.api_key_id, keyId));

      // Delete from api_request_logs
      await tx
        .delete(api_request_logs)
        .where(eq(api_request_logs.api_key_id, keyId));

      // Delete from api_key_permissions
      await tx
        .delete(api_key_permissions)
        .where(eq(api_key_permissions.api_key_id, keyId));

      // Finally delete the API key
      const [deletedKey] = await tx
        .delete(api_keys)
        .where(eq(api_keys.id, keyId))
        .returning();

      return deletedKey;
    });
  } catch (error) {
    // Rethrow known errors
    if (error.message === 'API key not found' || 
        error.message === 'Cannot delete admin API key') {
      throw error;
    }
    
    // Handle database errors
    console.error('Database error while deleting API key:', error);
    throw new Error('Failed to delete API key');
  }
}

// Get top IP addresses by request count
export async function getTopIpAddresses(limit = 10) {
  const thresholdEndpoints = [
    '/domains/by-threshold?threshold=51',
    '/domains/by-threshold?threshold=66',
    '/domains/by-threshold?threshold=80'
  ];

  return await db
    .select({
      ip_address: api_request_logs.ip_address,
      total_requests: sql`count(*)::int`,
      threshold_51: sql`count(CASE WHEN ${api_request_logs.endpoint} = '/domains/by-threshold?threshold=51' THEN 1 END)::int`,
      threshold_66: sql`count(CASE WHEN ${api_request_logs.endpoint} = '/domains/by-threshold?threshold=66' THEN 1 END)::int`,
      threshold_80: sql`count(CASE WHEN ${api_request_logs.endpoint} = '/domains/by-threshold?threshold=80' THEN 1 END)::int`,
      last_used: sql`max(${api_request_logs.timestamp})`
    })
    .from(api_request_logs)
    .where(
      and(
        isNotNull(api_request_logs.ip_address),
        inArray(api_request_logs.endpoint, thresholdEndpoints)
      )
    )
    .groupBy(api_request_logs.ip_address)
    .orderBy(desc(sql`count(*)`))
    .limit(limit);
}

