import { and, eq, sql } from "drizzle-orm";

import { db } from "../db.js";
import * as Schema from "../db/schema.js";
import { encryptString } from "../crypter.js";

export class AccountRepository {
  static async findById(accountId, serverId) {
    return await db.query.accounts.findFirst({
      where: and(
        eq(Schema.accounts.account_id, accountId),
        eq(Schema.accounts.server_id, serverId),
      ),
    });
  }

  static async findOrCreate({
    id,
    username,
    display_name,
    access_token,
    server,
  }) {
    const encryptedAccessToken = encryptString(access_token);
    const result = await db
      .insert(Schema.accounts)
      .values({
        server_id: server.id,
        account_id: id,
        username: username,
        display_name: display_name,
        access_token: encryptedAccessToken,
      })
      .onConflictDoUpdate({
        target: [Schema.accounts.server_id, Schema.accounts.account_id],
        set: {
          display_name: display_name,
          access_token: encryptedAccessToken,
          last_login_date: new Date(),
        },
      })
      .returning({ id: Schema.accounts.account_id });

    if (result.length === 1) {
      return result[0].id;
    } else {
      throw new Error(
        `Failed to findOrCreate Account record: ${id}, server: ${server.id}`,
      );
    }
  }

  static async deleteAccount(serverName, remoteAccountId) {
    if (!serverName || !remoteAccountId) {
      return;
    }

    await db.transaction(async (tx) => {
      const servers = await tx
        .select({
          id: Schema.servers.id,
          name: Schema.servers.server,
        })
        .from(Schema.servers)
        .where(eq(Schema.servers.server, serverName))
        .limit(1);

      if (servers.length !== 1) {
        return;
      }

      const serverId = servers[0].id;

      await tx
        .delete(Schema.accounts)
        .where(
          and(
            eq(Schema.accounts.server_id, serverId),
            eq(Schema.accounts.account_id, remoteAccountId),
          ),
        );

      const remainingAccounts = await tx
        .select({ remoteAccountId: Schema.accounts.account_id })
        .from(Schema.accounts)
        .where(eq(Schema.accounts.server_id, serverId))
        .limit(1);

      // If we have any accounts remaining, then replace the api_account_id with
      // one of the remaining account IDs
      if (remainingAccounts.length > 0) {
        await tx
          .update(Schema.server_config)
          .set({
            api_account_id: remainingAccounts[0].remoteAccountId,
          })
          .where(
            and(
              eq(Schema.server_config.server_id, serverId),
              eq(Schema.server_config.api_account_id, remoteAccountId),
            ),
          );
      } else {
        // Otherwise, wipe out the api_account_id, and disable pushes:
        await tx
          .update(Schema.server_config)
          .set({
            api_account_id: null,
            enabled: false,
          })
          .where(
            and(
              eq(Schema.server_config.server_id, serverId),
              eq(Schema.server_config.api_account_id, remoteAccountId),
            ),
          );
      }
    });
  }
}
