import punycode from "punycode/punycode.js";
import { eq, and, isNull, inArray, asc, sql } from "drizzle-orm";
import { db } from "../db.js";
import { getDomainDigest } from "../util.js";
import * as Schema from "../db/schema.js";

export class DomainsRepository {
  /**
   * Inserts a new domain into the database or updates the existing one if a conflict occurs on the domain field.
   *
   * @param {Object} params - The parameters for upserting a domain.
   * @param {string} params.domain - The domain to be inserted or updated in the database.
   * @param {string|null} [params.digest] - An optional digest of the domain. If not provided, it will be generated.
   * @returns {Promise<Object>} The inserted or updated domain record, including its id, domain, and digest.
   * @throws {Error} Throws an error if the upsert operation does not affect exactly one row.
   */
  static async upsert({ domain, digest }) {
    const domain_name = punycode.toASCII(domain);

    try {
      const result = await db
        .insert(Schema.domains)
        .values({
          domain: domain_name,
          digest: digest ?? getDomainDigest(domain),
        })
        // Don't use onConflictDoNothing since that aborts the upsert:
        .onConflictDoUpdate({
          target: Schema.domains.domain,
          set: {
            domain: domain_name,
          },
        })
        .returning({
          id: Schema.domains.id,
          domain: Schema.domains.domain,
          digest: Schema.domains.digest,
        });

      if (result.length === 1) {
        return result[0];
      } else {
        // we should never get here, but if we do, we want to know about it
        // we are throwing error here, becuase we are following repository pattern, calling of this method is expected to handle the error
        throw new Error(`DomainsRepository.upsert failed for: ${domain}`);
      }
    } catch (err) {
      throw new Error(`Error in upsert: ${err.message}`);
    }
  }

  static async find(domainId) {
    try {
      return await db.query.domains.findFirst({
        where: eq(Schema.domains.id, domainId),
      });
    } catch (err) {
      throw new Error(`Error finding domain by ID: ${err.message}`);
    }
  }

  static async findByDomain(domain, options) {
    const domain_name = punycode.toASCII(domain);
    try {
      return await db.query.domains.findFirst({
        where: eq(Schema.domains.domain, domain_name),
        ...options,
      });
    } catch (err) {
      throw new Error(`Error finding domain by name: ${err.message}`);
    }
  }

  static async findByDigest(digest) {
    try {
      return await db.query.domains.findFirst({
        where: eq(Schema.domains.digest, digest),
      });
    } catch (err) {
      throw new Error(`Error finding domain by digest: ${err.message}`);
    }
  }

   // this query is ordered ascending, because we want to prioritize the shortest
  static async findByServerAndVariants(serverId, variants) {
    if (variants.length === 0) {
      return [];
    }

    try {
      return await db
        .select({
          domain: Schema.domains.domain,
          server_id: Schema.blocks.server_id,
        })
        .from(Schema.blocks)
        .innerJoin(Schema.domains, eq(Schema.domains.id, Schema.blocks.domain_id))
        .where(
          and(
            eq(Schema.blocks.server_id, serverId),
            inArray(Schema.domains.domain, variants),
          ),
        )
        .orderBy(asc(sql`CHAR_LENGTH(${Schema.domains.domain})`));
    } catch (err) {
      throw new Error(`Error finding domains by server and variants: ${err.message}`);
    }
  }
}
