import { and, eq, sql } from "drizzle-orm";

import { db } from "../db.js";
import * as Schema from "../db/schema.js";
import { encryptString } from "../crypter.js";

export class ServerRepository {
  /**
   * @param {string} serverName
   */
  static async findByName(serverName) {
    try {
      const response = await db.query.servers.findFirst({
        where: eq(Schema.servers.server, serverName),
      });

      if (!response) {
        throw new Error(`Server not found in database: ${serverName}`);
      }

      return response;
    } catch (err) {
      throw new Error(`Error finding server by name: ${err.message}`);
    }
  }

  static async findAllByWipeRequested() {
    try {
      return await db.query.servers.findMany({
        where: eq(Schema.servers.block_wipe_in_progress, true),
      });
    } catch (err) {
      throw new Error(`Error finding servers with wipe requested: ${err.message}`);
    }
  }

  static async findAllPushEnabled() {
    try {
      return await db
        .select({
          server_id: Schema.servers.id,
          server: Schema.servers.server,
          api_account_id: Schema.server_config.api_account_id,
        })
        .from(Schema.servers)
        .innerJoin(
          Schema.server_config,
          eq(Schema.servers.id, Schema.server_config.server_id),
        )
        .where(
          and(
            eq(Schema.server_config.enabled, true),
            eq(Schema.servers.block_wipe_in_progress, false),
          ),
        );
    } catch (err) {
      throw new Error(`Error finding push-enabled servers: ${err.message}`);
    }
  }

  /**used for auth session, this find or create a new sever in db
   * @param {Object} params
   * @param {string} params.name
   * @param {string} params.url
   * @param {Object} params.credentials
   * @param {string} params.accountId
   */
  static async findOrCreate({ name, url, credentials, accountId }) {
    try {
      return await db.transaction(async (tx) => {
        const encryptedClientId = encryptString(credentials.client_id);
        const encryptedClientSecret = encryptString(credentials.client_secret);

        const serverId = await tx
          .insert(Schema.servers)
          .values({ server: name, url, verified: true })
          .onConflictDoUpdate({
            target: Schema.servers.server,
            set: {
              url,
              verified: true,
            },
          })
          .returning({
            id: Schema.servers.id,
          })
          .then((result) => {
            if (result.length === 1) {
              return result[0].id;
            } else {
              throw new Error(`Upsert failed for ${name}`);
            }
          });

        await tx
          .insert(Schema.server_config)
          .values({
            server_id: serverId,
            client_id: encryptedClientId,
            client_secret: encryptedClientSecret,
            api_account_id: accountId,
            list_type: "minimum",
          })
          .onConflictDoUpdate({
            target: Schema.server_config.server_id,
            set: {
              client_id: encryptedClientId,
              client_secret: encryptedClientSecret,
            },
          });

        return serverId;
      });
    } catch (err) {
      throw new Error(`Error in findOrCreate: ${err.message}`);
    }
  }

  static async updateApiAccount(serverId, accountId) {
    try {
      const accounts = await db
        .select({ access_token: Schema.accounts.access_token })
        .from(Schema.accounts)
        .where(
          and(
            eq(Schema.accounts.server_id, serverId),
            eq(Schema.accounts.account_id, accountId),
          ),
        );

      if (!accounts || !accounts[0].access_token) {
        const err = new Error(
          "Unable to update API Account as the selected account does not have a valid access token",
        );
        err.code = "INVALID_ACCOUNT";
        throw err;
      }

      await db
        .update(Schema.server_config)
        .set({ api_account_id: accountId })
        .where(eq(Schema.server_config.server_id, serverId));
    } catch (err) {
      throw new Error(`Error updating API account: ${err.message}`);
    }
  }

  static async updatePushStatus(serverId, newStatus) {
    try {
      await db
        .update(Schema.server_config)
        .set({ enabled: newStatus })
        .where(eq(Schema.server_config.server_id, serverId));
    } catch (err) {
      throw new Error(`Error updating push status: ${err.message}`);
    }
  }

  static async updateLastPushedAt(serverId) {
    try {
      await db
        .update(Schema.servers)
        .set({ last_server_push_date: sql`NOW()` })
        .where(eq(Schema.servers.id, serverId));
    } catch (err) {
      throw new Error(`Error updating last pushed at: ${err.message}`);
    }
  }

  static async deleteServer(serverId) {
    try {
      await this.updatePushStatus(serverId, false);

      await db.transaction(async (tx) => {
        await tx
          .delete(Schema.blocks)
          .where(eq(Schema.blocks.server_id, serverId));

        await tx
          .delete(Schema.server_config)
          .where(eq(Schema.server_config.server_id, serverId));

        await tx
          .delete(Schema.accounts)
          .where(eq(Schema.accounts.server_id, serverId));

        await tx.delete(Schema.servers).where(eq(Schema.servers.id, serverId));
      });
    } catch (err) {
      throw new Error(`Error deleting server: ${err.message}`);
    }
  }

  static async requestWipe(serverId) {
    try {
      await db.transaction(async (tx) => {
        await tx
          .update(Schema.server_config)
          .set({ enabled: false })
          .where(eq(Schema.server_config.server_id, serverId));

        await tx
          .update(Schema.servers)
          .set({ block_wipe_in_progress: true })
          .where(eq(Schema.servers.id, serverId));
      });
    } catch (err) {
      throw new Error(`Error requesting wipe: ${err.message}`);
    }
  }

  static async finishWipe(serverId) {
    try {
      await db
        .update(Schema.servers)
        .set({ block_wipe_in_progress: false })
        .where(eq(Schema.servers.id, serverId));
    } catch (err) {
      throw new Error(`Error finishing wipe: ${err.message}`);
    }
  }

  static async updateDenylist(serverId, listType) {
    try {
      await db
        .update(Schema.server_config)
        .set({ list_type: listType, enabled: false })
        .where(eq(Schema.server_config.server_id, serverId));
    } catch (err) {
      throw new Error(`Error updating denylist: ${err.message}`);
    }
  }
}
