import express from "express";
import * as AdminController from "../controllers/AdminController.js";
import { createApiKey } from "../queries.js";
import { db } from "../db.js";
import { api_keys } from "../db/schema.js";
import { eq } from "drizzle-orm";
import { adminAuthMiddleware } from '../middleware/adminAuthMiddleware.js';
import { adminApiKeyAuth } from '../middleware/apiKeyAuth.js';

export const adminRouter = express.Router();

// Apply admin auth middleware to all admin routes
adminRouter.use(adminAuthMiddleware);

// API key management endpoints
adminRouter.post('/api/keys/create', adminApiKeyAuth, async (req, res) => {
  try {
    const { description } = req.body;
    const apiKey = await createApiKey(description, false);
    
    res.json({
      success: true,
      message: 'API key created',
      key: apiKey.key
    });
  } catch (error) {
    res.status(500).json({
      success: false,
      error: 'Failed to create API key'
    });
  }
});

adminRouter.get('/api/keys', adminApiKeyAuth, async (req, res) => {
  try {
    const keys = await getApiUsageStats();
    res.json({ success: true, data: keys });
  } catch (error) {
    res.status(500).json({
      success: false,
      error: 'Failed to fetch API keys'
    });
  }
});

// Support legacy admin routes:
adminRouter.get("/", (req, res) => res.redirect("/admin/pending"));
adminRouter.get("/dashboard", (req, res) => res.redirect("/admin/pending"));
adminRouter.get("/domain_audits", (req, res) => res.redirect("/admin/audit"));

// New individual routes:
adminRouter.get("/pending", AdminController.pendingDomains);
adminRouter.get("/approved", AdminController.approvedDomains);
adminRouter.get("/inactive", AdminController.inactiveDomains);
adminRouter.get("/unknown", AdminController.unknownDomains);
adminRouter.get("/audit", AdminController.auditDomains);
adminRouter.get("/accounts", AdminController.accounts);

// Forms:
adminRouter.post("/approve/:domain_id", AdminController.approveDomain);
adminRouter.post("/unapprove/:domain_id", AdminController.unapproveDomain);

adminRouter.post("/activate/:domain_id", AdminController.activateDomain);
adminRouter.post("/deactivate/:domain_id", AdminController.deactivateDomain);

adminRouter.post("/approve_all", AdminController.approveAllPendingDomains);
adminRouter.post("/unapprove_all", AdminController.unapproveAllDomains);
adminRouter.post(
  "/deactivate_all",
  AdminController.deactivateAllPendingDomains,
);

// basic api endpoint for matching an unknown domain
adminRouter.post("/match/:digest", AdminController.guessDomainForDigest);

adminRouter.post(
  "/downloads/domain_audit",
  AdminController.downloadDomainAudit,
);

adminRouter.get("/error_logs", AdminController.errorLogsPage);

adminRouter.get("/domain-changes", AdminController.domainChanges);

adminRouter.post('/clear-error-logs', AdminController.clearErrorLogs);

adminRouter.get("/domain-history/:domain", AdminController.getDomainHistory);
