import express from 'express';
import {
  getDomainsAtLeast51Percent,
  getDomainsAtLeast66Percent,
  getDomainsAtLeast80Percent,
  getDomainsByMinimumThreshold,
  getApiUsageStats,
  createApiKey,
  getApiRequestStats,
  getApiKeyStats,
  getEndpointStats,
  getDailyStats,
  createApiKeyWithPermissions,
  getApiKeyPermissions,
  updateApiKeyPermissions,
  getRequestStatsByType,
  getThresholdUsageStats,
  toggleApiKeyStatus,
  incrementThresholdRequest,
  getApiKeyThresholdStats,
  getApiKeyDetails,
  getApiKeyRequestHistory,
  getApiKeyEndpointStats,
  getApiKeyDailyStats,
  getApiKeyHourlyDistribution,
  deleteApiKey,
  getTopIpAddresses
} from '../queries.js';
import { apiKeyAuth, adminApiKeyAuth } from '../middleware/apiKeyAuth.js';
import { getClientIP } from '../util.js';

const router = express.Router();

// Regular API endpoints with regular API key auth
router.get('/domains/by-threshold', apiKeyAuth, async (req, res) => {
  try {
    const { threshold, min } = req.query;
    const apiKeyId = req.apiKeyId;

    // Add debug logging
    console.log('Query parameters:', { threshold, min, apiKeyId });

    // If min is provided, use custom threshold
    if (min !== undefined) {
      const domains = await getDomainsByMinimumThreshold(Number(min));
      console.log('Custom threshold domains:', domains);
      return res.json({
        success: true,
        data: {
          threshold: 'custom',
          description: `Domains at or above ${min}%`,
          domains
        }
      });
    }

    let response = {};
    switch(threshold?.toLowerCase()) {
      case '51': {
        const domains = await getDomainsAtLeast51Percent();
        await incrementThresholdRequest(apiKeyId, '51');
        console.log('51% threshold domains:', domains);
        response = {
          threshold: '51',
          description: 'Domains at or above 51%',
          domains
        };
        break;
      }
      case '66': {
        const domains = await getDomainsAtLeast66Percent();
        await incrementThresholdRequest(apiKeyId, '66');
        console.log('66% threshold domains:', domains);
        response = { 
          threshold: '66',
          description: 'Domains at or above 66%',
          domains
        };
        break;
      }
      case '80': {
        const domains = await getDomainsAtLeast80Percent();
        await incrementThresholdRequest(apiKeyId, '80');
        console.log('80% threshold domains:', domains);
        response = {
          threshold: '80',
          description: 'Domains at or above 80%',
          domains
        };
        break;
      }
      default:
        // If no threshold specified, track all standard thresholds
        await Promise.all([
          incrementThresholdRequest(apiKeyId, '51'),
          incrementThresholdRequest(apiKeyId, '66'),
          incrementThresholdRequest(apiKeyId, '80')
        ]);
        
        response = {
          threshold_51: {
            description: 'Domains at or above 51%',
            domains: await getDomainsAtLeast51Percent()
          },
          threshold_66: {
            description: 'Domains at or above 66%',
            domains: await getDomainsAtLeast66Percent()
          },
          threshold_80: {
            description: 'Domains at or above 80%',
            domains: await getDomainsAtLeast80Percent()
          }
        };
    }

    res.json({
      success: true,
      data: response
    });

  } catch (error) {
    console.error('Error fetching domains by threshold:', error);
    res.status(500).json({
      success: false,
      error: 'Internal server error',
      details: error.message
    });
  }
});

// Admin endpoints with admin API key auth
router.post('/admin/api-keys/create', adminApiKeyAuth, async (req, res) => {
  try {
    const { description, permissions } = req.body;
    
    // Ensure permissions is an array
    const permissionsArray = Array.isArray(permissions) ? permissions : [];
    
    // Validate description
    if (!description) {
      return res.status(400).json({
        success: false,
        error: 'Description is required'
      });
    }

    // Validate that at least one permission is provided
    if (permissionsArray.length === 0) {
      return res.status(400).json({
        success: false,
        error: 'At least one permission must be specified',
        valid_permissions: ['threshold_51', 'threshold_66', 'threshold_80']
      });
    }

    // Validate that all provided permissions are valid
    const validPermissions = ['threshold_51', 'threshold_66', 'threshold_80'];
    const invalidPermissions = permissionsArray.filter(p => !validPermissions.includes(p));
    
    if (invalidPermissions.length > 0) {
      return res.status(400).json({
        success: false,
        error: 'Invalid permissions provided',
        invalid_permissions: invalidPermissions,
        valid_permissions: validPermissions
      });
    }

    const apiKey = await createApiKeyWithPermissions(description, permissionsArray);
    
    res.json({
      success: true,
      message: 'API key created',
      key: apiKey.key,
      description: apiKey.description,
      permissions: permissionsArray
    });
  } catch (error) {
    console.error('Error creating API key:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to create API key'
    });
  }
});

router.get('/admin/api-keys', adminApiKeyAuth, async (req, res) => {
  try {
    const apiKeys = await getApiUsageStats();
    
    // Get threshold stats for each API key
    const apiKeysWithThresholds = await Promise.all(apiKeys.map(async (key) => {
      const thresholdStats = await getApiKeyThresholdStats(key.key_id);
      
      return {
        ...key,
        threshold_requests: {
          threshold_51: thresholdStats.find(s => s.threshold === '51')?.request_count || 0,
          threshold_66: thresholdStats.find(s => s.threshold === '66')?.request_count || 0,
          threshold_80: thresholdStats.find(s => s.threshold === '80')?.request_count || 0
        }
      };
    }));
    
    res.json({
      success: true,
      data: apiKeysWithThresholds.map(key => ({
        key_id: key.key_id,
        description: key.description,
        key: key.key,
        created_at: key.created_at,
        last_used_at: key.last_used_at,
        total_requests: key.total_requests,
        is_active: key.is_active,
        is_admin: key.is_admin,
        permissions: key.permissions,
        threshold_requests: key.threshold_requests
      }))
    });
  } catch (error) {
    console.error('Error fetching API keys:', error);
    res.status(500).json({
      success: false,
      error: 'Internal server error',
      details: error.message
    });
  }
});

// Get overall API usage statistics
router.get('/admin/stats', adminApiKeyAuth, async (req, res) => {
  try {
    const stats = await getApiRequestStats();
    res.json({
      success: true,
      data: {
        total_requests: stats.total_requests,
        unique_keys: stats.unique_keys,
        active_keys: stats.active_keys,
        most_used_endpoint: stats.most_used_endpoint
      }
    });
  } catch (error) {
    console.error('Error fetching API stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch API statistics'
    });
  }
});

// Get stats for a specific API key
router.get('/admin/stats/key/:key_id', adminApiKeyAuth, async (req, res) => {
  try {
    const stats = await getApiKeyStats(parseInt(req.params.key_id));
    res.json({
      success: true,
      data: stats
    });
  } catch (error) {
    console.error('Error fetching key stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch key statistics'
    });
  }
});

// Get stats by endpoint
router.get('/admin/stats/endpoints', adminApiKeyAuth, async (req, res) => {
  try {
    const stats = await getEndpointStats();
    res.json({
      success: true,
      data: stats
    });
  } catch (error) {
    console.error('Error fetching endpoint stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch endpoint statistics'
    });
  }
});

// Gets daily usage stats
router.get('/admin/stats/daily', adminApiKeyAuth, async (req, res) => {
  try {
    const { days = 7 } = req.query; // Default to 7 days
    const stats = await getDailyStats(parseInt(days));
    res.json({
      success: true,
      data: stats
    });
  } catch (error) {
    console.error('Error fetching daily stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch daily statistics'
    });
  }
});

// this endpoint updates API key permissions
router.put('/admin/api-keys/:keyId/permissions', adminApiKeyAuth, async (req, res) => {
  try {
    const keyId = parseInt(req.params.keyId);
    const { permissions } = req.body;

    // Validate keyId
    if (isNaN(keyId)) {
      return res.status(400).json({
        success: false,
        error: 'Invalid API key ID format'
      });
    }

    // Validate permissions array
    if (!Array.isArray(permissions)) {
      return res.status(400).json({
        success: false,
        error: 'Permissions must be an array',
        valid_permissions: ['threshold_51', 'threshold_66', 'threshold_80']
      });
    }

    // Validate permission values
    const validPermissions = ['threshold_51', 'threshold_66', 'threshold_80'];
    const invalidPermissions = permissions.filter(p => !validPermissions.includes(p));
    
    if (invalidPermissions.length > 0) {
      return res.status(400).json({
        success: false,
        error: 'Invalid permissions provided',
        invalid_permissions: invalidPermissions,
        valid_permissions: validPermissions
      });
    }

    await updateApiKeyPermissions(keyId, permissions);
    
    res.json({
      success: true,
      message: 'Permissions updated successfully',
      key_id: keyId,
      permissions: permissions
    });
  } catch (error) {
    console.error('Error updating permissions:', error);
    
    // Handle specific errors
    if (error.message.includes('API key with ID')) {
      return res.status(404).json({
        success: false,
        error: 'API key not found'
      });
    }

    if (error.message.includes('Invalid API key ID')) {
      return res.status(400).json({
        success: false,
        error: 'Invalid API key ID'
      });
    }

    res.status(500).json({
      success: false,
      error: 'Failed to update permissions',
      details: env.isDevelopment ? error.message : undefined
    });
  }
});

// this endpoint fetches API key permissions
router.get('/admin/api-keys/:keyId/permissions', adminApiKeyAuth, async (req, res) => {
  try {
    const permissions = await getApiKeyPermissions(parseInt(req.params.keyId));
    
    res.json({
      success: true,
      data: permissions
    });
  } catch (error) {
    console.error('Error fetching permissions:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch permissions'
    });
  }
});

// this endpoint fetches request stats by method
router.get('/admin/stats/by-method', adminApiKeyAuth, async (req, res) => {
  try {
    const stats = await getRequestStatsByType();
    
    res.json({
      success: true,
      data: stats.map(stat => ({
        method: stat.method,
        total_requests: stat.total_requests,
        unique_keys: stat.unique_keys,
        last_used: stat.last_used,
        endpoints: stat.endpoints,
        success_rate: stat.success_rate
      }))
    });
  } catch (error) {
    console.error('Error fetching method stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch method statistics'
    });
  }
});

// this endpoint fetches threshold usage stats
router.get('/admin/stats/threshold-usage', adminApiKeyAuth, async (req, res) => {
  try {
    const stats = await getThresholdUsageStats();
    
    res.json({
      success: true,
      data: stats.map(stat => ({
        threshold: stat.threshold,
        total_requests: stat.total_requests,
        unique_keys: stat.unique_keys,
        last_used: stat.last_used,
        percentage: stat.percentage
      }))
    });
  } catch (error) {
    console.error('Error fetching threshold usage stats:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch threshold usage statistics'
    });
  }
});

// this endpoint to toggles API key status
router.post('/admin/api-keys/:keyId/toggle', adminApiKeyAuth, async (req, res) => {
  try {
    const keyId = parseInt(req.params.keyId);

    // Validate keyId
    if (isNaN(keyId)) {
      return res.status(400).json({
        success: false,
        error: 'Invalid API key ID format'
      });
    }

    const updatedKey = await toggleApiKeyStatus(keyId);
    
    res.json({
      success: true,
      message: `API key ${updatedKey.is_active ? 'activated' : 'deactivated'} successfully`,
      key_id: keyId,
      is_active: updatedKey.is_active
    });
  } catch (error) {
    console.error('Error toggling API key status:', error);
    
    // Handle specific errors
    if (error.message.includes('API key with ID')) {
      return res.status(404).json({
        success: false,
        error: 'API key not found'
      });
    }

    if (error.message.includes('Cannot toggle admin API key')) {
      return res.status(403).json({
        success: false,
        error: 'Cannot modify admin API key status'
      });
    }

    res.status(500).json({
      success: false,
      error: 'Failed to toggle API key status',
      details: env.isDevelopment ? error.message : undefined
    });
  }
});

// Add this new endpoint to get detailed API key stats
router.get('/admin/api-keys/:keyId/details', adminApiKeyAuth, async (req, res) => {
  try {
    const keyId = parseInt(req.params.keyId);

    // Validate keyId
    if (isNaN(keyId)) {
      return res.status(400).json({
        success: false,
        error: 'Invalid API key ID format'
      });
    }

    // Get all the data in parallel
    const [
      apiKey,
      permissions,
      thresholdStats,
      requestHistory,
      endpointStats,
      dailyStats,
      hourlyDistribution
    ] = await Promise.all([
      getApiKeyDetails(keyId),
      getApiKeyPermissions(keyId),
      getApiKeyThresholdStats(keyId),
      getApiKeyRequestHistory(keyId),
      getApiKeyEndpointStats(keyId),
      getApiKeyDailyStats(keyId),
      getApiKeyHourlyDistribution(keyId)
    ]);

    if (!apiKey) {
      return res.status(404).json({
        success: false,
        error: 'API key not found'
      });
    }

    res.json({
      success: true,
      data: {
        api_key: apiKey,
        permissions: permissions.map(p => p.permission),
        threshold_usage: thresholdStats.reduce((acc, stat) => {
          acc[`threshold_${stat.threshold}`] = {
            requests: stat.request_count,
            last_used: stat.last_used
          };
          return acc;
        }, {}),
        usage_stats: {
          total_requests: requestHistory.length,
          endpoints: endpointStats,
          daily_stats: dailyStats,
          hourly_distribution: hourlyDistribution
        },
        recent_requests: requestHistory.map(req => ({
          endpoint: req.endpoint,
          method: req.method,
          timestamp: req.timestamp,
          ip_address: getClientIP(req),
          user_agent: req.user_agent || 'unknown'
          
        }))
      }
    });

  } catch (error) {
    console.error('Error fetching API key details:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch API key details',
      details: process.env.NODE_ENV === 'development' ? error.message : undefined
    });
  }
});

// Add this new endpoint to delete an API key
router.delete('/admin/api-keys/:keyId', adminApiKeyAuth, async (req, res) => {
  try {
    const keyId = parseInt(req.params.keyId);

    // Validate keyId
    if (isNaN(keyId) || keyId <= 0) {
      return res.status(400).json({
        success: false,
        error: 'Invalid API key ID format. Must be a positive integer.'
      });
    }

    await deleteApiKey(keyId);
    
    res.json({
      success: true,
      message: 'API key deleted successfully',
      key_id: keyId
    });

  } catch (error) {
    console.error('Error deleting API key:', error);
    
    // Handle specific error cases
    switch(error.message) {
      case 'API key not found':
        return res.status(404).json({
          success: false,
          error: 'API key not found'
        });
      
      case 'Cannot delete admin API key':
        return res.status(403).json({
          success: false,
          error: 'Cannot delete admin API key'
        });
      
      default:
        return res.status(500).json({
          success: false,
          error: 'Failed to delete API key',
          details: process.env.NODE_ENV === 'development' ? error.message : undefined
        });
    }
  }
});

// Add this test endpoint temporarily
router.get('/test-headers', (req, res) => {
  res.json({
    headers: req.headers,
    clientIp: req.clientIp,
    expressIp: req.ip,
    ips: req.ips,
    cloudflareConnectingIp: req.headers['cf-connecting-ip'],
    xForwardedFor: req.headers['x-forwarded-for'],
    xRealIp: req.headers['x-real-ip'],
    doConnectingIp: req.headers['do-connecting-ip'],
    utilGetClientIP: getClientIP(req),
    rawSocket: req.socket?.remoteAddress
  });
});

// Add this new endpoint after the other admin stats endpoints
router.get('/admin/stats/top-ips', adminApiKeyAuth, async (req, res) => {
  try {
    const { limit = 10 } = req.query;
    
    // Validate limit
    const parsedLimit = parseInt(limit);
    if (isNaN(parsedLimit) || parsedLimit < 1 || parsedLimit > 100) {
      return res.status(400).json({
        success: false,
        error: 'Invalid limit. Must be a number between 1 and 100'
      });
    }

    const stats = await getTopIpAddresses(parsedLimit);
    
    res.json({
      success: true,
      data: stats.map(stat => ({
        ip_address: stat.ip_address,
        total_requests: stat.total_requests,
        requests_by_threshold: {
          '51%': stat.threshold_51,
          '66%': stat.threshold_66,
          '80%': stat.threshold_80
        },
        last_used: stat.last_used
      }))
    });
  } catch (error) {
    console.error('Error fetching top IP addresses:', error);
    res.status(500).json({
      success: false,
      error: 'Failed to fetch top IP statistics',
      details: process.env.NODE_ENV === 'development' ? error.message : undefined
    });
  }
});

export default router; 