import crypto from "crypto";
import { redis } from "./redis.js";
import { api } from "./api.js";
import {
  getServerConfig,
  wipeAccessTokens,
  getAccessTokens,
  updateServerAccountID,
  wipeAccountAccessToken,
  updateLastAccountTokenVerifyDate,
  getAccount,
} from "./queries.js";
import * as constants from "./constants.js";
import { decryptString, encryptString } from "./crypter.js";
import requestIp from 'request-ip';

export const checkServerPermissions = (serverName) => {
  if (constants.ALLOWED_SERVERS.length > 0) {
    // if this is an allowlisted server, only those on the list can pass
    if (!constants.ALLOWED_SERVERS.includes(serverName)) {
      return false;
    }
  }

  if (constants.DENIED_SERVERS.length > 0) {
    // otherwise, we then need to confirm if this is a denied server
    // this is mutually exclusive by design, you can't have both
    // allowlist and denylists for access to this application
    if (constants.DENIED_SERVERS.includes(serverName)) {
      return false;
    }
  }

  return true;
};

/**
 * @typedef ClientCredentials
 * @type {object}
 * @property {string} client_id
 * @property {string} client_secret
 */

/**
 * returns the encrypted credentials as pulled from the cache, if they are not in cache, then it checks the db, if found in the db, applied to redis cache
 * @param {string} server
 * @returns {Promise<ClientCredentials|undefined>}
 */
export async function getClientCredentials(server) {
  const credentials = await redis.hgetall(`credentials:${server}`);

  if (credentials && credentials?.client_id && credentials?.client_secret) {
    return {
      client_id: decryptString(credentials.client_id),
      client_secret: decryptString(credentials.client_secret),
    };
  }

  // Fetch the server config if we have it:
  const server_config = await getServerConfig(server);

  // Cache the credentials if we have server config:
  if (
    server_config &&
    server_config?.client_id &&
    server_config?.client_secret
  ) {
    /** @type {ClientCredentials} */
    const existingCredentials = {
      client_id: decryptString(server_config.client_id),
      client_secret: decryptString(server_config.client_secret),
    };

    // We don't care if setting the credentials cache throws:
    await setClientCredentials(server, existingCredentials).catch(() => {});

    return existingCredentials;
  }

  return undefined;
}

export async function setClientCredentials(server, credentials) {
  return await redis.hset(`credentials:${server}`, {
    client_id: encryptString(credentials.client_id),
    client_secret: encryptString(credentials.client_secret),
  });
}

export async function revokeAccessTokens(server_config) {
  const client_id = decryptString(server_config.client_id);
  const client_secret = decryptString(server_config.client_secret);

  const access_tokens = await getAccessTokens(server_config.server);

  // cleanly revoke access_tokens at the server
  try {
    for (const token of access_tokens) {
      const account = await getAccount(
        server_config.server_id,
        token.account_id,
      );
      await api.revoke_access_token(
        client_id,
        client_secret,
        server_config.server,
        decryptString(token.access_token),
        account,
      );
    }
  } catch (e) {
    console.error("error revoking access tokens", e);
  } finally {
    // wipe access tokens from the database
    await wipeAccessTokens(server_config.server);
  }
}

// designate a new account to be used for this server
// in the event the old access token was broken
// return true if a valid access token was found,
// false if not
export async function rotateAccessToken(server) {
  const server_config = await getServerConfig(server);
  let rotated = false;

  if (server_config && server_config.client_id && server_config.client_secret) {
    const client_id = decryptString(server_config.client_id);
    const client_secret = decryptString(server_config.client_secret);

    const access_tokens = await getAccessTokens(server);

    // // verify the credentials to retrieve an app-level token for the server used to verify the app credentials
    const verified = await api.get_client_credentials(server, {
      client_id,
      client_secret,
    });

    // this client_id itself is invalid
    // returning false will trigger a hard logout and purge of credentials
    if (verified?.error) {
      return false;
    }

    // check each token to see if it is valid
    for (const token of access_tokens) {
      if (token?.access_token) {
        const account = await getAccount(
          server_config.server_id,
          token.account_id,
        );
        const verified = await api.verify_app_credentials(
          server,
          decryptString(token.access_token),
          account,
          console,
        );

        if (verified?.error) {
          // we cannot even wipe this token as our access is broken
          // so we'll remove it from the db and try the next one
          await wipeAccountAccessToken(
            server_config.server_id,
            token.account_id,
          );
        } else if (!rotated) {
          // update api_account_id in config table to this valid token
          await updateServerAccountID(
            server_config.server_id,
            token.account_id,
          );

          rotated = true;
        }

        await updateLastAccountTokenVerifyDate(
          server_config.server_id,
          token.account_id,
        );
      }
    }
  }

  return rotated;
}

export async function purgeRedisClientCredentials(server) {
  await redis.del(`credentials:${server}`);
}

export async function sleep(ms) {
  return new Promise((resolve) => setTimeout(resolve, ms));
}

export function friendlySeverity(severity) {
  return severity === "noop"
    ? "None"
    : severity === "silence"
      ? "Limit"
      : "Suspend";
}

export function getDomainDigest(domain) {
  return crypto.createHash("sha256").update(domain).digest("hex").toString();
}

export function sanitizeServerInput(server) {
  if (server.startsWith("https://")) {
    try {
      return getServerUrl(server);
    } catch {
      return "";
    }
  }

  // Append the default protocol and try to parse:
  try {
    return getServerUrl(`https://${server}`);
  } catch {
    return "";
  }
}

function getServerUrl(input) {
  const serverUrl = new URL(input);
  serverUrl.pathname = "/";
  serverUrl.search = "";
  serverUrl.hash = "";

  return serverUrl.href;
}

// en-CA uses hyphens for the parts in the YYYY-MM-DD format:
const dateFormatter = new Intl.DateTimeFormat("en-CA", {
  year: "numeric",
  month: "2-digit",
  day: "2-digit",
  hour: "2-digit",
  minute: "2-digit",
  second: "2-digit",
  hourCycle: "h24",
  timeZoneName: "short",
  timeZone: "UTC",
});

export function formatPrivateComment() {
  const date = new Date();
  return `${constants.CLIENT_NAME} updated ${dateFormatter.format(date)}`;
}

export const getClientIP = (req) => {
  // If it's a stored request with just ip_address, return it
  if (req.ip_address) {
    return req.ip_address;
  }

  // If it's a live request, check headers in order of preference
  if (req.headers) {
    return req.headers['do-connecting-ip'] || 
           req.ip || 
           req.headers['x-forwarded-for'] || 
           req.socket?.remoteAddress ||
           'unknown';
  }

  // If neither case matches, return unknown
  return 'unknown';
};

// Helper function to validate IP address
function isValidIP(ip) {
  if (!ip) return false;
  
  // Remove IPv6 prefix if present
  if (ip.startsWith('::ffff:')) {
    ip = ip.substring(7);
  }
  
  // Basic IPv4 validation
  const ipv4Regex = /^(\d{1,3}\.){3}\d{1,3}$/;
  if (!ipv4Regex.test(ip)) return false;
  
  // Validate each octet
  const octets = ip.split('.');
  return octets.every(octet => {
    const num = parseInt(octet, 10);
    return num >= 0 && num <= 255;
  });
}

// Helper function to parse X-Forwarded-For header
function parseXForwardedFor(value) {
  if (!value) return null;
  
  // Split the string and clean each IP
  const ips = value.split(',').map(ip => {
    ip = ip.trim();
    // Handle IP:port format
    if (ip.includes(':')) {
      const [ipPart] = ip.split(':');
      return ipPart;
    }
    return ip;
  });
  
  // Return the first valid IP
  return ips.find(ip => isValidIP(ip)) || null;
}

// export function getClientIP(req) {
//   return req.headers['do-connecting-ip'] || 'unknown';
// }

// Helper function to check for private IP ranges
function isPrivateIP(ip) {
  // Convert IPv6 format if necessary
  if (ip.startsWith('::ffff:')) {
    ip = ip.substring(7);
  }
  
  const parts = ip.split('.').map(part => parseInt(part, 10));
  
  return (
    parts[0] === 10 || // 10.0.0.0/8
    (parts[0] === 172 && parts[1] >= 16 && parts[1] <= 31) || // 172.16.0.0/12
    (parts[0] === 192 && parts[1] === 168) || // 192.168.0.0/16
    parts[0] === 127 // 127.0.0.0/8
  );
}
