import { db } from "../db.js";
import { captureException } from '@sentry/node';
import { errorLogs } from "../db/schema.js"; // Import the errorLogs table schema

/**
 * Logs an error to the database and captures it with Sentry.
 *
 * @param {Error} err - The error object.
 * @param {Object} context - Additional context for the error.
 */
export async function logErrorToDatabase(err, context = {}) {
  try {
    // Log the error object to the console with distinct styling
    console.log("\x1b[31m%s\x1b[0m", "=== Error Object ===");
    console.dir(err, { depth: null, colors: true });
    console.log("\x1b[31m%s\x1b[0m", "====================");

    // Extract error components
    const errorMessage = err.message || "No message available";
    const errorStack = err.stack || "No stack trace available";
    const errorView = err.view ? JSON.stringify(err.view) : "No view data";

    // Capture the exception with Sentry
    captureException(err);

    // Extract userId from context
    const userId = context.userId || null;

    // Insert the error into the database
    await db.insert(errorLogs).values({
      message: errorMessage,
      severity: determineSeverity(err),
      context: JSON.stringify({
        ...context,
        stack: errorStack,
        view: errorView,
      }),
      user_id: userId, // Use the extracted userId
    });
  } catch (dbErr) {
    console.error("Failed to log error to database:", dbErr);
  }
}

/**
 * Determines the severity of an error.
 *
 * @param {Error} err - The error object.
 * @returns {string} The severity level of the error.
 */
function determineSeverity(err) {
  const errorString = err.message || JSON.stringify(err, Object.getOwnPropertyNames(err));
  if (errorString.includes("fatal") || errorString.includes("crash")) {
    return "critical";
  } else if (errorString.includes("timeout") || errorString.includes("failed")) {
    return "high";
  } else if (errorString.includes("deprecated") || errorString.includes("warning")) {
    return "medium";
  }
  return "low";
}

