import assert from "node:assert";
import { after, describe, it } from "node:test";
import { disconnect } from "./helpers.js";
import {
  generateSecretKey,
  decryptString,
  encryptString,
} from "../src/crypter.js";

describe("generateSecretKey", () => {
  after(async () => {
    await disconnect();
  });

  it("generates a different secret keys for different salts", () => {
    const a = generateSecretKey("test salt 1");
    const b = generateSecretKey("test salt 2");
    assert.notStrictEqual(a, b);
  });

  it("generates a repeatable secret key with default key length", () => {
    const a = generateSecretKey("test salt");
    const b = generateSecretKey("test salt");

    assert.strictEqual(a, b);
  });

  it("generates the same keys for the same key lengths and same salt", () => {
    const a = generateSecretKey("test salt", 128);
    const b = generateSecretKey("test salt", 128);

    assert.strictEqual(a, b);
  });

  it("generates different keys for differ key lengths but same salt", () => {
    const a = generateSecretKey("test salt", 68);
    const b = generateSecretKey("test salt", 128);

    assert.notStrictEqual(a, b);
  });
});

describe("database encryption", () => {
  describe("encryptString", () => {
    it("returns a string", () => {
      const encrypted = encryptString("hello world");

      assert.ok(typeof encrypted === "string");
      assert.ok(encrypted.includes(":"));
    });

    it("returns a string containing the encrypted data, iv, and authTag", () => {
      const encrypted = encryptString("hello world");
      const parts = encrypted.split(":");

      assert.strictEqual(parts.length, 3);
    });

    it("uses a unique iv for each encryption", () => {
      const a = encryptString("hello world");
      const b = encryptString("hello world");

      assert.notStrictEqual(a, b);

      // Extract the iv from the encrypted strings:
      const a_iv = a.split(":", 3)[1];
      const b_iv = b.split(":", 3)[1];

      assert.notStrictEqual(a_iv, b_iv);
    });
  });

  describe("decryptString", () => {
    it("decrypts a value from encryptString", () => {
      const encrypted = encryptString("hello world");
      const decrypted = decryptString(encrypted);

      assert.strictEqual(decrypted, "hello world");
    });

    it("fails to decrypt a value if the encrypted value has been tampered with", () => {
      const encrypted = encryptString("hello world");
      const tampered = "tamper" + encrypted.slice(6);

      assert.throws(
        () => {
          decryptString(tampered);
        },
        {
          message: /Unsupported state or unable to authenticate data/,
        },
      );
    });

    it("fails to decrypt a value if the auth tag has been tampered with", () => {
      const encrypted = encryptString("hello world");
      const tampered = encrypted.slice(0, -6) + "tamper";

      assert.throws(
        () => {
          decryptString(tampered);
        },
        {
          message: /Unsupported state or unable to authenticate data/,
        },
      );
    });
  });
});
