import assert from "node:assert";
import { describe, it, after } from "node:test";
import { disconnect } from "./helpers.js";
import { parseCsv, writeCsv } from "../src/csv.js";

after(async () => {
  disconnect();
});

describe("csv parsing", (t) => {
  describe("Mastodon CSV", () => {
    const data = `
#domain,#severity,#reject_media,#reject_reports,#public_comment,#obfuscate
076.moe,suspend,true,false,,false
076.ne.jp,suspend,false,true,,false
,,,,,
101010.pl,suspend,false,false,"Test Comment",true
    `.trim();

    it("parses successfully and does not reject", () => {
      assert.doesNotReject(parseCsv(data, "mastodon_csv"));
    });

    it("parses the correct number of domain blocks", (t, done) => {
      parseCsv(data, "mastodon_csv").then((domain_blocks) => {
        console.log({ domain_blocks });
        assert.strictEqual(domain_blocks.length, 3);
        done();
      });
    });
  });

  describe("FediBlockHole CSV", () => {
    const data = `
"domain","severity","reject_media","reject_reports","private_comment","public_comment","obfuscate"
"qoto.org","suspend",True,True,,,True
"sealion.club","suspend",True,True,,,True
"develop.gab.com","suspend",True,True,,,True
"gab.ai","suspend",True,True,,,True
`.trim();

    it("parses successfully and does not reject", () => {
      assert.doesNotReject(parseCsv(data, "fediblockhole"));
    });

    it("parses the correct number of domain blocks", (t, done) => {
      parseCsv(data, "fediblockhole").then((domain_blocks) => {
        console.log({ domain_blocks });
        assert.strictEqual(domain_blocks.length, 4);
        done();
      });
    });
  });

  describe("Domain Only CSV", () => {
    const data = `
13bells.com
1611.social
4aem.com
5dollah.click
adachi.party
adtension.com
annihilation.social
`.trim();

    it("parses successfully and does not reject", () => {
      assert.doesNotReject(parseCsv(data, "domain_only_csv"));
    });

    it("parses the correct number of domain blocks", (t, done) => {
      parseCsv(data, "domain_only_csv").then((domain_blocks) => {
        console.log({ domain_blocks });
        assert.strictEqual(domain_blocks.length, 7);
        done();
      });
    });
  });
});

describe("csv writing", () => {
  it("correctly creates a CSV string", async () => {
    const domains = [
      {
        domain: "piazza.today",
        severity: "suspend",
        reject_media: false,
        reject_reports: false,
        public_comment: "",
        obfuscate: false,
      },
      {
        domain: "pibvt.net",
        severity: "silence",
        reject_media: false,
        reject_reports: false,
        public_comment: "",
        obfuscate: false,
      },
      {
        domain: "pieville.net",
        severity: "suspend",
        reject_media: true,
        reject_reports: false,
        public_comment: "Test public comment",
        obfuscate: false,
      },
    ];

    assert.doesNotReject(() => writeCsv(domains));

    const csv = await writeCsv(domains);

    assert.equal(
      csv,
      "#domain,#severity\npiazza.today,suspend\npibvt.net,silence\npieville.net,suspend\n",
    );
  });
});
